/*
 * Layer.java
 *
 * Created on August 9, 2000, 3:41 PM
 */
 
package saaps.ann.mlff;

import saaps.ann.Data;
import java.io.Serializable;

/** 
 * This class represents a layer in a multi layer feed forward neural
 * network. The layer consist of a collection of <code>Neuron</class>
 * objects.
 *
 * @author  Peter Wintoft
 * @version 1.0 2000-08-11
 * @see Neuron
 * @see Network
 */
public class Layer extends Object implements Serializable {

  //Array of neurons.
  private Neuron[] neurons;
  //The number of neurons.
  private int numOfNeurons;
  //The input data.
  private transient Data inputData;
  //The output data.
  private transient Data outputData;
  //Flag indicating whether the layer has been run for the new data.
  private transient boolean hasBeenRun = false;
  
  /** Creates new Layer */
  public Layer() {
  }
  
  /**
  Create a layer of neurons.
  
  @param  n The number of neurons.
  */
  public Layer(int n)
  {
    numOfNeurons = n;
    neurons = new Neuron[n];
  }
  
  /**
  Create a layer of neurons.
  
  @param  n An array of neurons
  */
  public Layer(Neuron[] n)
  {
    neurons = n;
    numOfNeurons = n.length;
  }
  
  /**
  Get a neuron from the layer.
  
  @param  n The neuron index.
  @return The neuron.
  */
  public Neuron getNeuron(int n) {
    return neurons[n];
  }
  
  /**
  Get the number of neurons.
  
  @return The number of neurons.
  */
  public int getNumOfNeurons() {
    return neurons.length;
  }
  
  /**
  Set the input data to the layer.
  
  @param  d The input data
  */
  public void setInput(Data d)
  {
    inputData = d;
    hasBeenRun = false;
  }
  
  /**
  Get the output data from the layer.
  <p>
  If new input data has been presented to the layer then
  the layer is first run.
  <p>
  The input data to the layer is distributed to each neuron in the
  layer. Each neuron then processes the data. The number of outputs
  from the layer equals the number of neurons in the layer.
  
  @return The output data from the layer
  */
  public Data getOutput()
  {
    if(hasBeenRun)
      return outputData;
    else
    {
      run();
      return outputData;
    }
  }
  
  /**
  Run the layer.
  <p>
  The input data to the layer is distributed to each neuron in the
  layer. Each neuron then processes the data. The number of outputs
  from the layer equals the number of neurons in the layer.
  */
  private void run()
  {
    outputData = new Data(numOfNeurons,inputData.getNumOfExamples());
    for (int n=0;n<numOfNeurons;n++)
    {
      neurons[n].setInput(inputData);
      Data out = neurons[n].getOutput();
      outputData.setConnection(n,out.getConnection(0));
    }
    hasBeenRun = true;
  }
  
  public String toString() {
    StringBuffer buf = new StringBuffer(this.getClass().getName() + "\n");
    buf.append("Number of neurons = " + numOfNeurons + "\n");
    buf.append("Has been run = " + hasBeenRun + "\n");
    return buf.toString();
  }

}