package saaps.dbt;

import java.io.*;
import java.util.*;
import java.text.*;

import saaps.util.*;

/**
 * ACEDB extends the Database class. It contains
 * the read and write methods that are used in
 * Database. The field is a vector of ACE objects.
 *
 *@authors Peter Wintoft, Leif Kalla
 *@version 1.1 1999-12-21 (Use new ACEmag class)
 *@version 1.0 990913
 *@see ACEmag.html
 */
public class ACEmagDB extends Database {

  private double dataGapValueDefinition = -999.9;

  // Static Fields used to initiate the MetaData
  private static String name="ACEmagDB";
  private static String description="ACE magnetic field data";
  private static String longDescription="The ACE magnetic field data consist of " +
    "1 minute resolution solar wind magnetic field data. " +
    "The components are Bx, By, and Bz in GSM coordinate system. " +
    "The magnitude (Bt) and the direction of the magnetic field are also included.";

  private static Calendar st=Calendar.getInstance(TimeZone.getTimeZone("UTC"),new Locale("sv","SE"));
  private static Calendar et=Calendar.getInstance(TimeZone.getTimeZone("UTC"),new Locale("sv","SE"));
  private static boolean ct=true;
  private static long step=60000; //One minute timestep 
  private static String[] fieldNames = { "Bx" , "By" , "Bz" , "Bt" , "Latitude" , "Longitude" };
  private static String[] fieldUnits = { "nT" , "nT" , "nT" , "nT" , "Deg"      , "Deg"};
  private static String[] fieldTypes = { "Double", "Double", "Double","Double","Double","Double" };

  // Fields.
  private ACEmag ace;
  private Vector acev;

  // Constructors.
  public ACEmagDB() {
    //
    dataGapValue=dataGapValueDefinition;
  }
  public ACEmagDB(int i) {
    //
    dataGapValue=dataGapValueDefinition;
    //    acev = ACEV;
    //store(acev);
  }
  /**
   * When a new class is added to the database this
   * constructor should be used which includes the
   * information that goes into the meta data file.
   *
   * The arguments are: Vector of ACE objects, the class name,
   * a short description, continuos time series (false/true), and
   * if a continous time series the time step.
   */
  public ACEmagDB(Vector ACEV) throws IOException {
    //, String des, boolean ct, long step) {
    //
    //
    dataGapValue=dataGapValueDefinition;
    acev = ACEV;
    raf = new RandomAccessFile(baseDir + "ACEmagDB.data","rw");
    // Then get its metadata
    try { // To load the meta data for this database from file
      meta = new MetaData("ACEmagDB");
      isSet=true;
    }
    catch (IOException e) {
      System.err.println(e);
      System.out.println("Here we should handle the situation when no DB exists");
      try {
  meta = setDefaultMeta(((DatabaseObject) (acev.elementAt(0))).getDate(),((DatabaseObject) (acev.elementAt(acev.size()-1))).getDate()); //Takes static fields from XDB

  isSet=true;
      }
      catch (IOException ee) {
  System.err.println(ee);
  ee.printStackTrace();
      }
    }
    catch (ClassNotFoundException e) {
      System.err.println(e);
      e.printStackTrace();
    }

    store((Vector) acev); //Need to type cast so that store(Object obj) isn't used
  }

  // Methods.
  public Object read() throws IOException {
    ace = new ACEmag(raf.readLong(),raf.readDouble(),raf.readDouble(),raf.readDouble(),raf.readDouble(),raf.readDouble(),raf.readDouble());
    return (Object) ace;
  }

  public void myHello() {
    System.out.println("My ACEmagDB Hello");
  }

  public void write(Object obj) throws IOException {
    ace = (ACEmag) obj;
    raf.writeLong(ace.getDate().getTime());
    raf.writeDouble(ace.getBx());
    raf.writeDouble(ace.getBy());
    raf.writeDouble(ace.getBz());
    raf.writeDouble(ace.getBt());
    raf.writeDouble(ace.getLat());
    raf.writeDouble(ace.getLong());
  }
  // Implemented for Database abstract getDate
  public Date getDate(Object obj) {
    return ((ACEmag) obj).getDate();
  }

  /** This method is implemented for Database abstract metod.
   * It retrieves the latest data from the dataprovider and returns the vector of Database objects.
   */
  public Vector getLatestDatabaseObjects() {
    Vector vec = new Vector(8*30);
    Date mdate = new Date();

    ACEmag ace = new ACEmag(mdate,1.0,2.0,3.0,4.0,5.0,6.0);

    for(int i=0;i<8*30;i++) {
      vec.addElement(ace);
      mdate.setDate( (int) (mdate.getDate() + 10));
      ace.setDate(mdate);
      ace.setBx(i);
      ace.setBy(i);
      ace.setBz(i);
    }
    vec.trimToSize();
    return vec;
  }

  /** 
   * This method is used for importing ASCII data to our database.
   * The format of the data is as output from ExportAscii application.
   * It will fill a buffer of databaseobjects and then try to store it,
   * read again and store until it is finished or an exception occures.
   */
  protected void importAscii1(String inputfile) throws Exception {
    final int MAXSIZE=1048576; // bytes , 1Mb 
    //final int MAXSIZE=1024; // bytes ,

    Date inputDateTime;
    String tmpVal;
    double tmpKp;


    // Create a Vector of Kp objects.
    Vector vec = new Vector(8*30);

    try {
      //Name of database objects
      String objectName = meta.getName().substring(0, meta.getName().length()-2);
      short numFields = (short) meta.getFieldNames().length;
      double[] data = new double[numFields];

      //set default maximum number of objects
      int maxObjs = 100;
      if(meta.getCont()) { //Calculate if we have continous data
	long objSize = DatabaseObject.getInstance(objectName,new Date()).getObjectSize();
	maxObjs = (int) (MAXSIZE/objSize);
      }
      int cnt=0;

      // Input stream.
      BufferedReader streamIn = new BufferedReader(new FileReader(inputfile));
      String textLine = "";
      textLine = streamIn.readLine();
      // Go through the file for all lines starting with
      // "#", ":", or " ".
      while (!textLine.startsWith("1") && !textLine.startsWith("2")) {
	textLine = streamIn.readLine();
      }

      while(textLine != null) {
	while ((textLine != null) && ( cnt < maxObjs) ) {
	  cnt++;
	  // Read the date using - as the field separator.
	  String dateStr = textLine.substring(0,4) + "." +
	    textLine.substring(5,7) + "." +
	    textLine.substring(8,10) + " " + textLine.substring(11,13) + ":" + textLine.substring(14,16) + ":" + textLine.substring(17); //Last item to end of string

	  SimpleDateFormat formatter = (SimpleDateFormat) SimpleDateFormat.getDateTimeInstance(0,0,new Locale("sv","SE"));

	  formatter.setTimeZone(TimeZone.getTimeZone("UTC"));
	  formatter.applyPattern("yyyy.MM.dd HH:mm:ss");
     
	  // Parse the string back into a Date.
	  ParsePosition pos = new ParsePosition(0);
	  inputDateTime = formatter.parse(dateStr, pos);
	
	// Assume that observations are at 00, 03, 06, ...
	  Calendar cal = Calendar.getInstance();
	  cal.setTimeZone(formatter.getTimeZone());
	  cal.setTime(inputDateTime);
	  //The data value
	  tmpVal = textLine.substring(20).trim();
	  StringTokenizer tokens = new StringTokenizer(tmpVal);
	  int numTok = tokens.countTokens(); 
	  for(int i=0;i<numTok;i++) {
	    String value = tokens.nextToken();
	    System.out.println(value);

	    if(value.equals(new String("NaN"))) {
	      data[i] = Double.NaN;
	    }
	    else {
	      data[i] = Double.valueOf(value).doubleValue();
	    }
	    if (data[i] == dataGapValue ) {
	      data[i] = Double.NaN;
	    }
	  }

	  //System.out.println("Date/tome:" + formatter.format(inputDateTime) + " " + tmpKp);
	  vec.addElement(DatabaseObject.getInstance(objectName,cal.getTime(),data));
	  //Get the next line
	  textLine = streamIn.readLine();	
	}
	//Store our buffer of objects
	vec.trimToSize();
	// Store the data in the database
	ObjectTimeSeries ts = new ObjectTimeSeries(vec);

	if(!isSet())
	  System.out.println("MetaData is unavailable \nWill try to create it.");
	System.out.println("Storing:" + cnt);
	store(ts);

	//Reset the vector and counter
	//vec.removeAllElements();
	vec = new Vector(8*30);
	cnt=0;
      }
      streamIn.close();
    }
    catch (IOException e) {
      System.err.println("IO error: " + e);
    }
    catch (NumberFormatException e) {
      System.err.println("Number format error: " + e);
    }
    catch (IllegalTimeSeriesException e) {
      System.err.println(e);
    }

  }

  /** Creates a metadata file for this class with the default values set
   *to the static values defined above. OBS!, check calendar ....
   */
  protected MetaData setDefaultMeta(Date sd, Date ed) throws IOException {

    st.setTime(sd);
    st.clear(st.MILLISECOND);
    et.setTime(ed);
    et.clear(et.MILLISECOND);

    MetaData m = new MetaData(name,description, longDescription,  st, et, ct, step,fieldNames,fieldUnits,fieldTypes);
    return m;
  }

  /**
   * Returns an "empty" metadata object. All fields but Start and end time
   * has valid values (as set in the respective database class).
   * The times will be set to that what is supplied.
   * OBS! This is not stored in the MetaData file !!
   */
  protected MetaData getEmptyMetaData(Date sd, Date ed) {
    MetaData m = new MetaData();
    st.clear(st.MILLISECOND);
    st.setTime(sd);
    et.clear(et.MILLISECOND);
    et.setTime(ed);
    m.setMetaData(name,description, longDescription,  st, et, ct, step, fieldNames,fieldUnits,fieldTypes );
    return m;
  }

}
