package saaps.dbt;

import java.io.*;
import java.util.*;
import java.text.*;

import saaps.util.*;

/**
 * AEDB extends the Database class. It contains
 * the read and write methods that are used in
 * Database. The field is a vector of AE objects.
 *
 *@authors Peter Wintoft.
 *@version 1.0 2001-03-05
 *@see AE
 */
public class AEDB extends Database {

  private double dataGapValueDefinition = 9999.0;

  // Static Fields used to initiate the MetaData
  private static String name = "AEDB";
  private static String description = "Hourly AE, AL, AO, and AU indices.";
  private static String longDescription = "The AE, AL, AO, and AU indices covers the period 1957 to 1988.";
  private static Calendar st=Calendar.getInstance(TimeZone.getTimeZone("UTC"),new Locale("sv","SE"));
  private static Calendar et=Calendar.getInstance(TimeZone.getTimeZone("UTC"),new Locale("sv","SE"));
  private static boolean ct=true;
  private static long step=3600000L; //One hour timestep 
  private static String[] fieldNames = { "AE", "AL", "AO", "AU" };
  private static String[] fieldUnits = { "nT", "nT", "nT", "nT" };
  private static String[] fieldTypes = { "Double", "Double", "Double", "Double" };

  // Fields.
  private AE ae;
  private Vector aeVec;

  // Constructors.
  public AEDB() {
    //
    this.dataGapValue = this.dataGapValueDefinition;
  }
  
  public AEDB(int i) {
    //
    this.dataGapValue = this.dataGapValueDefinition;
  }

  /**
   * When a new class is added to the database this
   * constructor should be used which includes the
   * information that goes into the meta data file.
   *
   * The arguments are: Vector of AE objects, the class name,
   * a short description, continuos time series (false/true), and
   * if a continous time series the time step.
   */
  public AEDB(Vector aeVec) throws Exception {
    this.dataGapValue = this.dataGapValueDefinition;
    this.aeVec = aeVec;
    this.raf = new RandomAccessFile(this.baseDir + this.name + ".data","rw");
    // Then get its metadata
    try { // To load the meta data for this database from file
      this.meta = new MetaData(this.name);
      this.isSet=true;
    }
    catch (IOException e) {
      System.err.println(e);
      System.out.println("Here we should handle the situation when no DB exists");
      try {
        this.meta = setDefaultMeta(((DatabaseObject) (this.aeVec.elementAt(0))).getDate(),
                              ((DatabaseObject) (this.aeVec.elementAt(this.aeVec.size()-1))).getDate()); //Takes static fields from XDB
        this.isSet=true;
      }
      catch (IOException ee) {
        System.err.println(ee);
        ee.printStackTrace();
      }
    }
    catch (ClassNotFoundException e) {
      System.err.println(e);
      e.printStackTrace();
    }

    store((Vector) this.aeVec); //Need to type cast so that store(Object obj) isn't used
  }

  // Methods.
  public Object read() throws IOException {
    long time = this.raf.readLong();
    double[] data = new double[AE.numberOfIndexes];
    for (int i=0;i<data.length;i++) {
      data[i] = this.raf.readDouble();
    }
    this.ae = new AE(time,data);
    return (Object) this.ae;
  }

  public void myHello() {
    System.out.println("My " + this.name + " Hello");
  }

  public void write(Object obj) throws IOException {
    this.ae = (AE) obj;
    this.raf.writeLong(this.ae.getDate().getTime());
    double[] data = this.ae.getDataArray();
    for (int i=0;i<data.length;i++) {
      this.raf.writeDouble(data[i]);
    }
  }
  // Implemented for Database abstract getDate
  public Date getDate(Object obj) {
    return ((AE) obj).getDate();
  }

  /** This method is implemented for Database abstract metod.
   * It retrieves the latest data from the dataprovider and returns the vector of Database objects.
   */
  public Vector getLatestDatabaseObjects() {
    Vector vec = new Vector(8*30);
    Date mdate = new Date();

    double[] tmp = {1.0, 2.0, 3.0, 4.0};
    AE ae = new AE(mdate,tmp);

    for(int i=0;i<8*30;i++) {
      vec.addElement(ae);
      mdate.setDate( (int) (mdate.getDate() + 10));
      ae.setDate(mdate);
//      ae.setAE(i);
    }
    vec.trimToSize();
    return vec;
  }

  /** 
   * This method is used for importing ASCII data to our database.
   * The format of the data is as output from ExportAscii application.
   * It will fill a buffer of databaseobjects and then try to store it,
   * read again and store until it is finished or an exception occures.
   */
  protected void importAscii1(String inputfile) throws Exception {
    final int MAXSIZE=1048576; // bytes , 1Mb 
    //final int MAXSIZE=1024; // bytes ,

    Date inputDateTime;
    String tmpVal;
    double tmpData;


    // Create a Vector of Dst objects.
    Vector vec = new Vector(8*30);

    try {
      //Name of database objects
      String objectName = this.meta.getName().substring(0, this.meta.getName().length()-2);
      short numFields = (short) this.meta.getFieldNames().length;
      double[] data = new double[numFields];

      //set default maximum number of objects
      int maxObjs = 100;
      if(this.meta.getCont()) { //Calculate if we have continous data
	long objSize = DatabaseObject.getInstance(objectName,new Date()).getObjectSize();
	maxObjs = (int) (MAXSIZE/objSize);
      }
      int cnt=0;

      // Input stream.
      BufferedReader streamIn = new BufferedReader(new FileReader(inputfile));
      String textLine = "";
      textLine = streamIn.readLine();
      // Go through the file for all lines starting with
      // "#", ":", or " ".
      while (!textLine.startsWith("1") && !textLine.startsWith("2")) {
	textLine = streamIn.readLine();
      }

      while(textLine != null) {
	while ((textLine != null) && ( cnt < maxObjs) ) {
	  cnt++;
	  // Read the date using - as the field separator.
	  String dateStr = textLine.substring(0,4) + "." +
	    textLine.substring(5,7) + "." +
	    textLine.substring(8,10) + " " + textLine.substring(11,13) + ":" + textLine.substring(14,16) + ":" + textLine.substring(17); //Last item to end of string

	  SimpleDateFormat formatter = (SimpleDateFormat) SimpleDateFormat.getDateTimeInstance(0,0,new Locale("sv","SE"));

	  formatter.setTimeZone(TimeZone.getTimeZone("UTC"));
	  formatter.applyPattern("yyyy.MM.dd HH:mm:ss");
     
	  // Parse the string back into a Date.
	  ParsePosition pos = new ParsePosition(0);
	  inputDateTime = formatter.parse(dateStr, pos);
	
	// Assume that observations are at 00, 03, 06, ...
	  Calendar cal = Calendar.getInstance();
	  cal.setTimeZone(formatter.getTimeZone());
	  cal.setTime(inputDateTime);
	  //The data value
	  tmpVal = textLine.substring(20).trim();
	  StringTokenizer tokens = new StringTokenizer(tmpVal);
	  int numTok = tokens.countTokens(); 
	  for(int i=0;i<numTok;i++) {
	    String value = tokens.nextToken();
	    System.out.println(value);

	    if(value.equals(new String("NaN"))) {
	      data[i] = Double.NaN;
	    }
	    else {
	      data[i] = Double.valueOf(value).doubleValue();
	    }
	    if (data[i] == dataGapValue ) {
	      data[i] = Double.NaN;
	    }
	  }

	  //System.out.println("Date/tome:" + formatter.format(inputDateTime) + " " + tmpKp);
	  vec.addElement(DatabaseObject.getInstance(objectName,cal.getTime(),data));
	  //Get the next line
	  textLine = streamIn.readLine();	
	}
	//Store our buffer of objects
	vec.trimToSize();
	// Store the data in the database
	ObjectTimeSeries ts = new ObjectTimeSeries(vec);

	if(!isSet())
	  System.out.println("MetaData is unavailable \nWill try to create it.");
	System.out.println("Storing:" + cnt);
	store(ts);

	//Reset the vector and counter
	//vec.removeAllElements();
	vec = new Vector(8*30);
	cnt=0;
      }
      streamIn.close();
    }
    catch (IOException e) {
      System.err.println("IO error: " + e);
    }
    catch (NumberFormatException e) {
      System.err.println("Number format error: " + e);
    }
    catch (IllegalTimeSeriesException e) {
      System.err.println(e);
    }

  }

  /** Creates a metadata file for this class with the default values set
   *to the static values defined above. OBS!, check calendar ....
   */
  protected MetaData setDefaultMeta(Date sd, Date ed) throws IOException {

    this.st.setTime(sd);
    this.st.clear(st.MILLISECOND);
    this.et.setTime(ed);
    this.et.clear(et.MILLISECOND);

    MetaData m = new MetaData(name,description, longDescription,  st, et, ct, step,fieldNames,fieldUnits,fieldTypes);
    return m;
  }

  /**
   * Returns an "empty" metadata object. All fields but Start and end time
   * has valid values (as set in the respective database class).
   * The times will be set to that what is supplied.
   * OBS! This is not stored in the MetaData file !!
   */
  protected MetaData getEmptyMetaData(Date sd, Date ed) {
    MetaData m = new MetaData();
    this.st.clear(st.MILLISECOND);
    this.st.setTime(sd);
    this.et.clear(et.MILLISECOND);
    this.et.setTime(ed);
    m.setMetaData(name,description, longDescription,  st, et, ct, step, fieldNames,fieldUnits,fieldTypes );
    return m;
  }

}
