/** 
 *
 * Title: CheckImportData.java
 * Author: Leif Kalla
 * Description:	Import data from a file and store it in our database.
 * Currently supports file syntax as exported by ExportAscii
 */

package saaps.dbt;

import java.io.*;
import java.util.*;
import java.text.*;

import saaps.util.*;

public class CheckImportData  {

  static public void main(String[] args) {
    CheckImportData checker = new CheckImportData();

    Database dbRef;

    if(args.length!=3) {
      System.out.println("Usage: java saaps.dbt.CheckImportData <Db class name> <dataFile> <corrected file>");
      System.exit(0);
    }

    dbRef = Database.getInstance(args[0]);
    String dataFile = args[1];
    String outputfile = args[2];
    checker.importAscii(dbRef,dataFile,outputfile);

    System.out.println("Finished " + args[0]);
  }


  /** 
   * This method is used for importing ASCII data to our database.
   * The format of the data is as output from ExportAscii application.
   * It will fill a buffer of databaseobjects and then try to store it,
   * read again and store until it is finished or an exception occures.
   */
  protected void importAscii(Database dbRef,String inputfile,String outputfile) {
    final int MAXSIZE=524288000;//500 Mb //1048576; // bytes , 1Mb 
    //final int MAXSIZE=1024; // bytes ,
    final long MAXDIFFTIME = (long) 8035200*1000; // 3 months
    Date inputDateTime;
    String tmpVal;
    double tmpKp;
    int j;

    // Create a Vector of Kp objects.
    Vector vec = new Vector(8*30);

    try {
      //Name of database objects
      String objectName = dbRef.meta.getName().substring(0, dbRef.meta.getName().length()-2);
      short numFields = (short) dbRef.meta.getFieldNames().length;
      double[] data = new double[numFields];

      //set default maximum number of objects
      int maxObjs = 100;
      if(dbRef.meta.getCont()) { //Calculate if we have continous data
	long objSize = DatabaseObject.getInstance(objectName,new Date()).getObjectSize();
	maxObjs = (int) (MAXSIZE/objSize);
      }
      int cnt=0;

      // Input stream.
      BufferedReader streamIn = new BufferedReader(new FileReader(inputfile));
      String textLine = "";

      //Output stream
      PrintWriter out = new PrintWriter(new BufferedWriter(new FileWriter(outputfile)));

      SimpleDateFormat formatter = (SimpleDateFormat) SimpleDateFormat.getDateTimeInstance(0,0,new Locale("sv","SE"));

      formatter.setTimeZone(TimeZone.getTimeZone("UTC"));
      formatter.applyPattern("yyyy.MM.dd HH:mm:ss");
      ParsePosition pos = new ParsePosition(0);

      Calendar cal = Calendar.getInstance();
      cal.setTimeZone(formatter.getTimeZone());
      Calendar tmpCal=Calendar.getInstance(TimeZone.getTimeZone("GMT"),new Locale("sv","SE"));

      textLine = streamIn.readLine();
      // Go through the file for all lines starting with
      // "#", ":", or " ".
      while (!textLine.startsWith("1") && !textLine.startsWith("2")) {
	textLine = streamIn.readLine();
      }

      while(textLine != null) {
	while ((textLine != null) && ( cnt < maxObjs) ) {
	  cnt++;
	  // Read the date using - as the field separator.
	  String dateStr = textLine.substring(0,4) + "." +
	    textLine.substring(5,7) + "." +
	    textLine.substring(8,10) + " " + textLine.substring(11,13) + ":" + textLine.substring(14,16) + ":" + textLine.substring(17); //Last item to end of string
     
	  // Parse the string back into a Date.
	  pos.setIndex(0);
	  inputDateTime = formatter.parse(dateStr, pos);
	
	  cal.setTime(inputDateTime);

	  //The data value
	  tmpVal = textLine.substring(20).trim();
	  StringTokenizer tokens = new StringTokenizer(tmpVal);
	  int numTok = tokens.countTokens(); 
	  for(int i=0;i<numTok;i++) {
	    String value = tokens.nextToken();
	    //System.out.println(value);

	    if(value.equals(new String("NaN"))) {
	      data[i] = Double.NaN;
	    }
	    else {
	      data[i] = Double.valueOf(value).doubleValue();
	    }
	  }

	  //System.out.println("Date/tome:" + formatter.format(inputDateTime) + " " + tmpKp);
	  vec.addElement(DatabaseObject.getInstance(objectName,cal.getTime(),data));
	  //System.out.println(((DatabaseObject)vec.elementAt(cnt-1)).toString());
	  if(cnt>1) {
	    long timeDiff = ((DatabaseObject) vec.elementAt(cnt-1)).getDate().getTime()- ((DatabaseObject) vec.elementAt(cnt-2)).getDate().getTime(); //ms
	    if( timeDiff != dbRef.meta.getTimeStep()) {
	      if( timeDiff < 0 ) { // Time jumps backward, we don't correct that type of errors
		System.out.println("Time jumps backward, 1:" + formatter.format(((DatabaseObject) vec.elementAt(cnt-1)).getDate()));
		System.out.println("Time jumps backward, 2:" + formatter.format(((DatabaseObject) vec.elementAt(cnt-2)).getDate()));
		//remove this object
		vec.remove(cnt-1);
		cnt--;
	      }
	      else if( timeDiff < dbRef.meta.getTimeStep()) { //In the middle
		System.out.println("Time is in middle, 1:" + formatter.format(((DatabaseObject) vec.elementAt(cnt-1)).getDate()));
		System.out.println("Time is in middle, 2:" + formatter.format(((DatabaseObject) vec.elementAt(cnt-2)).getDate()));
		//remove this object
		vec.remove(cnt-1);
		cnt--;
	      }
	      else { //Jumps forward
		  System.out.println("Time jumps forward, 1:" + formatter.format(((DatabaseObject) vec.elementAt(cnt-1)).getDate()));
		System.out.println("Time jumps forward, 2:" + formatter.format(((DatabaseObject) vec.elementAt(cnt-2)).getDate()));
		if( timeDiff > MAXDIFFTIME )  { //3 months
		    System.err.println("Time diff > 3 months (diff:" + timeDiff + " limit:" + MAXDIFFTIME + ") exit");
		    System.exit(-1);
		}
		//Fill empty data into file

		int addThisMany = (int) (timeDiff/dbRef.meta.getTimeStep());
		System.out.println("will add data from t1:" + formatter.format(((DatabaseObject) vec.elementAt(cnt-2)).getDate()) + " to t2:" + formatter.format(((DatabaseObject) vec.elementAt(cnt-1)).getDate()) + " num items:" + addThisMany);
		if(addThisMany > 43200) {
		  System.out.println("Too many DatabaseObjects needed:" + addThisMany);
		  System.exit(-1);
		}
		tmpCal.setTime(((DatabaseObject) vec.elementAt(cnt-2)).getDate()); //Increase this with the timeStep in MetaData
		
		//OK start filling the array with the correct objects
		//The empty ones
		for(j=0;j<addThisMany;j++) {
		  System.out.println("tmpCal:" + tmpCal.getTime());
		  //Insert an object at this point
		  vec.add(cnt-1,DatabaseObject.getInstance(objectName,tmpCal.getTime()));
		  out.println(vec.elementAt(cnt-1));		  
		  System.out.println("Added: " + ((DatabaseObject) vec.elementAt(cnt-1)).toString() + " timeStep:" + dbRef.meta.getTimeStep());

		  cnt++; //Increase the vec counter
		  tmpCal.add(Calendar.MILLISECOND,(int) dbRef.meta.getTimeStep()); //millisec
		}

	      }
	    }
	    else 
	      out.println(vec.elementAt(cnt-2));
	  }

	  //Get the next line
	  textLine = streamIn.readLine();
	}
	//Store our buffer of objects
	vec.trimToSize();
	out.println(vec.elementAt(vec.size()-1)); //The last element
	// Store the data in the database

	//Reset the vector and counter
	vec.removeAllElements();
	vec = new Vector(8*30);
	cnt=0;
      }
      out.close();
      streamIn.close();
    }
    catch (IOException e) {
      System.err.println("IO error: " + e);
    }
//     catch (NumberFormatException e) {
//       System.err.println("Number format error: " + e);
//     }
//     catch (IllegalTimeSeriesException e) {
//       System.err.println(e);
//     }

  }

}

