/*
 * @(#)//$Id: Database.java,v 1.16 2002/01/18 12:34:03 saaps Exp $
 *
 * Copyright (c) 2000 Swedish Institute of Space Physics.
 * Box 812, 981 28 Kiruna, Sweden
 *
 * This software is ... what ?
 *
 *
 *	LOG HISTORY of this file:
 *      $Log: Database.java,v $
 *      Revision 1.16  2002/01/18 12:34:03  saaps
 *      Added ExportLatest.
 *
 *      Revision 1.15  2001/10/15 07:16:20  saaps
 *      Fix short description setting for TimeSeries
 *
 *      Revision 1.14  2001/09/20 11:11:23  saaps
 *      Bug fixes:
 *      - faulty insertion of data that was exactly one timestep from the last item in the db
 *      - incorrect handling of data assignment in LANL_1990_095 (modified to deep copy)
 *      - integer overflow in calculation of number of data items to insert (ObjectTimeSeries.fitToDatabase())
 *      - faulty handling of TimeSeries...Exceptions in Database store method. (When real problem occurs
 *        we now throw a DatabaseException in the end). Modified several classes to handle this.
 *
 *      Revision 1.13  2001/09/10 08:13:19  saaps
 *      Added finalize method to Database.
 *
 *      Revision 1.12  2001/06/21 13:53:06  saapscvs
 *      removed error msg
 *
 *      Revision 1.11  2001/06/18 13:24:55  saapscvs
 *      fixed retrieveHere "before" bug
 *
 *      Revision 1.10  2001/06/18 12:08:36  saapscvs
 *      Added methods to retrieve
 *      latest X data
 *
 *      Revision 1.9  2001/06/18 06:43:42  saaps
 *      Added code so that the baseDir string always ends with an "/".
 *
 *      Revision 1.8  2001/04/12 09:42:07  saapscvs
 *      Added methods for data retrieval.
 *      Removed old getTimeSeries methods.
 *
 *      Revision 1.7  2000/10/27 12:06:39  saapscvs
 *      Modified retrieveSEA to use superposed epoch analysis
 *
 *      Revision 1.6  2000/10/18 11:23:21  saapscvs
 *      fieldNames and fieldUnits are now set from MetaData
 *
 *      Revision 1.5  2000/10/04 13:56:54  saapscvs
 *      Removed fieldIndex bug and updated the RMI with the new retrieveSEA.
 *
 *      Revision 1.4  2000/10/03 14:14:32  saapscvs
 *      Redesigned retrieveSEA to get additional information
 *      about the average process.
 *
 *      Revision 1.3  2000/10/02 13:22:42  saapscvs
 *      Added methods to support averaging and "SEA" (superposed
 *      epoch analysis)
 *
 *      Revision 1.2  2000/09/28 10:42:56  saapscvs
 *      Documented Database.java DatabaseRMI.java DatabaseRMIImpl.java
 *      MetaData.java
 *
 *      Revision 1.1.1.1  2000/09/28 08:43:52  saaps
 *      Imported sources
 *
 *
 */

package saaps.dbt;

import java.io.*;
import java.text.*;
import java.util.*;
import java.lang.reflect.*;
import saaps.util.*;

/**
 * Database is the abstract base class for all saaps data classes.
 * It provides storing and retrieving of the data defined by the
 * classes that extend Database.
 * <p>
 * Information about the data is stored in the class <code>MetaData</code>
 * which is initialized from the *.meta file when an instance of 
 * <code>Database</code> is created.
 * <p>
 * The static method <code>getInstance(String)</code> can be used to get
 * an instance of <code>Database</code>, it will open the data file
 * in read-write mode.
 * The other static method <code>getInstance(String, boolean)</code> should
 * be used when you want to open the data file in read only mode.
 * <p>
 * @author    Leif Kalla (leif@irf.se)
 * @version   $Revision: 1.16 $
 */
public abstract class Database {

  /**
   * Absolute path to the database.
   * The <code>dataDirVar</code> contains the name of the property
   * that will reset the path if it's defined.
   */
  protected static String baseDir = "/space/saaps/Database/";

  /** An environment property name that defines the path to database. */
  public static String dataDirVar="saaps.data";

  public MetaData meta;     /* We need it public so that we can use it's
			       methods */

  protected double dataGapValue = 0.0;     //Is set by the extending class

  /**
   * Indicates if this instance of <code>Database</code> has valid
   * <code>MetaData</code>.
   */
  protected boolean isSet=false; // Have we initiated all params like meta

  /**
   * Reading and writeing of data is done through a RandomAccessFile
   * class.
   */
  protected RandomAccessFile raf;

  /* No Constructors defined in this abstract class !
   * You will have to use the static methods getInstance(...)
   * to get an object.
   */
  
  /**
   * Close the DB when all references to this object are lost.
   * OBS! To really ensure that this method is run you must do the following
   * <code>System.gc();
   * System.runFinalization();</code>
   */
  protected void finalize() throws Throwable {
      super.finalize();
      raf.close();
  }

  /** 
   * Runs method getInstance with the default action to
   * open database for read/write.
   * @param myDB the database to be loaded
   * @return a <code>Database</code> object
   */
  public static Database getInstance(String myDB) {
    return(getInstance(myDB,false));
  }

  /**
   * Gets a <code>Database</code> object.
   * Also retrieves the <code>MetaData</code> information from file and
   * opens the randomaccessfile (data file).
   * <p>
   * If loading of <code>MetaData</code> fails, the default <code>MetaData</code>
   * is set using {@link #getEmptyMetaData(Date, Date)}.
   * In this case the {@link #isSet} variable is set to false.
   * <p>
   * @param myDB the database to be loaded
   * @param readOnlyAccess the access mode
   * @return a <code>Database</code> object
   */
  public static Database getInstance(String myDB, boolean readOnlyAccess) {
    Database theDB=null;

    //Get the database location
    baseDir=System.getProperty(dataDirVar,baseDir);

    //Make sure that it ends with a "/".
    if(!baseDir.endsWith("/")) {
        baseDir += "/";
    }

    try {
      Class classObject = Class.forName("saaps.dbt." + myDB);
      theDB = (Database) classObject.newInstance();
      //OBS! take care to close it, check for exception
      if(readOnlyAccess)
	theDB.raf = new RandomAccessFile(baseDir + myDB + ".data","r");
      else
	theDB.raf = new RandomAccessFile(baseDir + myDB + ".data","rw");
    }
    catch (Exception e) {
      e.printStackTrace();
    }
    /* We successfully got the object.
       Then get its metadata */
    try { // to load the meta data for this database from file
      theDB.meta = new MetaData(myDB);
      theDB.isSet=true;
    }
    catch (IOException e) {
      System.err.println(e);
      System.out.println("Here we should handle the situation when no DB exists");
      //Get the other meta data so that we know if it's continous , timestep etc.
      theDB.meta = theDB.getEmptyMetaData(new Date(), new Date());
      theDB.isSet=false;    //Indicate that meta data isn't really valid
    }
    catch (ClassNotFoundException e) {
      System.err.println(e);
      e.printStackTrace();
    }
    return theDB;
  }

  /*
    XXX NOT USED any more, Instead ... 
    ObjectTimeSeries -> TimeSeries -> store(ts) !! 

    Note: This serves a good example on how to dynamically instantiate
    an object with a constructor that takes an argument.
    
  */
  /*
    public static Database getInstance(String myDB, Vector dbo) {
    Database theDB=null;

    //Get the database location
    baseDir=System.getProperty(dataDirVar,baseDir);
    try {
      Class classObject = Class.forName("saaps.dbt." + myDB);
      //Try to get a constructor with argument Vector for this database
      //Note that Constructor is from java.lang.reflect
      Class[] theArgs = { Class.forName("java.util.Vector") };
      Constructor dbConstructor = classObject.getConstructor(theArgs);

      Object[] objArgs = { dbo };
      theDB = (Database) dbConstructor.newInstance( objArgs );
      //OBS! take care to close it, check for exception
      theDB.raf = new RandomAccessFile(baseDir + myDB + ".data","rw");
    }
    catch (Exception e) {
      e.printStackTrace();
    }
    return theDB;
  } 
  */

  /** 
   * Close the database.
   * <p>
   * <b>Note:</b> we don't have any open method yet so you will
   * have to get a new instance of this database to get an open file.
   */
  public void close() throws IOException {
    raf.close();
  }

  /** 
   * Indicates if the <code>MetaData</code> is set and the data file
   * is open. If <code>MetaData</code> is set by
   * {@link #getEmptyMetaData(Date , Date )} the parameter <code>isSet</code>
   * is false.
   * @return <code>true</code> if the <code>MetaData</code> is set
   */
  public boolean isSet() {
    return isSet;
  }


  /**
   * Positions in start or end of file.
   * @param start if <code>true</code> position in beginning of file
   * @return the position
   */
  protected final long position(boolean start)  throws IOException {
    long pos;
    pos = (start ? 0 : raf.length());
    raf.seek(pos);
    return (pos);
  }

  /**
   * Runs position with default action to position before the
   * search date. See also {@link #position(Date , boolean , boolean )}.
   * @param d the search date
   * @param fromStart if true then serach file from beginning else proceed 
   * from current file position.
   * @return the position for the data item that follows this one
   */
  private final long position(Date d, boolean fromStart)  throws IOException {
    return(position(d,fromStart,true));
  }

  /**
   * Positions the <code>RandomAccessFile</code> at the location closest
   * to the supplied date and returns the location of the object after this.
   * Position will try to use <code>MetaData</code> to do a faster search.
   * <p>
   * If the date is after last date in file the fileposition is left 
   * at EOF and an IOException is thrown.
   * @param d the date to be found.
   * @param fromStart if true then serach file from beginning, not
   * using <code>MetaData</code>, else proceed from current file position.
   * @param before if true position before date else at or after date
   * @return the position for the data item that follows this one
   */
  private final long position(Date d, boolean fromStart, boolean before)  throws IOException {
    long currentFp=0;
    long previousFp=0;
    long trailFp=0;
    long debugFp=0;

    currentFp= raf.getFilePointer();
    previousFp=currentFp;
    trailFp=currentFp;
    debugFp=currentFp;

    if (fromStart) { //don't use meta data for search
      raf.seek(0);
      
      currentFp = 0;
      previousFp= 0;
      trailFp=0;

      try {
	Object obj = read();
	currentFp=raf.getFilePointer();
	while ((getDate(obj)).before(d)) {
	  /* if there are no more objects to read the 
	   * IOException is thrown. */
	  obj = read();
	  trailFp=debugFp;
	  debugFp=previousFp;
	  previousFp=currentFp;
	  currentFp=raf.getFilePointer();
	}
	
      }
      catch (EOFException e) {
	/* the filepointer is left at EOF and no IOException is thrown !
	 * XXX check for a better implementation ...
	 * earlier an exception was thrown but to enable creation of
	 * datafile from scratch (without metadata) the exception was
	 * removed
	 */
      }
    }
    else {     //If not from start
      /* Try to use meta data if available (should always be true) */
      if(isSet()) {
	Date ed = meta.getEndDate();
	Date sd = meta.getStartDate();

	if(d.after(ed)) { //This date is after the end date in our file
	  raf.seek(raf.length());
	  throw( new EOFException("Date is after last item in Database"));
	}
	else if(d.before(sd)) { //This date is before first date in file
	  currentFp = 0;
	  previousFp= 0;
	  trailFp=0;
	}
	else if(meta.getCont()) {
	  /* If continous data we'll try to calculate position in "raf"
	   * this will improve the speed
	   */
	  // diff in milliseconds 
	  long tmp = (long) ((long) d.getTime() - (long) sd.getTime());
	  long nObjs = (long) tmp/meta.getTimeStep();
	  long remainder = tmp%meta.getTimeStep();
	  
 	  //System.out.println("nObjs: " + nObjs + "Remainder: " + remainder);
	  if(remainder>0) nObjs+=1;
	  String[] tmpStr = meta.getFieldNames();
	  //Number of objects * number of fields * filedSize
	  currentFp = (nObjs+1)*(tmpStr.length+1)*8;
	  previousFp= currentFp - (tmpStr.length+1)*8;
	  trailFp= previousFp - (tmpStr.length+1)*8;
	}
	else { //If none above do the next best thing
	  /* XXX Here we will do "quick search" but for now 
	   * we do the slow thing */
	  raf.seek(0);
	  currentFp = 0;
	  previousFp= 0;
	  trailFp=0;

	  try {
	    Object obj = read();
	    currentFp=raf.getFilePointer();
	    while ((getDate(obj)).before(d)) {
	      /* if there are no more objects to read the 
	       * IOException is thrown. */
	      obj = read();
	      trailFp=previousFp;
	      previousFp=currentFp;
	      currentFp=raf.getFilePointer();
	    }
	  }
	  catch (EOFException e) {
	    /* the filepointer is left at EOF and no IOException is thrown !
	     * XXX check for a better implementation ...
	     * earlier an exception was thrown but to enable creation of
	     * datafile from scratch (without metadata) the exception was
	     * removed
	     */
	  }

	}
      }
      else { //If meta is not set
	/* Do a slow search */
	raf.seek(0);
	currentFp = 0;
	previousFp= 0;
	trailFp=0;

	try {
	  Object obj = read();
	  currentFp=raf.getFilePointer();
	  while ((getDate(obj)).before(d)) {
	    /* if there are no more objects to read the IOException is thrown. */
	    obj = read();
	    trailFp=previousFp;
	    previousFp=currentFp;
	    currentFp=raf.getFilePointer();
	  }
	  
	}
	catch (EOFException e) {
	  /* the filepointer is left at EOF and no IOException is thrown !
	   * XXX check for a better implementation ...
	   * earlier an exception was thrown but to enable creation of
	   * datafile from scratch (without metadata) the exception was
	   * removed
	   */
	}
      }
    }

    /* do the real seek */
    if(before) {
      raf.seek(trailFp);
    }
    else {
      raf.seek(previousFp);
    }
    return currentFp; //Return a "pointer" to the object after this one
  }

  /**
   * Gets the first database object in the database.
   * @return the first <code>DatabaseObject</code> from the database
   */
  public DatabaseObject getFirst() throws IOException {
    long savePointer;

    savePointer = raf.getFilePointer();
    raf.seek(0); // Go to beginning of file
    DatabaseObject tmp = (DatabaseObject) read();
    raf.seek(savePointer);
    return tmp;
  }

  /**
   * Gets the last database object in the database
   * @return the last <code>DatabaseObject</code> from the database
   */
  public DatabaseObject getLast() throws IOException {
    long savePointer;

    savePointer = raf.getFilePointer();
    raf.seek(0); // Go to beginning of file

    DatabaseObject tmp = (DatabaseObject) read();
    long dataSize = tmp.getDataSize(); //Size of data from one object
    raf.seek(raf.length()-dataSize); //Position before last item
    tmp = (DatabaseObject) read();
    raf.seek(savePointer);
    return tmp;
  }


  /**
   * Gets all objects in the database (returned in a <code>Vector</code>).
   * <b>Note:</b> We should take care when using this so that we don't 
   * exhaust memory.
   * You should use one of the other retrieve methods when you want
   * to load data.
   * @return a <code>Vector</code> containing <b>all</b> 
   * <code>DatabaseObject</code>s. 
   */
  public final Vector retrieve() throws IOException {
    long savePointer;

    Vector vec = new Vector(50); // Start with 50 elements
    savePointer = raf.getFilePointer();

    raf.seek(0);
    try {
      while(true) {
	vec.addElement((Object) read());
      }
    }
    catch (IOException e) {
      raf.seek(savePointer); // Restore position
      if(vec.size()==0) { //We didn't get any data
	throw ((IOException) e);
      }
    }
    raf.seek(savePointer); // Restore position
    vec.trimToSize();
    return vec;
  }

  /**
   * Gets the object before this date.
   * @param sd the search date
   * @return object before the search date.
   */
  public final Object retrieve(Date sd) throws IOException {
    Vector vec = retrieve(sd,sd);
    return ( vec.elementAt(0));
  }

  /**
   * Gets the objects between the two dates. If the first date is
   * after the second date, only the object at the first date will
   * be returned.
   * @param sd start date for search
   * @param ed end date for search
   * @return a <code>Vector</code> containing the
   * <code>DatabaseObject</code>s between the dates.
   */
  public final Vector retrieve(Date sd, Date ed) throws IOException {
    //We should throw an exception if no data is found
    //perhaps something like DataNotFoundException
    return (retrieve(sd,ed,true));
  }

  /**
   * Gets the <code>DatabaseObject</code>s between the dates.
   * If the first date is after the second date, only the object at 
   * the first date will be returned.
   * @param sd start date for search
   * @param ed end date for search
   * @param before if <code>true</code> then the first object returned
   * is before the start search date.
   * @return a <code>Vector</code> containing the
   * <code>DatabaseObject</code>s between the dates.
   */
  public final Vector retrieve(Date sd, Date ed, boolean before) throws IOException {
    /* XXX We should throw an exception if no data is found
     * perhaps something like DataNotFoundException
     */
    Vector res = null;

    long savePointer = raf.getFilePointer();

    if (sd.after(ed))
      ed = sd;

    try {
      /* sd , fromstart , before */
      position(sd,false,before); //Try using metadata for retrieve
    }
    catch (IOException e) {
      raf.seek(savePointer); //Restore raf position
      System.err.println(e + " " + sd + ", No such start date.");
      throw(new IOException(e + " ,DataNotFoundException"));
    }

    /* XXX Use Metadata here to setup vector and ... ???? */
    res = new Vector(100);
    Object obj = null;
    try {
      obj = read();
      while ((getDate(obj)).before(ed) || (getDate(obj)).equals(ed)) {
	res.addElement(obj);
	obj = read();
      }
    }
    catch (IOException e) {  // End date does not exist.
      //System.err.println(e + " " + sd + ", No such end date.");
      raf.seek(savePointer); //Restore raf position
      if(res.size()==0) { // We didn't get any data !!
	/* I suppose this check is already implemented above if sd not found
	 * But it doesn't hurt
	 */
	throw(new IOException(e + " ,DataNotFoundException"));
      }
      else { // We got something ... return at least this
	res.trimToSize();
	return res;
      }
    }

    if(res.size()==0) { //We didn't find any data throw our exception
      throw(new IOException("DataNotFoundException"));
    }
    raf.seek(savePointer); //Restore raf position
    res.trimToSize();
    return res;
  }

  /**
   * Gets an <code>ObjectTimeSeries</code> containing data between
   * the dates.
   * See also {@link #retrieve(Date, Date, boolean)}
   * @param sd start date for search
   * @param ed end date for search
   * @param before if <code>true</code> then the first object returned
   * is before the start search date.
   * @return the ObjectTimeSeries
   */
  public final ObjectTimeSeries retrieveOTS(Date sd, Date ed, boolean before)
    throws IOException, IllegalTimeSeriesException {
    return (new ObjectTimeSeries(retrieve(sd,ed,before)));
  }

  /**
   * Gets a <code>TimeSeries</code> containing data between
   * the dates.
   * See also {@link #retrieve(Date, Date, boolean)} and 
   * {@link #retrieveOTS(Date, Date, boolean)}.
   * @param sd start date for search
   * @param ed end date for search
   * @param before if <code>true</code> then the first object returned
   * is before the start search date.
   * @return the TimeSeries
   */
  public final TimeSeries retrieveTS(Date sd, Date ed, boolean before)
    throws IOException {
        TimeSeries ts = new TimeSeries(retrieve(sd,ed,before),meta.getFieldNames(),meta.getFieldUnits());
        ts.setShortDescription(meta.getDescr());
        return (ts);
  }

  /**
   * Gets a <code>TimeSeries</code> containing data between
   * the dates, the data is resampled if needed to the sampleinterval stated.
   * See also {@link #retrieve(Date, Date, boolean)} and 
   * {@link #retrieveOTS(Date, Date, boolean)}.
   * @param sd start date for search
   * @param ed end date for search
   * @param before if <code>true</code> then the first object returned
   * is before the start search date.
   * @param sampleInterval timestep that the data should have [ms]
   * @return the TimeSeries
   */
  public final TimeSeries retrieveTS(Date sd, Date ed, boolean before,long sampleInterval)
    throws IOException, TimeSeriesException {
      if(meta.getTimeStep()!=sampleInterval) {
	  TimeSeries ts = new TimeSeries(retrieve(sd,ed,before),meta.getFieldNames(),meta.getFieldUnits());
	  ts.resample(sampleInterval/2,sampleInterval,-sampleInterval/2,sampleInterval/2);
	  return ts;
      }
      else
	  return (new TimeSeries(retrieve(sd,ed,before),meta.getFieldNames(),meta.getFieldUnits()));

  }

  /**
   * Gets a <code>TimeSeries</code> containing data from this time
   * and this many objects back or forward.
   * OBS! This works only on database with equidistant data.
   * @param sd start date for search
   * @param num this many objects, if negative backward else forward
   * @param before if <code>true</code> then the first object returned
   * is before the start search date.
   * @return the TimeSeries
   */
  public final TimeSeries retrieveHere(Date sd, int num, boolean before)
  throws IOException, DatabaseException, MetaNotSetException, TimestepException {
      if(!isSet())
          throw (new MetaNotSetException("no meta data available"));
      if(!meta.getCont())
          throw ( new TimestepException("Not continous data"));
        /* if called with num=0 we return the last anyway */
      if(num==0)
          return retrieveLatest();
      int wantThisMany=num;
      if(!before)
          num = num>=0?(num-1):(num+1);
      else
          num = num>=0?(num-2):(num+2);
          //set end date! Note: if neg. ed is before sd see the return statement
          Date ed = new Date(sd.getTime()+num*meta.getTimeStep());
          TimeSeries ts = sd.before(ed)?new TimeSeries(retrieve(sd,ed,before),meta.getFieldNames(),meta.getFieldUnits()):new TimeSeries(retrieve(ed,sd,before),meta.getFieldNames(),meta.getFieldUnits());
          if(ts.getLength() != java.lang.Math.abs(wantThisMany))
              throw new DatabaseException("Cannot get "+java.lang.Math.abs(wantThisMany)+" items! Got "+ts.getLength());
          return (ts);
  }

/**
 * Gets a <code>TimeSeries</code> containing data from the end of the db
 * and this many objects back.
 * Note: If num is negative it will try to fetch data into the future and
 * it will most certainly fail - no problem the user will get a message.
 * OBS! This works only on database with equidistant data.
 * @param num this many objects, if negative backward else forward
 * @return the TimeSeries
 */
  public final TimeSeries retrieveLatest(int num)
  throws IOException, DatabaseException, MetaNotSetException, TimestepException {
      Date sd = getLast().getDate();
      return retrieveHere(sd,-1*num,false);
  }

/**
 * Gets a <code>ObjectTimeSeries</code> containing data from the end of the db
 * and milliseconds back.
 * OBS! This works only on database with equidistant data.
 * Note: This call will fail if you try to fetch more than 2E16 objects.
 * @param ms this many milliseconds of data
 * @param millis an extra field to distinguish this method
 * @return the TimeSeries
 */
  public final ObjectTimeSeries retrieveLatest(long ms,boolean millis)
  throws IOException, DatabaseException, MetaNotSetException, TimestepException,IllegalTimeSeriesException {
      Date sd = getLast().getDate();
      /* calculate the number of objects */
      long num = ((long)ms/meta.getTimeStep());
      
      return retrieveHereOTS(sd,((int)(-1*num)),false);
  }
  
  /**
   * Gets a <code>ObjectTimeSeries</code> containing data from this time
   * and this many objects back or forward.
   * OBS! This works only on database with equidistant data.
   * @param sd start date for search
   * @param num this many objects, if negative backward else forward
   * @param before if <code>true</code> then the first object returned
   * is before the start search date.
   * @return the ObjectTimeSeries
   */
  public final ObjectTimeSeries retrieveHereOTS(Date sd, int num, boolean before)
  throws IOException, DatabaseException, MetaNotSetException, TimestepException,IllegalTimeSeriesException {
      if(!isSet())
          throw (new MetaNotSetException("no meta data available"));
      if(!meta.getCont())
          throw ( new TimestepException("Not continous data"));
        /* if called with num=0 we return the last anyway */
      if(num==0) {
          Vector vec = new Vector();
          vec.addElement(getLast());
          return (new ObjectTimeSeries(vec));
      }
      int wantThisMany=num;
      if(!before)
          num = num>=0?(num-1):(num+1);
      else
          num = num>=0?(num-2):(num+2);
          //set end date! Note: if neg. ed is before sd see the return statement
          Date ed = new Date(sd.getTime()+num*meta.getTimeStep());
          ObjectTimeSeries ts = sd.before(ed)?new ObjectTimeSeries(retrieve(sd,ed,before)):new ObjectTimeSeries(retrieve(ed,sd,before));
          if(ts.getLength() != java.lang.Math.abs(wantThisMany))
              throw new DatabaseException("Cannot get "+java.lang.Math.abs(wantThisMany)+" items! Got "+ts.getLength());
          return (ts);
  }

  
/**
 * Gets a <code>TimeSeries</code> containing data from the last object in the
 * database.
 * @param num this many objects, if negative backward else forward
 * @return the TimeSeries
 */
  public final TimeSeries retrieveLatest() throws IOException {
      Vector vec = new Vector(1);
      vec.add(getLast());
      return new TimeSeries(vec,meta.getFieldNames(),meta.getFieldUnits());
  }

  /**
   * Retrieves a TimeSeries object containing the averaged data for the
   * input event list.
   * <p>
   * Every date/time in the <code>events</code> list is used as time
   * reference for data retrieval, the <code>start</code> and 
   * <code>end</code> parameters define the time window for which 
   * data is retrieved.
   * If data for any of these time windows is missing an IOException 
   * is thrown.
   * The data loaded for every event is averaged, time is set from the 
   * middle of the event data. A <code>TimeSeries</code> object 
   * containing the data field indicated is created from the averaged 
   * data. 
   * @param fieldIndex  index of the data field
   * @param events array of event times
   * @param start event window starts here (normally negative [milliseconds])
   * @param end event window ends here (normally positiv [milliseconds])
   * @return <code>TimeSeries</code> object with averaged 
   * <code>fieldName</code> data for start to end time for every event
   * @exception TimeSeriesIndexException when <code>fieldIndex</code>
   * is out of bounds
   * @exception IOException when data for any of the event times are missing
   */
  public TimeSeries retrieveAVE(int fieldIndex,Date[] events,long start, long end)
    throws IOException, TimeSeriesIndexException, IllegalTimeSeriesException {

    Vector theData = new Vector(50);

    Calendar sd = Calendar.getInstance(TimeZone.getTimeZone("UTC"),new Locale("sv","SE"));
    Calendar ed = Calendar.getInstance(TimeZone.getTimeZone("UTC"),new Locale("sv","SE"));

    /* An IOException is thrown if start date is before start of data
     * in file and if start date is after end of data in file.
     * XXX here we should get an DataNotFoundException instead !! 
     */
    for(int i=0;i<events.length;i++) {
      sd.setTime(new Date(events[i].getTime()+start));
      ed.setTime(new Date(events[i].getTime()+end));

      /* add one DatabaseObject (the averaged object ) */
      theData.add(((ObjectTimeSeries) retrieveOTS(
						  sd.getTime(),
						  ed.getTime(),
						  false)
		   ).average());
    }
    theData.trimToSize();
    /* strip off the other fields and return the TimeSeries */
    TimeSeries ts = new TimeSeries(theData,fieldIndex);
    ts.setShortDescription("SEA field " + fieldIndex + " of " + meta.getDescr());
    return (ts);
  }

  /**
   * Retrieves a TimeSeries object containing the averaged data for the
   * input event list.
   * <p>
   * Every date/time in the <code>events</code> list is used as time
   * reference for data retrieval, the <code>start</code> and 
   * <code>end</code> parameters define the time window for which 
   * data is retrieved.
   * If data for any of these time windows is missing an IOException 
   * is thrown.
   * The data loaded for every event is averaged, time is set from the 
   * middle of the event data. A <code>TimeSeries</code> object 
   * containing the data field indicated, the number of datapoints 
   * used in the average and the standard deviation  is created from 
   * the averaged data.
   * @param fieldIndex  index of the data field
   * @param events array of event times
   * @param start event window starts here (normally negative [milliseconds])
   * @param end event window ends here (normally positiv [milliseconds])
   * @param withAdditionalInfo if <code>false</code> run the other method
   * {@link #retrieveAVE(int, Date[], long, long)}.
   * @return <code>TimeSeries</code> object with averaged 
   * <code>fieldName</code> data for start to end time for every event
   * @exception TimeSeriesIndexException when <code>fieldIndex</code>
   * is out of bounds
   * @exception IOException when data for any of the event times are missing
   */
  public TimeSeries retrieveAVE(int fieldIndex,Date[] events,long start, long end, boolean withAdditionalInfo)
    throws IOException, TimeSeriesIndexException, IllegalTimeSeriesException {

    /* Run the simpler one if num data points and
     * standard deviation not needed.
     */
    if(!withAdditionalInfo)
      return retrieveAVE(fieldIndex,events,start,end);

    Vector theData = new Vector(50);
    Vector theTimes = new Vector(50);

    Calendar sd = Calendar.getInstance(TimeZone.getTimeZone("UTC"),new Locale("sv","SE"));
    Calendar ed = Calendar.getInstance(TimeZone.getTimeZone("UTC"),new Locale("sv","SE"));
    Calendar dpTime = Calendar.getInstance(TimeZone.getTimeZone("UTC"),new Locale("sv","SE"));

    /* An IOException is thrown if start date is before start of data
     * in file and if start date is after end of data in file.
     * XXX here we should get an DataNotFoundException instead !! 
     */
    for(int i=0;i<events.length;i++) {
      sd.setTime(new Date(events[i].getTime()+start));
      ed.setTime(new Date(events[i].getTime()+end));

      /* add one DatabaseObject (the averaged object ) */
      theData.add(((TimeSeries) retrieveTS(
					   sd.getTime(),
					   ed.getTime(),
					   false)
		   ).average(dpTime,true));
      theTimes.add(dpTime.getTime());
    }
    theData.trimToSize();
    theTimes.trimToSize();
    /*                               length      , data nodp stddev */
    double[][] tmpData = new double[theData.size()][3];
    int nodf = java.lang.reflect.Array.getLength(theData.elementAt(0))/3;
    if(fieldIndex > (nodf-1))
      throw (new TimeSeriesIndexException("field index:" + fieldIndex +
					  " out of bounds"));

    /* copy data into the reduced array */
    for(int i=0;i<theData.size();i++) {
      tmpData[i][0] = java.lang.reflect.Array.
	getDouble(theData.elementAt(i),fieldIndex); //data
      tmpData[i][1] = java.lang.reflect.Array.
	getDouble(theData.elementAt(i),fieldIndex+
		  TimeSeries.aveArrDptOffset*nodf); //nodp
      tmpData[i][2] = java.lang.reflect.Array.
	getDouble(theData.elementAt(i),fieldIndex+
		  TimeSeries.aveArrStdvOffset*nodf); //stddev
    }
    Date[] tmpTimes = new Date[theTimes.size()];
    theTimes.copyInto(tmpTimes);

    String[] fieldName = new String[3];
    String[] fieldUnit = new String[3];
    fieldName[0] = new String("Index " + fieldIndex + " of " +
			      meta.getDescr());
    fieldName[1] = new String("no. data points");
    fieldName[2] = new String("standard deviation");
    fieldUnit[0] = new String("double");
    fieldUnit[1] = new String("double");
    fieldUnit[2] = new String("double");

    TimeSeries ts = new TimeSeries(tmpTimes,tmpData,fieldName,fieldUnit);
    ts.setShortDescription("SEA field " + fieldIndex + " of " + meta.getDescr());
    return (ts);
  }

  /**
   * Retrieves a TimeSeries object containing the averaged data for the
   * input event list.
   * <p>
   * Every date/time in the <code>events</code> list is used as time
   * reference for data retrieval, the <code>start</code> and 
   * <code>end</code> parameters define the time window for which 
   * data is retrieved.
   * If data for any of these time windows is missing an IOException 
   * is thrown.
   * The data field indicated is averaged over the events.
   * A <code>TimeSeries</code> object 
   * containing the data field indicated, the number of datapoints 
   * used in the average and the standard deviation  is created from 
   * the averaged data.
   * @param fieldIndex  index of the data field
   * @param events array of event times
   * @param start event window starts here (normally negative [milliseconds])
   * @param end event window ends here (normally positiv [milliseconds])
   * @param withAdditionalInfo Not used yet
   * @return <code>TimeSeries</code> object with averaged 
   * <code>fieldName</code> data for every event.
   * @exception TimeSeriesIndexException when <code>fieldIndex</code>
   * is out of bounds
   * @exception IOException when data for any of the event times are missing
   */
    public TimeSeries retrieveSEA(int fieldIndex,Date[] events,long start, long end,boolean withAdditionalInfo)
	throws IOException, TimeSeriesIndexException, IllegalTimeSeriesException {

	Calendar sd = Calendar.getInstance(TimeZone.getTimeZone("UTC"),new Locale("sv","SE"));
	Calendar ed = Calendar.getInstance(TimeZone.getTimeZone("UTC"),new Locale("sv","SE"));
	Calendar dpTime = Calendar.getInstance(TimeZone.getTimeZone("UTC"),new Locale("sv","SE"));

    /* An IOException is thrown if start date is before start of data
     * in file and if start date is after end of data in file.
     * XXX here we should get an DataNotFoundException instead !! 
     */
    TimeSeries[] tsArr = new TimeSeries[events.length];
    for(int i=0;i<events.length;i++) {
      sd.setTime(new Date(events[i].getTime()+start));
      ed.setTime(new Date(events[i].getTime()+end));
      tsArr[i] = retrieveTS(sd.getTime(),ed.getTime(),false);
    }
    /* assume same length for all TimeSeries */
    double[][] data = new double[((TimeSeries) tsArr[0]).getLength()][3];
    Date[] tmpTimes = new Date[((TimeSeries) tsArr[0]).getLength()];

    /* calculate average, no. data pts, std.dev. */
    /* first loop over the data points */
    for(int j=0;j<((TimeSeries) tsArr[0]).getLength();j++) {
	/* second loop over the events */
	for(int i=0;i<tsArr.length;i++) {
	    double tmp = ((TimeSeries) tsArr[i]).getData(j,fieldIndex);
	    /* Skip NaN:s */
	    if(!Double.isNaN(tmp)) {
		data[j][0]+=tmp;      //the data itself
		data[j][1]++;        // Add one to the counts
		data[j][2] += Math.pow(tmp,2);      //std.dev. container
	    }
 	}
    }
    for(int j=0;j<((TimeSeries) tsArr[0]).getLength();j++) {
	if (data[j][1]<1) { //All datapoints were NaN
	    data[j][0] =  Double.NaN;
	    data[j][2] =  Double.NaN;
	}
	else {
	    data[j][0]/=data[j][1];
	    data[j][2]=Math.sqrt((data[j][2] - (data[j][1]*Math.pow(data[j][0],2)))/(data[j][1]-1));
	}
    }

    /* Create the times */
    long step = (end-start)/(((TimeSeries) tsArr[0]).getLength()-1);
    long stepStart = -1*((end-start)/2);
    for(int i=0;i<((TimeSeries) tsArr[0]).getLength();i++) {
	tmpTimes[i]= new Date(stepStart);
	stepStart+=step;
    }
    
    String[] fieldName = new String[3];
    String[] fieldUnit = new String[3];
    fieldName[0] = new String("Index " + fieldIndex + " of " +
			      meta.getDescr());
    fieldName[1] = new String("no. data points");
    fieldName[2] = new String("standard deviation");
    fieldUnit[0] = new String("AVE");
    fieldUnit[1] = new String("AVE");
    fieldUnit[2] = new String("AVE");

    TimeSeries ts = new TimeSeries(tmpTimes,data,fieldName,fieldUnit);
    ts.setShortDescription("SEA field " + fieldIndex + " of " + meta.getDescr());
    return (ts);
  }

  /**
   * Gets a TimeSeries object.
   * The <code>limit</code> parameter will be the maximum memory 
   * consumption of the resultimg <code>TimeSeries</code> object,
   * in other words the data will be resampled to fit into that memory.
   * <p>
   * <b>Note:</b> This method only handles data with constant timestep.
   * @param sd start date for search
   * @param ed end date for search
   * @param limit limits the size of the TimeSeries object
   * @return <code>TimeSeries</code> contains data between the dates, possibly reduced
   */
  public TimeSeries getTimeSeries(Date sd, Date ed, long limit) throws IOException,MetaNotSetException, TimestepException , DatabaseException {

    /* XXX this code is not tested properly !!! */

    TimeSeries res = null;

    if(!isSet()) 
      throw (new MetaNotSetException("no meta"));

    if(!meta.getCont())
      throw ( new TimestepException("Not continous data"));

    long objSize = getFirst().getObjectSize();
    int maxObjs = (int) (limit/objSize);

    long totalSize =   ((long) ((ed.getTime()-sd.getTime())
				/meta.getTimeStep()))*objSize;

    /* check if we need to reduce the data */
    if(totalSize<=limit) {
      res = retrieveTS(sd, ed,false);
      res.setShortDescription(meta.getDescr());
      res.setLongDescription(meta.getDescrLong());
      return res;
    }

    long sampleInterval= (long) (ed.getTime()-sd.getTime())/maxObjs;

//     long dt0 = sampleInterval/2; 
//     long wa = -sampleInterval/2;
//     long wb = sampleInterval/2;

    /* setup data buffer */
    Vector oBuffer = new Vector(maxObjs);
    /* Setup timeing for data retrieval */
    Date rdStart = new Date(sd.getTime());
    Date rdEnd = new Date(sd.getTime()+sampleInterval);

    for(int i=0;i<maxObjs;i++) { //Go through the intervals
      /* DEBUG System.out.println("retrieved " + vec.size() +
	 " objects" + " first date:" + 
	 ((DatabaseObject) vec.elementAt(0)).getDate() + 
	 " last:" + ((DatabaseObject) vec.elementAt(vec.size()-1)).getDate()); */
      oBuffer.addElement(average(retrieve(rdStart,rdEnd,false)));
      
      /* Next period
       * step time forward with the new sample interval
       */
      // System.out.println("Start:" + rdStart + " end:" + rdEnd);

      rdStart.setTime(rdStart.getTime()+sampleInterval);
      rdEnd.setTime(rdEnd.getTime()+sampleInterval);
    }

    oBuffer.trimToSize();
    res = new TimeSeries(oBuffer);
    res.setShortDescription(meta.getDescr());
    res.setLongDescription(meta.getDescrLong());
    return res;
  }

  /**
   * Gets a TimeSeries object.
   * If needed, the data is resampled to match the required sample interval.
   * 
   * <b>Note:</b> This method only handles data with constant timestep.
   * @param sd start date for search
   * @param ed end date for search
   * @param sampleInterval data is resampled to this interval if needed [ms]
   * @return <code>TimeSeries</code> contains data between the dates, possibly reduced
   */
  public TimeSeries getResampledTimeSeries(Date sd, Date ed, long sampleInterval) throws IOException,MetaNotSetException, TimestepException , DatabaseException {

    /* XXX this code is not tested properly !!! */

    TimeSeries res = null;

    if(!isSet()) 
      throw (new MetaNotSetException("no meta"));

    if(!meta.getCont())
      throw ( new TimestepException("Not continous data"));

    long objSize = getFirst().getObjectSize();

    int maxObjs = (int) ((ed.getTime()-sd.getTime())/sampleInterval);
    if(maxObjs < 1)
	throw new DatabaseException("Too few data items! 0");

    /* check if resample is needed */
    if(sampleInterval == meta.getTimeStep()) {
	res = retrieveTS(sd, ed,false);
	res.setShortDescription(meta.getDescr());
	res.setLongDescription(meta.getDescrLong());
	return res;
    }

    System.out.println("Doing resample, interval:" + sampleInterval);
    /* setup data buffer */
    Vector oBuffer = new Vector(maxObjs);

    /* Setup timeing for data retrieval */
    Date rdStart = new Date(sd.getTime());
    Date rdEnd = new Date(sd.getTime()+sampleInterval);

    for(int i=0;i<maxObjs;i++) { //Go through the intervals
      /* DEBUG System.out.println("retrieved " + vec.size() +
	 " objects" + " first date:" + 
	 ((DatabaseObject) vec.elementAt(0)).getDate() + 
	 " last:" + ((DatabaseObject) vec.elementAt(vec.size()-1)).getDate()); */
      oBuffer.addElement(average(retrieve(rdStart,rdEnd,false)));
      
      /* Next period
       * step time forward with the new sample interval
       */
      // System.out.println("Start:" + rdStart + " end:" + rdEnd);

      rdStart.setTime(rdStart.getTime()+sampleInterval);
      rdEnd.setTime(rdEnd.getTime()+sampleInterval);
    }

    res = new TimeSeries(oBuffer);
    res.setShortDescription(meta.getDescr());
    res.setLongDescription(meta.getDescrLong());
    return res;
  }

  /**
   * Averages this data.
   * <b>Note:</b> This method only handles data with constant timestep.
   * @param vec a <code>Vector</code> containing the data to be averaged
   * @return the averaged <code>DatabaseObject</code> with time set to
   * the middle of the input data
   */
  private DatabaseObject average(Vector vec) {

    int fields = meta.getFieldNames().length;
    double data[] = new double[fields];
    Date date;
    int cnt[] = new int[fields]; //Count the number of data points

    /* Get the ave time */
    //date = ((DatabaseObject) vec.elementAt((int) vec.size()/2)).getDate();
    date = new Date( ((DatabaseObject) vec.elementAt(0)).getDate().getTime() + ( ((DatabaseObject) vec.elementAt(vec.size()-1)).getDate().getTime() - ((DatabaseObject) vec.elementAt(0)).getDate().getTime() )/2);
    for(int i=0;i<vec.size();i++) {
      for(int j=0;j<fields;j++) {
	/* No check for overflow ... should not occur */
	double tmp = ((DatabaseObject) vec.elementAt(i)).getData(j);
	/* Skip NaN:s */
	if(Double.isNaN(tmp)) {
	  data[j]=0;
	}
	else {
	  data[j]+= tmp;
	  cnt[j]++;
	}
      }
    }
    
    /* Average */
    for(int j=0;j<fields;j++) {
      data[j]/= cnt[j];
    }
    String tmp = meta.getName();
    return (DatabaseObject.getInstance(tmp.substring(0,tmp.length()-2),date,data));
  }

  /**
   * Averages this data.
   * <b>Note:</b> This method only handles data with constant timestep.
   * @param vec a <code>Vector</code> containing the data to be averaged
   * @param ave defines the number of objects that will be averaged
   * @return a <code>Vector</code> containing <code>DatabaseObject</code>s
   * averaged according the ave parameter
   * @exception DatabaseException even number of elements
   */
  private DatabaseObject[] average(Vector vec, int ave) throws DatabaseException {

    String dbName = meta.getName();
    int fields = meta.getFieldNames().length;
    double data[] = new double[fields];
    Date date;
    int cnt[] = new int[fields]; //Count the number of data points
    int numObjPerPeriod = (int) (ave/meta.getTimeStep());
    int numPeriods =  vec.size()/numObjPerPeriod;

    if((vec.size()%numObjPerPeriod)!=0)
      throw (new DatabaseException("Not even number of items in vector"));

    DatabaseObject[] dboArr = new DatabaseObject[numPeriods]; 
    for(int k=0;k<numPeriods;k++) {
      /* Get the ave time */
      date = ((DatabaseObject) vec.elementAt((int)((int) numObjPerPeriod/2 + k*numObjPerPeriod))).getDate();

      for(int i=k*numObjPerPeriod;i<(k+1)*numObjPerPeriod;i++) {
	for(int j=0;j<fields;j++) {
	  /* No check for overflow ... should not occur */
	  double tmp = ((DatabaseObject) vec.elementAt(i)).getData(j);
	  /* Skip NaN:s */
	  if(Double.isNaN(tmp)) {
	    data[j]=0;
	  }
	  else {
	    data[j]+= tmp;
	    cnt[j]++;
	  }
	}
      }
    
      /* Average */
      for(int j=0;j<fields;j++) {
	data[j]/= cnt[j];
      }
      dboArr[k]=DatabaseObject.getInstance(dbName.substring(0,dbName.length()-2),date,data);
    }
    return dboArr;
  }

  /**
   * Gets the <code>dataGapValue</code>.
   * @return the <code>dataGapValue</code>.
   */
  protected double getDataGapValue() {
    /* XXX the dataGapValue should be changed to an
     * array
     */
    return dataGapValue;
  }

  /**
   * Stores <code>DatabaseObject</code>s to file without checking
   * the input data.
   * <p>
   * <b>Note:</b>You should not use this version of the store method !
   * See {@link #store(ObjectTimeSeries)}
   * @param vec data to be stored
   */
  public final void store(Vector vec) {

    //Check if we have metadata set,
    //if not try to load it from file (check if file exists)
    //if file doesn't exist create the first metadata.
    //Check if this object is already stored
    // boolean isThisInDatabase(obj)
    //
    try {
      //positionInFile(obj) //raf
      try {
	//OBS!!! If date is before start of data in file
	//the whole file is overwritten !!!!
	//date, from beginning, after or at this date
	//We shoulden't use position to go to the rigth place i8n the file
	//but a mathod that can handle insertion of data.

        //position(((DatabaseObject) vec.elementAt(0)).getDate(),true,false);
        boolean usemeta = true; //Position in Database with help of meta data
        //           sd, fromstart , before
        position(((DatabaseObject) vec.elementAt(0)).getDate(),false,false);

      }
      catch (IOException ee) {
	//We shoulden't use position to go to the rigth place i8n the file
	//but a mathod that can handle insertion of data.
	System.err.println(ee);
	ee.printStackTrace();
	System.out.println("This exception could be thrown if we cannot access the file or if we reach end of file. If EOF we can continue to write down our data.");
	//Safe .. go to end of file
	raf.seek(raf.length());

      }
      for(int i=0;i<vec.size();i++) {
	write(vec.elementAt(i));
      }
      //Update of MetaData
      //OBS!!! This assumes that the data is appended to the database
      DatabaseObject dboF = getFirst();
      DatabaseObject dboL = getLast();  //get Last object in file
      setDefaultMeta(dboF.getDate(),dboL.getDate());

      raf.close();
    }
    catch (IOException e) {
      System.err.println(e);
      e.printStackTrace();
    }
  }

  /**
   * Stores <code>DatabaseObject</code>s to file.
   * This version of store will do the necessary checks on 
   * the ObjectTimeSeries so that it is safe to write it to
   * our database.
   * Currently it will only allow writing if the TimeSeries follows
   * the rules for the specific database.
   * Data will not be written if it's not inside the start and
   * end time in the <code>MetaData</code>
   * (end time can be endtime + timestep).
   * @param ts <code>ObjectTimeSeries</code> containing the data
   * thats about to be stored
   */
  public final void store(ObjectTimeSeries ts) throws DatabaseException {

    /* Get a Class object from this object and then the Name of the class */
    Class myClass = this.getClass();
    String className = myClass.getName();
    String shortName = className.substring(className.lastIndexOf('.')+1);

    /* Now we can run the ObjectTimeSeries validation method to check 
     * if it's safe to write this to the Database
     */
    try {

      try {
	meta = new MetaData(shortName);
	ts.checkDbValid(meta);
	/*           sd, fromstart , before */
	position(ts.getDate(0),false,false);
      }	
      catch (IOException ee) {
	System.err.println(ee);
	/* This happens if the MetaData could not be loaded.
	 * Trying to create a MetaData which represents times in 
	 * our ObjectTimeSeries object ...
	 * this will hopefully be updated when the data write is
	 * finished below ... 
	 */
	System.out.println("Loading empty meta data ...");
	meta = getEmptyMetaData(ts.getDate(0),ts.getDate(ts.getLength()-1));
	/* OBS!! The metadata is not stored above !!! */
	System.out.println("got it!");
	System.out.println("Validate it ...");
	ts.checkDbValid(meta);
	System.out.println("done!");
	System.out.println("positioning (slow)");
	/* slow position, note fromstart */
	/*       sd, fromstart , before */
	position(ts.getDate(0),true,false);
	System.out.println("done!");
      }

    }
    catch (TimeSeriesEODException e) {
      /* If this happens we have either a datagap between incoming data and that 
       * what we have in the database or the incoming data is exactly one timestep away
       * from the last item in the database.
       * For now we accept a gap of one month.
       * We will create a new ObjectTimeSeries with empty values
       * and try to insert it if it's needed.
       */
        if((ts.getDate(0).getTime() - meta.getEndDate().getTime())
        /1000 > 60*60*24*30) {
            System.out.println("Your input data is too far into the future. No update !");
            throw new saaps.dbt.DatabaseException("Your input data is too far into the future. No update !");
        }
        else {            
            try {
            Date lastInDb = meta.getEndDate();
            /* If the first objects date/time is exactly one timestep away
             * we don't need to insert stuff just to move the raf to the end of the db.
             * Else we do the insertion ....
             */
                if((ts.getDate(0).getTime() - lastInDb.getTime()) == meta.getTimeStep()) {
                    raf.seek(raf.length());
                }
                else {
                    System.out.println("Will insert empty data for period:" + lastInDb + " to " + ts.getDate(0));
                    ts = ts.fitToDatabase(meta,lastInDb);
                    System.out.println("New data is: " + ts.getDate(0) + " " + ts.getDate(ts.getLength()-1));
                    raf.seek(raf.length());
                }
            }
            catch(IllegalTimeSeriesException er) {
                System.err.println(er);
                er.printStackTrace();
                throw new saaps.dbt.DatabaseException(er.toString());
            }
        /* XXX check this ! */
            catch (IOException er) {
                System.err.println(er);
                er.printStackTrace();
                throw new saaps.dbt.DatabaseException(er.toString());
            }
            
        }
    }
    catch (TimeSeriesSequenceException e) {
        System.err.println(e);
        throw new saaps.dbt.DatabaseException(e.toString());
    }
    catch (TimeSeriesContinuityException e) {
        System.err.println(e);
        throw new saaps.dbt.DatabaseException(e.toString());
        //e.printStackTrace();
    }
    catch (IllegalTimeSeriesException e) {
        System.err.println(e);
        throw new saaps.dbt.DatabaseException(e.toString());
        //e.printStackTrace();
    }
    catch (IOException e) {
        System.err.println(e);
        e.printStackTrace();
        throw new saaps.dbt.DatabaseException(e.toString());
    }
    catch (ClassNotFoundException e) {
        System.err.println(e);
        e.printStackTrace();
        throw new saaps.dbt.DatabaseException(e.toString());
    }
    
    try {
        for(int i=0;i<ts.getLength();i++) {
            write(ts.getDataObject(i));
        }
      /* Update of MetaData
       * OBS!!! This assumes that the data is appended to the database
       * OBS!! What happens if the write cannot complete and we get an
       * exception and the MetaData is not updated ???
       * Anyway ... if we are able to update the MetaData the info. comes
       * from the file itself due to the getFirst and getLast methods ...
       */
        DatabaseObject dboF = getFirst();
        DatabaseObject dboL = getLast();
        //System.out.println("dbo size:" + dboF.getObjectSize() + " data size:" + dboF.getDataSize() );
        setDefaultMeta(dboF.getDate(),dboL.getDate());
        //raf.close();
    }
    catch (IOException eee) {
        System.err.println(eee);
        eee.printStackTrace();
    }
  }

  /** 
   * Imports ASCII data to our database.
   * The format of the data is as output from ExportAscii application.
   * It will fill a buffer of <code>DatabaseObject</code>s and then try
   * to store it, read again and store until it is finished or an 
   * exception occures.
   * <p>
   * <b>Note:</b> Currently the buffering doesn't work ! vectors
   * below 200Mb in size is imported correctly.
   * @param inputfile the file containing the data to be imported
   */
  protected void importAscii(String inputfile) throws DatabaseException, Exception {
    final int MAXSIZE=209715200; //200Mb //1048576; bytes , 1Mb 524288000;//500 Mb

    Date inputDateTime;
    String tmpVal;
    double tmpKp;

    Vector vec = new Vector(8*30);

    try {
      /* get name of database objects */
      String objectName = meta.getName().substring(0, meta.getName().length()-2);
      short numFields = (short) meta.getFieldNames().length;
      double[] data = new double[numFields];

      SimpleDateFormat formatter = (SimpleDateFormat) SimpleDateFormat.getDateTimeInstance(0,0,new Locale("sv","SE"));

      formatter.setTimeZone(TimeZone.getTimeZone("UTC"));
      formatter.applyPattern("yyyy.MM.dd HH:mm:ss");

      Calendar cal = Calendar.getInstance();
      cal.setTimeZone(formatter.getTimeZone());

      /* set default maximum number of objects */
      int maxObjs = 100;
      if(meta.getCont()) { //Calculate if we have continous data
	long objSize = DatabaseObject.getInstance(objectName,new Date()).getObjectSize();
	maxObjs = (int) (MAXSIZE/objSize);
      }
      int cnt=0;

      BufferedReader streamIn = new BufferedReader(new FileReader(inputfile));
      String textLine = "";
      textLine = streamIn.readLine();
      /* Go through the file for all lines starting with
       * "#", ":", or " ".
       */
      while (!textLine.startsWith("1") && !textLine.startsWith("2")) {
	textLine = streamIn.readLine();
      }

      while(textLine != null) {
	while ((textLine != null) && ( cnt < maxObjs) ) {
	  cnt++;
	  /* Read the date using - as the field separator. */
	  String dateStr = textLine.substring(0,4) + "." +
	    textLine.substring(5,7) + "." +
	    textLine.substring(8,10) + " " + 
	    textLine.substring(11,13) + ":" + 
	    textLine.substring(14,16) + ":" + 
	    textLine.substring(17); //Last item to end of string
     
	  /* Parse the string back into a Date. */
	  ParsePosition pos = new ParsePosition(0);
	  inputDateTime = formatter.parse(dateStr, pos);
	
	  cal.setTime(inputDateTime);
	  tmpVal = textLine.substring(20).trim();
	  StringTokenizer tokens = new StringTokenizer(tmpVal);
	  int numTok = tokens.countTokens();
	  for(int i=0;i<numTok;i++) {
	    String value = tokens.nextToken();
	    if(value.equals(new String("NaN"))) {
	      data[i] = Double.NaN;
	    }
	    else {
	      data[i] = Double.valueOf(value).doubleValue();
	    }
	    /* XXX problem with OMNIDB */
	    if (data[i] == dataGapValue ) {
	      data[i] = Double.NaN;
	    }
	  }
	  vec.addElement(DatabaseObject.getInstance(objectName,cal.getTime(),data));
	  textLine = streamIn.readLine();
	} //end while textline,maxobjs

	/* Store our buffer of objects */
	vec.trimToSize();
	ObjectTimeSeries ts=null;

	if(!isSet()) { //No metadata use this vector data
	  meta = getEmptyMetaData(((DatabaseObject) vec.elementAt(0)).getDate(), ((DatabaseObject) vec.elementAt(vec.size()-1)).getDate());
	}
	ts = new ObjectTimeSeries(vec);

        if(!isSet())
	  System.out.println("MetaData is unavailable \nWill try to create it.");

	store(ts);

	if(!isSet()) { //If no meta reload it
	  meta = new MetaData(meta.getName());
	  //System.out.println("The DB meta: " + theDB.meta.getDescr() );
	  isSet=true;
	}
	/* Reset the vector and counter */
	vec = new Vector(8*30);
	cnt=0;
      }
      streamIn.close();
    }
    catch (IOException e) {
      System.err.println("IO error: " + e);
    }
    catch (NumberFormatException e) {
      System.err.println("Number format error: " + e);
    }
    catch (IllegalTimeSeriesException e) {
      System.err.println(e);
    }
    catch (ClassNotFoundException e) {
      System.err.println(e);
    }
  }

  /* Abstract methods for reading and writing from/to the files. */

  /**
   * Gets the object at the current point in file.
   * @return an object from the database
   * @exception IOException when reading fails
   */
  public abstract Object read() throws IOException;

  /**
   * Writes data at the current point in file.
   * @param dbo an object containg data to be written.
   * @exception IOException when writeing fails
   */
  public abstract void write(Object dbo) throws IOException;

  /**
   * Gets the date from this object.
   * See also {@link saaps.dbt.DatabaseObject#getDate()}.
   * @param obj a database object
   * @return the <code>Date</code> for the input object
   */
  public abstract Date getDate(Object obj);

  /**
   * <b> Do not use this method it will be removed ! </b>
   */
  public abstract Vector getLatestDatabaseObjects();

  /**
   * Sets the <code>MetaData</code> for the specific database and 
   * stores it to file.
   * @param sd start time of data in datafile
   * @param ed end time of data in datafile
   * @return the new <code>MetaData</code> object
   * @exception IOException when save of <code>MetaData</code> fails
   */
  protected abstract MetaData setDefaultMeta(Date sd, Date ed) throws IOException;

  /**
   * Gets a <code>MetaData</code> object with the dates set to
   * the supplied dates.
   * @param sd fake start date
   * @param ed fake end date
   * @return the fake <code>MetaData</code> object
   */
  protected abstract MetaData getEmptyMetaData(Date sd, Date ed);

}
