package saaps.dbt;

import java.rmi.Remote;
import java.rmi.RemoteException;

import java.util.Vector;
import java.util.Date;
import java.io.IOException;

import saaps.util.TimeSeries;
import saaps.util.ObjectTimeSeries;
import saaps.util.TimeSeriesIndexException;
import saaps.util.IllegalTimeSeriesException;
import saaps.util.EventList;
import saaps.util.ModelList;
import saaps.util.Model;
import saaps.saam.BestPredModel;
import saaps.saam.ElFluenceLevel;
import saaps.saam.SumKp;
import saaps.saam.SumKpLevel;
import saaps.network.*;

/**
 * Remote interface for <code>Database</code> access.
 *
 * @author Leif Kalla
 */
public interface DatabaseRMI extends Remote {

  /**
   * Gets a list of available <code>Database</code>s.
   * @return <code>Vector</code> of <code>Database</code> names
   * @exception RemoteException network/io problem
   */
  Vector getDatabaseList() throws RemoteException;

  /**
   * Gets the <code>MetaData</code> for this <code>Database</code>.
   * @param theDb <code>Database</code> name
   * @return <code>MetaData</code> for this <code>Database</code>
   * @exception RemoteException network/io problem
   */
  MetaData getMetaData(String theDb) throws RemoteException;

  /**
   * Gets a <code>Vector</code> of <code>Database</code> objects.
   * @param theDb <code>Database</code> name
   * @param sd start date/time
   * @param ed end date/time
   * @exception RemoteException network/io problem
   * @exception IOException when io problem or no data for this time
   */
  Vector getData(String theDb, Date sd, Date ed) 
    throws RemoteException, IOException;

  /**
   * Gets a <code>TimeSeries</code> object.
   * @param theDb <code>Database</code> name
   * @param sd start date/time
   * @param ed end date/time
   * @return <code>TimeSeries</code>
   * @exception RemoteException network/io problem
   * @exception IOException when io problem or no data for this time
   */
  TimeSeries getTimeSeries(String theDb, Date sd, Date ed) 
    throws RemoteException, IOException;

  /**
   * Gets an <code>ObjectTimeSeries</code> containing data between
   * the dates.
   * @param theDB SAAPS database name
   * @param sd start date for search
   * @param ed end date for search
   * @param before if <code>true</code> then the first object returned
   * is before the start search date.
   * @return the ObjectTimeSeries
   */
  ObjectTimeSeries retrieveOTS(String theDB,Date sd, Date ed, boolean before)
    throws IOException, IllegalTimeSeriesException;

  /**
   * Retrieves a TimeSeries object containing the averaged data for the
   * input event list. See also 
   * {@link Database#retrieveSEA(int, Date[], long, long, boolean )}.
   * @param theDB name of the SAAPS <code>Database</code>
   * @param fieldIndex  index of the data field
   * @param events array of event times
   * @param start event window starts here (normally negative [milliseconds])
   * @param end event window ends here (normally positiv [milliseconds])
   * @param withAdditionalInfo if <code>true</code> get also the number of 
   * data points and the standard deviation from the average.
   * @return <code>TimeSeries</code> object with <code>fieldName</code>
   * data for start to end time for every event
   * @exception TimeSeriesIndexException when <code>fieldIndex</code>
   * is out of bounds
   * @exception IOException when data for any of the event times are missing
   */
  TimeSeries retrieveSEA(String theDB, int fieldIndex,Date[] events,
			 long start, long end, boolean withAdditionalInfo)
    throws RemoteException, TimeSeriesIndexException, IllegalTimeSeriesException, IOException;

  /**
   * Gets the parameters for the neural network.
   * @param netParam network parameter name
   * @return <code>NetworkParam</code>
   * @exception RemoteException network/io problem
   */
  NetworkParam getNetworkParam(String netParam) throws RemoteException;

  /**
   * Gets the normalization parameters needed for the preprocessing
   * of a <code>TimeSeries</code> object before it is input to a network.
   * @param normParam normalization parameter name
   * @param nof number of fields
   * @return <code>NormParam</code>
   * @exception RemoteException network/io problem
   */
  NormParam getNormParam(String normParam, int nof) throws RemoteException;

  String[][] getBestPredModel(EventList eventList,EventList nonEventList,int statistics) throws RemoteException;
  
  String [] getModelListString() throws RemoteException;

  Model getModel(int m) throws RemoteException;
  
  TimeSeries getModelTS(int i,Date d1, Date d2) throws RemoteException;

  TimeSeries getModelTS(int i,Date predDate) throws RemoteException;
  
  TimeSeries retrieveElFluence() throws RemoteException;

  TimeSeries retrieveElFluence(Date d) throws RemoteException;
  
  ElFluenceLevel getOptimizedElFluenceLevel(EventList events,EventList nonEvents) throws RemoteException, IOException;

  ElFluenceLevel getElFluenceLevel(EventList events,EventList nonEvents,int[] td,double[] lev) throws RemoteException, IOException;

  SumKpLevel sumKpLevelOptimize(SumKpLevel skl) throws RemoteException;
  SumKpLevel sumKpLevelCalcStat(SumKpLevel skl) throws RemoteException;
  
  TimeSeries getAnom002ModelTS(Date d1, Date d2) throws RemoteException;
}
