package saaps.dbt;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.*;
import java.text.*;
import java.util.*;

import java.io.IOException;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
import java.text.DateFormat;
import java.util.Date;
import java.util.Calendar;
import java.util.Locale;
import java.util.TimeZone;
import java.util.Vector;

import java.rmi.Naming;
import java.rmi.RemoteException;
import java.rmi.RMISecurityManager;
import java.rmi.server.UnicastRemoteObject;

import saaps.sapm.Anom002Model;
import saaps.sapm.ElFluence;
import saaps.util.TimeSeries;
import saaps.util.ObjectTimeSeries;
import saaps.util.TimeSeriesIndexException;
import saaps.util.IllegalTimeSeriesException;
import saaps.util.EventList;
import saaps.util.ModelList;
import saaps.util.Model;
import saaps.saam.BestPredModel;
import saaps.saam.ElFluenceLevel;
import saaps.saam.SumKp;
import saaps.saam.SumKpLevel;
import saaps.network.*;

/**
 * Provides a RMI server for <code>Database</code> access.
 * Through the use of the <code>DatabaseRMI</code> interface
 * it's possible to:
 * <p>
 * get information of available databases,
 * <p>
 * get <code>MetaData</code> for a specific databse,
 * <p>
 * get different types of data from the database.
 *
 * @author Leif Kalla
 */
public class DatabaseRMIImpl extends UnicastRemoteObject
  implements DatabaseRMI {

  /**
   * Runs the constructor of the superclass <code>UnicastRemoteObject</code>.
   * @exception RemoteException network/io problem
   */
  public DatabaseRMIImpl() throws RemoteException {
    super();
  }
  
  /**
   * Gets a list of available <code>Database</code>s.
   * @return <code>Vector</code> of <code>Database</code> names
   */
  public Vector getDatabaseList() throws RemoteException {
    return (DatabaseItems.list());
  }

  /**
   * Gets the <code>MetaData</code> for this <code>Database</code>.
   * @param theDb <code>Database</code> name
   * @return <code>MetaData</code> for this <code>Database</code>
   */
  public MetaData getMetaData(String theDb) throws RemoteException {

    MetaData meta=null;

    try { // To load the meta data for this database from file
      meta = new MetaData(theDb);
    }
    catch (IOException e) {
      System.err.println(e);
      e.printStackTrace();

      System.out.println("Here we should handle the situation when no DB exists");
    }
    catch (ClassNotFoundException ee) {
      System.err.println(ee);
      ee.printStackTrace();
    }
    return meta;
  }

  /**
   * Gets a <code>Vector</code> of <code>Database</code> objects.
   * @param theDb <code>Database</code> name
   * @param sd start date/time
   * @param ed end date/time
   * @exception IOException when io problem or no data for this time
   */
  public Vector getData(String theDb,Date sd,Date ed) throws RemoteException, IOException {

    Database selDb = null;
    Vector theData = null;

    selDb = Database.getInstance(theDb,true);

    /* An IOException is thrown is start date is before start of data
     * in file.
     * XXX here we should get an DataNotFoundException instead !! 
     */
    theData = selDb.retrieve(sd,ed,false);
    return theData;
  }

  /**
   * Gets a <code>TimeSeries</code> object.
   * @param theDb <code>Database</code> name
   * @param sd start date/time
   * @param ed end date/time
   * @return <code>TimeSeries</code>
   * @exception IOException when io problem or no data for this time
   */
  public TimeSeries getTimeSeries(String theDb,Date sd,Date ed) throws RemoteException, IOException {

    Database selDb;

    selDb = Database.getInstance(theDb,true);

    /* An IOException is thrown if start date is before start of data
     * in file and if start date is after end of data in file.
     * XXX here we should get an DataNotFoundException instead !! 
     */
    TimeSeries ts = selDb.retrieveTS(sd,ed,false);
    selDb.close();
    return (ts);
  }

  /**
   * Gets an <code>ObjectTimeSeries</code> containing data between
   * the dates.
   * @param theDB SAAPS database name
   * @param sd start date for search
   * @param ed end date for search
   * @param before if <code>true</code> then the first object returned
   * is before the start search date.
   * @return the ObjectTimeSeries
   */
  public ObjectTimeSeries retrieveOTS(String theDB,Date sd, Date ed, boolean before)
    throws IOException, IllegalTimeSeriesException {
    Database selDB;

    selDB = Database.getInstance(theDB,true);
    ObjectTimeSeries ots = selDB.retrieveOTS(sd,ed,before);
    selDB.close();

    return ots;
  }

  /**
   * Retrieves a TimeSeries object containing the averaged data for the
   * input event list. See also 
   * {@link Database#retrieveSEA(int, Date[], long, long, boolean )}.
   * @param theDB name of the SAAPS <code>Database</code>
   * @param fieldIndex  index of the data field
   * @param events array of event times
   * @param start event window starts here (normally negative [milliseconds])
   * @param end event window ends here (normally positiv [milliseconds])
   * @param withAdditionalInfo if <code>true</code> get also the number of 
   * data points and the standard deviation from the average.
   * @return <code>TimeSeries</code> object with <code>fieldName</code>
   * data for start to end time for every event
   * @exception TimeSeriesIndexException when <code>fieldIndex</code>
   * is out of bounds
   * @exception IOException when data for any of the event times are missing
   */
  public TimeSeries retrieveSEA(String theDB,int fieldIndex,Date[] events,long start,long end,boolean withAdditionalInfo) throws RemoteException, TimeSeriesIndexException, IllegalTimeSeriesException, IOException {
    Database selDB;

    selDB = Database.getInstance(theDB,true);
    TimeSeries ts = selDB.retrieveSEA(fieldIndex ,events,start,end,withAdditionalInfo);
    selDB.close();

    return ts;
  }

  /**
   * Gets the parameters for the neural network.
   * @param netParam network parameter name
   * @return <code>NetworkParam</code>
   */
  public NetworkParam getNetworkParam(String netParam) throws RemoteException {

    /* We don't really need to give the applet the possibility to
     * choose which file to use. (It doesn't know anything about 
     * the files anyway, yet ...)
     * i.e String netParam could be omitted
     */
    return (new NetworkParam(netParam));
  }

  /**
   * Gets the normalization parameters needed for the preprocessing
   * of a <code>TimeSeries</code> object before it is input to a network.
   * @param normParam normalization parameter name
   * @param nof number of fields
   * @return <code>NormParam</code>
   */
  public NormParam getNormParam(String normParam,int nof) throws RemoteException {

    /* We don't really need to give the applet the possibility to
     * choose which file to use. (It doesn't know anything about 
     * the files anyway, yet ...)
     * i.e String normParam could be omitted
     */

    return(new NormParam(normParam, nof));
  }

  /**
   * Gets the names of the best prediction model for this input.
   * @param eventList 
   * @param nonEventList 
   * @param statistics determines BestPredModel operation
   * @return Strings with best model names
   */
  public String[][] getBestPredModel(EventList eventList,EventList nonEventList,int statistics) throws RemoteException {
      String linCorr = "Linear correlation";
      String condProb = "Conditional probability";

      BestPredModel bestPredModel = new BestPredModel();

      bestPredModel.setEventList(eventList);
      bestPredModel.setNonEventList(nonEventList);

      bestPredModel.update();
      bestPredModel.runModels();
      bestPredModel.setStatistics(statistics);
      bestPredModel.computeStatistics();
      bestPredModel.sortModels();
      String[][] result = new String[bestPredModel.getNumOfModels()][2];
      for (int i=0;i<bestPredModel.getNumOfModels();i++) {
          result[i][0] = new String(bestPredModel.getModelName(i));
          result[i][1] = new String(bestPredModel.toString(i));
      }
      return result;
  }

  /**
   * Gets the available model names.
   * @return model names
   */
  public String [] getModelListString() throws RemoteException {
      ModelList ml = new ModelList();
      return (ml.getListString());
  }
    
  /**
   * Gets the specified Model.
   * @param m the Model index
   * @return the selected Model
   */
  public Model getModel(int m) throws RemoteException {
      ModelList ml = new ModelList();
      return (ml.getModel(m));
  }
  
    /**
     * Gets a TimeSeries for this <code>Model</code> between the dates.
     * @param i <code>ModelList</code> index to <code>Model</code>
     * @param d1 start date
     * @param d2 end date
     * @return Resulting TimeSeries from this Model run
     */
    public TimeSeries getModelTS(int i,Date d1, Date d2) throws RemoteException {
	return (new ModelList()).getModelTS(i,d1,d2);
    }

    /**
     * Gets a TimeSeries for this <code>Model</code> at this prediction date.
     * @param i <code>ModelList</code> index to <code>Model</code>
     * @param predDate prediction date
     * @return Resulting TimeSeries from this Model run
     */
    public TimeSeries getModelTS(int i,Date predDate) throws RemoteException {
	return (new ModelList()).getModelTS(i,predDate);
    }

    public ElFluenceLevel getOptimizedElFluenceLevel(EventList events,EventList nonEvents) throws RemoteException, IOException {
        ElFluenceLevel efl = new ElFluenceLevel();
        efl.setEventList(events);
        efl.setNonEventList(nonEvents);
        efl.optimize();
        return efl;
    }

    public ElFluenceLevel getElFluenceLevel(EventList events,EventList nonEvents,int[] td,double[] lev) throws RemoteException, IOException {
        ElFluenceLevel efl = new ElFluenceLevel();
        efl.setEventList(events);
        efl.setNonEventList(nonEvents);
        efl.setTimeDelay(td);        
        efl.setLevel(lev);
        efl.calcStat();
        return efl;
    }

    /**
     * Runs the optimize method of the SumKpLevel class.
     * The data should be setup before the reference is passed here.
     * @param skl the object should be ready for optimize method
     */
    public SumKpLevel sumKpLevelOptimize(SumKpLevel skl) throws RemoteException {
        skl.optimize();
        return skl;
    }

    /**
     * Runs the calcStat method of the SumKpLevel class.
     * The data should be setup before the reference is passed here.
     * @param skl the object should be ready for calcStat method
     */
    public SumKpLevel sumKpLevelCalcStat(SumKpLevel skl) throws RemoteException {
        skl.calcStat();
        return skl;
    }

  /**
   * The main for the RMI server.
   * Creates an instance of the default security manager and
   * binds the <code>DatabaseRMIImpl</code> to the registry.
   */
  public static void main(String args[]) {

    /* XXX check how to run the remote method invocation daemon
     * so that this need not run all the time.
     */

    if (System.getSecurityManager() == null) {
      System.setSecurityManager(new RMISecurityManager());
    }
    try {
      DatabaseRMIImpl obj = new DatabaseRMIImpl();
      /* Bind this object instance to the name "DatabaseRMIServer" */
      Naming.rebind("//localhost:2001/DatabaseRMIServer", obj); //sol.irfl.lu.se
      System.out.println("DatabaseRMIServer bound in registry");
    } catch (Exception e) { // remote exception is thrown here ??
      System.out.println("DatabaseRMIImpl err: " + e.getMessage());
      e.printStackTrace();
    }
  }
  
  public TimeSeries retrieveElFluence() throws RemoteException {
    ElFluence fluence = new ElFluence();
    return fluence.getFluence();
  }

  public TimeSeries retrieveElFluence(Date d) throws RemoteException {
    ElFluence fluence = new ElFluence(d);
    return fluence.getFluence();
  }
  
  public TimeSeries getAnom002ModelTS(Date d1,Date d2) throws RemoteException {
      TimeSeries ts = null;
      Anom002Model model = new Anom002Model();
      model.readTimeSeries(d1,d2);
      model.run();
      return model.getOutput();
  }
}
