package saaps.dbt;

import java.io.*;
import java.util.*;

/**
 * GOES10partDB extends the Database class. It contains
 * the read and write methods that are used in
 * Database. The field is a vector of ACE objects.
 *
 *@author Peter Wintoft
 *@version 1.0 2000-02-22
 *@see GOES10part
 */
public class GOES10partDB extends Database {

  private static String cName = "GOES10part";
  private static final double dataGapValueDefinition = -1.00e+05;

  // Static Fields used to initiate the MetaData
  private static String name= cName + "DB";
  private static String description = "GOES-10 particle data";
  private static String longDescription = "The GOES-10 particle data contain " +
    "geosynchronous electron and proton flux data with 5 minute resolution. " +
    "The electron flux cover three energy levels: >0.6MeV, >2MeV, and >4MeV. " +
    "The >4MeV range is mostly empty. " +
    "The proton flux cover 6 energy levels: >1MeV, >5MeV, >10MeV, >30MeV, >50MeV, and >100MeV. " +
    "The longitude of the satellite is also given. ";

  private static Calendar st=Calendar.getInstance(TimeZone.getTimeZone("UTC"),new Locale("sv","SE"));
  private static Calendar et=Calendar.getInstance(TimeZone.getTimeZone("UTC"),new Locale("sv","SE"));
  private static boolean ct=true;
  private static long step=5*60000; //Five minute timestep
  private static String[] fieldNames = { "P>1 MeV", "P>5 MeV", "P>10 MeV", "P>30 MeV", "P>50 MeV",
                                            "P>100 MeV", "E>0.6 MeV", "E>2.0 MeV", "E>4.0 MeV", "Longitude" };
  private static String[] fieldUnits = { "1/(cm^2 s sr)", "1/(cm^2 s sr)", "1/(cm^2 s sr)", "1/(cm^2 s sr)", "1/(cm^2 s sr)",
                                            "1/(cm^2 s sr)", "1/(cm^2 s sr)", "1/(cm^2 s sr)", "1/(cm^2 s sr)", "Deg" };
  private static String[] fieldTypes = { "Double", "Double", "Double","Double","Double",
                                          "Double", "Double", "Double", "Double", "Double" };

  // Fields.
  private GOES10part goes;
  private Vector goesv;

  // Constructors.
  public GOES10partDB() {
    
    dataGapValue=dataGapValueDefinition;
  }
  public GOES10partDB(int i) {
    
    dataGapValue=dataGapValueDefinition;
  }
  /**
   * When a new class is added to the database this
   * constructor should be used which includes the
   * information that goes into the meta data file.
   *
   * The arguments are: Vector of GOES objects, the class name,
   * a short description, continuos time series (false/true), and
   * if a continous time series the time step.
   */
  public GOES10partDB(Vector GOESV) throws IOException {
    
    dataGapValue=dataGapValueDefinition;
    goesv = GOESV;
    raf = new RandomAccessFile(baseDir + cName + "DB.data","rw");
    // Then get its metadata
    try { // To load the meta data for this database from file
      meta = new MetaData(cName + "DB");
      isSet=true;
    }
    catch (IOException e) {
      System.err.println(e);
      System.out.println("Here we should handle the situation when no DB exists");
      try {
  meta = setDefaultMeta(((DatabaseObject) (goesv.elementAt(0))).getDate(),((DatabaseObject) (goesv.elementAt(goesv.size()-1))).getDate()); //Takes static fields from XDB

  isSet=true;
      }
      catch (IOException ee) {
  System.err.println(ee);
  ee.printStackTrace();
      }
    }
    catch (ClassNotFoundException e) {
      System.err.println(e);
      e.printStackTrace();
    }

    store((Vector) goesv); //Need to type cast so that store(Object obj) isn't used
  }

  // Methods.
  public Object read() throws IOException {
    goes = new GOES10part(raf.readLong(),
                          raf.readDouble(),raf.readDouble(),raf.readDouble(),raf.readDouble(),raf.readDouble(),
                          raf.readDouble(),raf.readDouble(),raf.readDouble(),raf.readDouble(),raf.readDouble());
    return (Object) goes;
  }

  public void myHello() {
    System.out.println("My" + cName + "DB Hello");
  }

  public void write(Object obj) throws IOException {
    goes = (GOES10part) obj;
    raf.writeLong(goes.getDate().getTime());
    raf.writeDouble(goes.getP1());
    raf.writeDouble(goes.getP5());
    raf.writeDouble(goes.getP10());
    raf.writeDouble(goes.getP30());
    raf.writeDouble(goes.getP50());
    raf.writeDouble(goes.getP100());
    raf.writeDouble(goes.getE06());
    raf.writeDouble(goes.getE2());
    raf.writeDouble(goes.getE4());
    raf.writeDouble(goes.getLongitude());
  }
  // Implemented for Database abstract getDate
  public Date getDate(Object obj) {
    return ((GOES10part) obj).getDate();
  }

  /** This method is implemented for Database abstract metod.
   * It retrieves the latest data from the dataprovider and returns the vector of Database objects.
   */
  public Vector getLatestDatabaseObjects() {
    Vector vec = new Vector(24);
    Date mdate = new Date();

    GOES10part goes = new GOES10part(mdate,1.0,2.0,3.0,4.0,5.0,6.0,7.0,8.0,9.0,10.0);

    for(int i=0;i<8*30;i++) {
      vec.addElement(goes);
      mdate.setDate( (int) (mdate.getDate() + 10));
      goes.setDate(mdate);
      goes.setP1(i);
      goes.setP5(i);
      goes.setP10(i);
      goes.setP30(i);
      goes.setP50(i);
      goes.setP100(i);
      goes.setE06(i);
      goes.setE2(i);
      goes.setE4(i);
      goes.setLongitude(i);
    }
    vec.trimToSize();
    return vec;
  }

  /** Creates a metadata file for this class with the default values set
   *to the static values defined above. OBS!, check calendar ....
   */
  protected MetaData setDefaultMeta(Date sd, Date ed) throws IOException {

    st.setTime(sd);
    st.clear(st.MILLISECOND);
    et.setTime(ed);
    et.clear(et.MILLISECOND);

    MetaData m = new MetaData(name,description, longDescription,  st, et, ct, step,fieldNames,fieldUnits,fieldTypes);
    return m;
  }
  /**
   * Returns an "empty" metadata object. All fields but Start and end time
   * has valid values (as set in the respective database class).
   * The times will be set to that what is supplied.
   * OBS! This is not stored in the MetaData file !!
   */
  protected MetaData getEmptyMetaData(Date sd, Date ed) {
    MetaData m = new MetaData();
    st.clear(st.MILLISECOND);
    st.setTime(sd);
    et.clear(et.MILLISECOND);
    et.setTime(ed);
    m.setMetaData(name,description, longDescription, st, et, ct, step, fieldNames, fieldUnits, fieldTypes );
    return m;
  }

}
