package saaps.dbt;

import java.util.Calendar;
import java.util.Date;
import java.util.Locale;
import java.util.TimeZone;
import java.text.DateFormat;
import java.text.NumberFormat;

import java.io.Serializable;
import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.lang.Integer;

/**
* A class for Kp objects.
*
* The class contains the date, the Kp value, and a status flag
* indicating the type of Kp. There are four possible values on
* that the status flag can have:
* <ul>
* <li> 1 = final Kp,
* <li> 2 = estimated Kp from SEC,
* <li> 3 = nowcasted Kp from Lund,
* <li> 4 = 3-hour forecasted Kp from Lund.
* </ul>
*
* @author Peter Wintoft
* @version 1.2
* @version 1.1 990718
* @version 1.0 990604
*/
public class Kp extends DatabaseObject implements Serializable
{
  // Fields
  private static int numberOfIndexes = 2;
  private Date d;
  private double kp;
  private double status;
  
  // Constructors
  public Kp()
  {
  }

  /**
   * This constructor is used if you want an "empty" object with this date.
   */
  public Kp(Date D) {
    this.d = D;
    this.kp = Double.NaN;
    this.status = 1.0;
  }

  /**
  * Construct a Kp object by giving the date and the two data field
  * values. The data fields are Kp and status.
  *
  *@param D The date.
  *@param kp The Kp value.
  *@param status The status.
  */
  public Kp(Date D, double kp, double status) {
    this.d = D;
    this.kp = kp;
    this.status = status;
  }

  /**
  * Construct a Kp object by giving the date in milliseconds and the two data field
  * values. The data fields are Kp and status.
  *
  *@param l The date in milliseconds.
  *@param kp The Kp value.
  *@param status The status.
  */
  public Kp(long l, double kp, double status) {
    this(new Date(l),kp,status);
  }

  /**
  * Construct a Kp object by giving the date and the two data field
  * values. The data fields are Kp and status.
  *
  *@param D The date.
  *@param data Array with the the two data field values.
  */
  public Kp(Date D, double[] data) {
    this.d = D;
    this.kp = data[0];
    this.status = data[1];
  }

  /**
  * Construct a Kp object by giving the date and the Kp value. The
  * status is assumed to be 1.
  *
  *@param D The date.
  *@param KP The Kp value.
  *@deprecated Replaced by {@link Kp(Date,double[])}.
  */
  public Kp(Date D, double KP)
  {
    this.d = D;
    this.kp = KP;
    this.status = 1;
  }

  /**
  * Construct a Kp object by giving the date in milliseconds
  * and the two data field
  * values. The data fields are Kp and status.
  *
  *@param L The date in milliseconds.
  *@param data Array with the the two data field values.
  */
  public Kp(long L, double[] data)
  {
    this.d = new Date(L);
    this.kp = data[0];
    this.status = data[1];
  }
  
  /**
  * Construct a Kp object by giving the date in milliseconds
  * and the Kp value. The status is assumed to be 1.
  *
  *@param L The date in milliseconds.
  *@param KP The Kp value.
  *@deprecated Replaced by {@link Kp(long,double[])}.
  */
  public Kp(long L, double KP)  // 990718
  {
    this.d = new Date(L);
    this.kp = KP;
    this.status = 1;
  }
  
  //Implemented for DatabaseObject getDate
  public int getNumberOfDataIndexes() {
    return (this.numberOfIndexes);
  }
  //Implemented for DatabaseObject getDate
  public Date getDate()
  {
    return this.d;
  }
  public double getKp()
  {
    return this.kp;
  }
  public double getStatus()
  {
    return this.status;
  }
  // This method is the generic form of accessing data.
  // This should be changed in the future to return
  // either Vector or double[] containing both kp and status.
  public double getData()
  {
  	return kp;
  }

  //Implemented for DatabaseObject getData
  public double getData(int index) {

    switch (index) { //No breks allowed due to return statement
    case 0 :
      return(this.kp);
    case 1 :
      return(this.status);
    default :
      System.out.println("Index out of bounds for Kp");
      return (0.0);
    }
  }

  /**
   * This method returns the size of the data that is to be written to file
   */
  public long getDataSize() throws IOException {

    ByteArrayOutputStream baos = new ByteArrayOutputStream();  
    DataOutputStream    dos  = new DataOutputStream(baos);  
    dos.writeLong(this.d.getTime());
    dos.writeDouble(this.kp);  
    dos.writeDouble(this.status);  
    dos.flush();  
    long size = baos.size();  // now you know how large ....
    return size;
  }

  public void setDate(Date D)
  {
    this.d = D;
  }
  public void setDate(long L)  // 990718
  {
    this.d = new Date(L);
  }
  public void setKp(double KP)
  {
    this.kp = KP;
  }
  public void setStatus(double s)
  {
    this.status = s;
  }

  public String toString() {
    // Set the date format.
//     DateFormat df = DateFormat.getDateTimeInstance(DateFormat.DEFAULT,DateFormat.DEFAULT,new Locale("sv","SE"));
//     df.setTimeZone(TimeZone.getTimeZone("UTC"));

    // Find the selected dates.
    Calendar cal = Calendar.getInstance(TimeZone.getTimeZone("GMT"));
    cal.clear(cal.MILLISECOND);
    cal.setTime(this.d);

    NumberFormat nf = NumberFormat.getInstance();
    nf.setMinimumIntegerDigits(2);

    return ( cal.get(Calendar.YEAR) + " " + nf.format(cal.get(Calendar.MONTH)+1) + " " + nf.format(cal.get(Calendar.DAY_OF_MONTH)) + " " + nf.format(cal.get(Calendar.HOUR_OF_DAY)) + " " + nf.format(cal.get(Calendar.MINUTE)) + " " + nf.format(cal.get(Calendar.SECOND)) + " " + Double.toString(this.kp) + " " + Double.toString(this.status));

  }

}
