package saaps.dbt;

import java.util.Calendar;
import java.util.Date;
import java.util.Locale;
import java.util.TimeZone;
import java.util.Vector;
import java.text.DateFormat;
import java.text.NumberFormat;

import java.io.Serializable;
import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.lang.Integer;

/**
A class for LANL 1990-095 object.
<p>
The class contains the date and LANL data fields.
The following fields are included:
<ul>
  <li>Latitude</li>
  <li>Longitude east</li>
  <li>Radius (Re)</li>
  <li>E21.6-28.9 eV</li>
  <li>E87-114 eV</li>
  <li>E0.793-1.04 keV</li>
  <li>E9.23-12.14 keV</li>
  <li>E75-105 keV</li>
  <li>E0.75-1.1 MeV</li>
  <li>E7.8-10.8 MeV</li>
</ul>

@author Peter Wintoft
@version 1.0 2001-02-13
*/
public class LANL1990_095 extends DatabaseObject implements Serializable
{
  // Fields
  private static int numberOfIndexes = 10;
  private Date d;
  private double[] data = new double[numberOfIndexes];

  // Constructors
  public LANL1990_095()
  {
  }
  
  public LANL1990_095(Date D) {
    d = D;
    for (int i=0;i<data.length;i++)
      data[i] = Double.NaN;
  }
  
  public LANL1990_095(Date D,double[] DATA)
  {
    d = D;
    data = (double[]) DATA.clone();
  }
  public LANL1990_095(long L,double[] DATA)
  {
    d = new Date(L);
    data = (double[]) DATA.clone();
  }

  //Implemented for DatabaseObject getDate
  public int getNumberOfDataIndexes() {
    return (numberOfIndexes);
  }
  //Implemented for DatabaseObject getDate
  public Date getDate()
  {
    return d;
  }

  public Vector getData()
  {
    Vector vec = new Vector(numberOfIndexes);
    for (int i=0;i<numberOfIndexes;i++)
      vec.addElement(new Double(data[i]));
    return vec;
  }

  //Implemented for DatabaseObject getData
  public double getData(int index) {
    return data[index];
  }

  /**
   * This method returns the size of the data that is to be written to file
   */
  public long getDataSize() throws IOException {

    ByteArrayOutputStream baos = new ByteArrayOutputStream();  
    DataOutputStream    dos  = new DataOutputStream(baos);  

    dos.writeLong(d.getTime());
    for (int i=0;i<numberOfIndexes;i++)
      dos.writeDouble(data[i]);  

    dos.flush();  
    long size = baos.size();  // now you know how large ....
    return size;
  }

  public void setDate(Date D)
  {
    d = D;
  }
  public void setDate(long L)
  {
    d = new Date(L);
  }
  public void setData(double[] DATA)
  {
    data = (double []) DATA.clone();
  }
  public void setData(double DATA, int index)
  {
    data[index] = DATA;
  }

  public String toString() {
    // Find the selected dates.
    Calendar cal = Calendar.getInstance(TimeZone.getTimeZone("GMT"));
    cal.clear(cal.MILLISECOND);
    cal.setTime(d);

    NumberFormat nf = NumberFormat.getInstance();
    nf.setMinimumIntegerDigits(2);

    StringBuffer sb = new StringBuffer(cal.get(Calendar.YEAR) + " " + nf.format(cal.get(Calendar.MONTH)+1) + " " + nf.format(cal.get(Calendar.DAY_OF_MONTH)) + " " + nf.format(cal.get(Calendar.HOUR_OF_DAY)) + " " + nf.format(cal.get(Calendar.MINUTE)) + " " + nf.format(cal.get(Calendar.SECOND)));
    for (int i=0;i<numberOfIndexes;i++)
      sb.append(" " + Double.toString(data[i]));

    return ( sb.toString() );

  }

}
