package saaps.dbt;

import java.io.*;
import java.net.*;
import java.text.*;
import java.util.*;

import saaps.util.*;

/**
 * LatestHourlyACE downloads the ACE mag and swe data and stores it in the database.
 * FTP address:
 *   ftp://sec.noaa.gov/pub/lists/ace2/YYYYMM_ace_mag_1h.txt
 *   ftp://sec.noaa.gov/pub/lists/ace2/YYYYMM_ace_swepam_1h.txt
 *@author Leif Kalla
 *@version 1.0 2001-10-16
 */
public class LatestHourlyACE extends Object{
    
    private static String urlDir = "ftp://sec.noaa.gov/pub/lists/ace2/";
    private Calendar calendar = Calendar.getInstance(TimeZone.getTimeZone("GMT"));
    private SimpleDateFormat formatter;
    private Date now;
    
    public LatestHourlyACE() {
        now = new Date();
        formatter = (SimpleDateFormat)
        SimpleDateFormat.getDateTimeInstance(0,0,new Locale("sv","SE"));
        
        formatter.setTimeZone(TimeZone.getTimeZone("UTC"));
        formatter.applyPattern("yyyyMM");
    }

    public LatestHourlyACE(String y,String m) {
        Calendar cal = Calendar.getInstance(TimeZone.getTimeZone("GMT"));
        cal.clear(Calendar.MILLISECOND);
        cal.set(Integer.parseInt(y), Integer.parseInt(m)-1, 1, 0, 0, 0);
        now = cal.getTime();
        formatter = (SimpleDateFormat)
        SimpleDateFormat.getDateTimeInstance(0,0,new Locale("sv","SE"));
        
        formatter.setTimeZone(TimeZone.getTimeZone("UTC"));
        formatter.applyPattern("yyyyMM");
    }
    
    public String getLatestMagURL() {
        return (new String(urlDir+formatter.format(now)+"_ace_mag_1h.txt"));
    }

    public String getLatestSweURL() {
        return (new String(urlDir+formatter.format(now)+"_ace_swepam_1h.txt"));
    }

    public static void parseMagDataString(Vector data,String txt,Calendar cal) throws DatabaseException {
        StringTokenizer tokens = new StringTokenizer(txt);
        double dataArr[] = new double[4];
        
        int numTok = tokens.countTokens();
        if(numTok != 13) {
            throw new DatabaseException("Not correct number of tokens.");
        }
        int year = Integer.parseInt(tokens.nextToken());
        int month = Integer.parseInt(tokens.nextToken());
        int day = Integer.parseInt(tokens.nextToken());
        
        int hour = Integer.parseInt(((String)(tokens.nextToken())).substring(0,2));
        cal.set(year, month-1, day, hour,0, 0);
        for(int i=0;i<3;i++)
            tokens.nextToken();
        
        for(int i=0;i<4;i++) {
            double dtmp = Double.parseDouble(tokens.nextToken());
            if(dtmp == -999.9)
                dtmp = Double.NaN;
            dataArr[i]= dtmp;
        }
        data.add(new ACEmagHourly(cal.getTime(),dataArr));
    }
    
    public static void parseSweDataString(Vector data,String txt,Calendar cal) throws DatabaseException {
        StringTokenizer tokens = new StringTokenizer(txt);
        double dataArr[] = new double[2];
        
        int numTok = tokens.countTokens();
        if(numTok != 10) {
            throw new DatabaseException("Not correct number of tokens.");
        }
        int year = Integer.parseInt(tokens.nextToken());
        int month = Integer.parseInt(tokens.nextToken());
        int day = Integer.parseInt(tokens.nextToken());
        
        int hour = Integer.parseInt(((String)(tokens.nextToken())).substring(0,2));
        cal.set(year, month-1, day, hour,0, 0);
        for(int i=0;i<3;i++)
            tokens.nextToken();
        
        for(int i=0;i<2;i++) {
            double dtmp = Double.parseDouble(tokens.nextToken());
            if(dtmp == -9999.9)
                dtmp = Double.NaN;
            dataArr[i]= dtmp;
        }
        data.add(new ACEsweHourly(cal.getTime(),dataArr));
    }

    
    /**
     * Get the mag data.
     */
    public void getMagData(String urlStr) {
        Date d;
        try {
            Vector data = new Vector(24*31);
            URL url = new URL(urlStr);
            URLConnection urlConn = url.openConnection();
            // Input stream.
            BufferedReader streamIn = new BufferedReader(
            new InputStreamReader(
            urlConn.getInputStream()));
            //System.out.println("Got input stream");
            String textLine = "";
            
            /* Find the datatype position */
            textLine = streamIn.readLine();
            while ((textLine = streamIn.readLine()) != null) {
                if(!textLine.startsWith("#----")) 
                    continue;
                else
                    break;
            }

            Calendar cal = Calendar.getInstance();
            cal.setTimeZone(TimeZone.getTimeZone("UTC"));
            //cal.set(year,month-1,1,0,0,0);
            cal.clear(cal.MILLISECOND);

            while((textLine = streamIn.readLine()) != null) {
                if(textLine.startsWith("20"))
                    parseMagDataString(data,textLine,cal);
            }
            data.trimToSize();
            
            ObjectTimeSeries ts = new ObjectTimeSeries(data);
            Database db = Database.getInstance("ACEmagHourlyDB");
            db.store(ts);
            
            streamIn.close();
        }
        catch (Exception e) {
            System.err.println("error: " + e);
            e.printStackTrace();
        }        
    }

    /**
     * Get the swe data.
     */
    public void getSweData(String urlStr) {
        Date d;
        try {
            Vector data = new Vector(24*31);
            URL url = new URL(urlStr);
            URLConnection urlConn = url.openConnection();
            // Input stream.
            BufferedReader streamIn = new BufferedReader(
            new InputStreamReader(
            urlConn.getInputStream()));
            //System.out.println("Got input stream");
            String textLine = "";
            
            /* Find the datatype position */
            textLine = streamIn.readLine();
            while ((textLine = streamIn.readLine()) != null) {
                if(!textLine.startsWith("#----")) 
                    continue;
                else
                    break;
            }

            Calendar cal = Calendar.getInstance();
            cal.setTimeZone(TimeZone.getTimeZone("UTC"));
            //cal.set(year,month-1,1,0,0,0);
            cal.clear(cal.MILLISECOND);

            while((textLine = streamIn.readLine()) != null) {
                if(textLine.startsWith("20"))
                    parseSweDataString(data,textLine,cal);
            }
            data.trimToSize();
            
            ObjectTimeSeries ts = new ObjectTimeSeries(data);
            Database db = Database.getInstance("ACEsweHourlyDB");
            db.store(ts);
            
            streamIn.close();
        }
        catch (Exception e) {
            System.err.println("error: " + e);
            e.printStackTrace();
        }        
    }

    public static void main(String[] args) {
        LatestHourlyACE ace = null;
        
        if((args.length > 0 ) && (args.length != 2)) {
            System.out.println("Usage: LatestHourlyACE or LatestHourlyACE YYYY MM");
            System.exit(0);
        }
        if(args.length == 2)
            ace = new LatestHourlyACE(args[0],args[1]);
        else
            ace = new LatestHourlyACE();
        
        System.out.println("Getting remote data...");
                
        ace.getMagData(ace.getLatestMagURL());
        ace.getSweData(ace.getLatestSweURL());
    }
}