package saaps.dbt;

import java.io.*;
import java.text.DateFormat;
import java.text.FieldPosition;
import java.util.*;

/**
 * <code>MetaData</code> is a class that provides methods to read
 * and write the meta data from or to the meta data file.
 * Each <code>Database</code> class has a meta data file.
 * <p>
 * <b>Note:</b> The set methods doesn't update the stored 
 * metadata, modifications are only made on the current object.
 * When all parameters are set you may create a new <code>MetaData</code>
 * object whith the modified <code>MetaData</code> object as input and
 * the new meta data will be saved to file.
 *
 * @author Leif Kalla
 * @author Peter Wintoft
 */
public class MetaData implements Serializable {

  /** Default path to the database. */
  private String baseDir = "/space/saaps/Database/";

  /* Private fields, no sense storing these thus transient */
  private transient ObjectInputStream in=null;
  private transient ObjectOutputStream out=null;

  /* These are our metadata, store them with Objectwriter */
  /** The class name. */
  private String cname=null;
  /** Short description of the data. */
  private String descr=null;
  /** Long description of the data. */
  private String descrLong=null;
  /** Time of the first data point. */
  private Calendar startTime;
  /** Time of the last data point. */
  private Calendar endTime;
  /** True if data is equidistant. */
  private boolean cTimeSteps=false;
  /** Time step [milliseconds] */
  private long timeStep=0;
  /** Name of each data field. */
  private String[] fieldNames=null;
  /** Unit of each data field. */
  private String[] fieldUnits=null;
  /** Type of each data field. */
  private String[] fieldTypes =null;
  ;
  /** Hashtable over the <code>fieldName</code>s. */
  private Hashtable selHash = null;

  /**
   * Set's the database location from the property selected by
   * <code>saaps.dbt.Database.dataDirVar</code>.
   */
  public MetaData() {
    baseDir=System.getProperty(saaps.dbt.Database.dataDirVar,baseDir);
    //Make sure that it ends with a "/".
    if(!this.baseDir.endsWith("/")) {
        this.baseDir += "/";
    }
  }

  /**
   * Loads meta data from file.
   * @param cn <code>Database</code> class name
   * @exception IOException when problem reading data from file
   * @exception ClassNotFoundException when <code>Database</code>
   * class not found
   */
  public MetaData(String cn) throws IOException ,ClassNotFoundException  {
    this();
    in = new ObjectInputStream(new FileInputStream(baseDir + cn + ".meta"));
    /* Create a meta object and fill in our fields for this object ..
     * This shoulde be possible to do easier ... this = ..
     */
    MetaData meta = new MetaData();
    meta = (MetaData) in.readObject();
    in.close();

    cname = meta.cname;
    descr = meta.descr;
    descrLong = meta.descrLong;
    startTime = meta.startTime;
    endTime = meta.endTime;
    cTimeSteps = meta.cTimeSteps;
    timeStep = meta.timeStep;
    fieldNames = meta.fieldNames;
    fieldUnits = meta.fieldUnits;
    fieldTypes = meta.fieldTypes;
    selHash = meta.selHash;
  }

  /**
   * Stores meta data to file.
   * @param cn <code>Database</code> class name
   * @param des short description
   * @param des2 long description
   * @param st date/time of first item in database
   * @param et date/time of last item in database
   * @param ct true if equidistant data
   * @param step time step in milliseconds
   * @param fn array of field names
   * @param ft array of field types
   */
  public MetaData(String cn, String des,String des2, Calendar st, Calendar et, boolean ct, long step, String[] fn, String[] fu,String[] ft) {

    this();
    cname = cn;
    descr = des;
    descrLong = des2;
    startTime = st;
    endTime = et;
    cTimeSteps = ct;
    timeStep = step;
    fieldNames = fn;
    fieldUnits = fu;
    fieldTypes = ft;
    selHash = new Hashtable(fieldNames.length);

    /* Setup the HashTable */
    for(int i=0;i<fieldNames.length;i++) {
      selHash.put(fieldNames[i],new Integer(i));
    }

    try {
      out = new ObjectOutputStream(new FileOutputStream(baseDir + cn + ".meta"));
      out.writeObject(this);// is this possible ?
      out.close();
    }
    catch (IOException e) {
      System.err.println(e);
    }
  }

  /**
   * Run constructor with default time step -1.
   * @param cn <code>Database</code> class name
   * @param des short description
   * @param des2 long description
   * @param st date/time of first item in database
   * @param et date/time of last item in database
   * @param ct true if equidistant data
   * @param fn array of field names
   * @param ft array of field types
   */
  public MetaData(String cn, String des,String des2, Calendar st, Calendar et, boolean ct,String[] fn, String[] fu,String[] ft) {
    this(cn,des,des2,st,et,ct,-1,fn,fu,ft);
  }

  /**
   * Run constructor with all arguments extracted from the 
   * input <code>MetaData</code> object.
   * Stores meta data to file.
   * @param m meta data
   */
  public MetaData(MetaData m) {
    this(m.cname,m.descr,m.descrLong,m.startTime,m.endTime,m.cTimeSteps,m.timeStep,m.fieldNames,m.fieldUnits,m.fieldTypes);
  }

  /**
   * Gets the class name.
   * @return class name
   */
  public String getName() {
    return cname;
  }

  /**
   * Gets the short description.
   * @return short description
   */
  public String getDescr() {
    return descr;
  }

  /**
   * Gets the long description.
   * @return long description
   */
  public String getDescrLong() {
    return descrLong;
  }

  /**
   * Gets the start date.
   * @return start date
   */
  public Date getStartDate() {
    return startTime.getTime();
  }

  /**
   * Gets the start date in <code>Calendar</code> object.
   * @return start date (Calendar)
   */
  public Calendar getStartCalendar() {
    return startTime;
  }

  /**
   * Gets the end date.
   * @return end date
   */
  public Date getEndDate() {
    return endTime.getTime();
  }

  /**
   * Gets the end date in <code>Calendar</code> object.
   * @return end date (Calendar)
   */
  public Calendar getEndCalendar() {
    return endTime;
  }

  /**
   * @return <code>true</code> if <code>Database</code> has 
   * equidistant data
   */
  public boolean getCont() {
    return cTimeSteps;
  }

  /**
   * @return time step of data
   */
  public long getTimeStep() {
    return timeStep;
  }

  /**
   * @return field names
   */
  public String[] getFieldNames() {
    return fieldNames;
  }

  /**
   * @return field units
   */
  public String[] getFieldUnits() {
    return fieldUnits;
  }

  /**
   * @return field types
   */
  public String[] getFieldTypes() {
    return fieldTypes;
  }

  /**
   * Gets the index of the input field name.
   * @param str data field name
   * @return the index of this field
   */
  public int getIndex(String str) {
    return (((Integer) selHash.get(str)).intValue());
  }
  
  /**
   * Sets the start date.
   * @param cal date/time of first item in the <code>Database</code>
   */
  public void setStartDate(Calendar cal) {
    startTime=cal;
  }

  /**
   * Sets the start date.
   * @param d start date/time of first item in <code>Database</code>
   */
  public void setStartDate(Date d) {

    Calendar calStart = Calendar.getInstance(TimeZone.getTimeZone("UTC"),new Locale("sv","SE"));
    calStart.setTime(d);
    startTime=calStart;
  }

  /**
   * Sets the end date.
   * @param cal end date/time of last item in <code>Database</code>
   */
  public void setEndDate(Calendar cal) {
    endTime=cal;
  }

  /**
   * Sets the end date.
   * @param d end date/time of last item in <code>Database</code>
   */
  public void setEndDate(Date d) {
    Calendar cal = Calendar.getInstance(TimeZone.getTimeZone("UTC"),new Locale("sv","SE"));
    cal.setTime(d);
    endTime=cal;
  }

  /**
   * Sets the start and end date/times.
   * @param sd date/time of first item in <code>Database</code>
   * @param ed date/time of last item in <code>Database</code>
   */
  public void setDates(Calendar sd, Calendar ed) {
    setStartDate(sd);
    setEndDate(ed);
  }

  /**
   * Sets the start and end date/times.
   * @param sd date/time of first item in <code>Database</code>
   * @param ed date/time of last item in <code>Database</code>
   */
  public void setDates(Date sd, Date ed) {
    setStartDate(sd);
    setEndDate(ed);
  }

  /**
   * Sets all fields in <code>MetaData</code>.
   * @param cn <code>Database</code> class name
   * @param des short description
   * @param des2 long description
   * @param st date/time of first item in database
   * @param et date/time of last item in database
   * @param ct true if equidistant data
   * @param step time step in milliseconds
   * @param fn array of field names
   * @param ft array of field types
   */
  public void setMetaData(String cn, String des,String des2, Calendar st, Calendar et, boolean ct, long step, String[] fn, String[] fu,String[] ft) {

    cname = cn;
    descr = des;
    descrLong = des2;
    startTime = st;
    endTime = et;
    cTimeSteps = ct;
    timeStep = step;
    fieldNames = fn;
    fieldUnits = fu;
    fieldTypes = ft;

    /* Setup the HashTable */
    selHash = new Hashtable(fieldNames.length);

    for(int i=0;i<fieldNames.length;i++) {
      selHash.put(fieldNames[i],new Integer(i));
    }

  }

  /**
   * Public version of store.
   * See {@link #store}
   */
  public void flush() {
    store();
  }

  /**
   * Stores the current <code>MetaData</code> object to file.
   */
  private void store() {
    if(cname!=null) {
      try {
	out = new ObjectOutputStream(new FileOutputStream(baseDir + cname + ".meta"));
	out.writeObject(this);
	out.close();
      }
      catch (IOException e) {
	System.err.println(e);
      }
    }
  }
}
