package saaps.dbt;

import java.io.*;
import java.util.*;

/**
 * OMNIDB extends the Database class. It contains
 * the read and write methods that are used in
 * Database. The field is a vector of OMNI objects.
 *
 *@author Peter Wintoft
 *@version 1.0 2000-04-14
 *@see OMNI
 */
public class OMNIDB extends Database {
  // Absolute path to the database.
  //  private String baseDir = "/space/saaps/Database/";

  private static String cName = "OMNI";

  private static final double dataGapValueDefinition = 999.9;
  private static double[] dataGapValues =
    //F      B      Bx     ByGSE  BzGSE  ByGSM  BzGSM
    { 999.9, 999.9, 999.9, 999.9, 999.9, 999.9, 999.9,
    //sigmaF sigmaB sigmaBx sigmaBy sigmaBz
      999.9, 999.9, 999.9,  999.9,  999.9,
    //T          N      V       sigmaT     sigmaN sigmaV
      9999990.0, 999.9, 9999.0, 9999990.0, 999.9, 9999.0,
    //Dst    P1        P2        P4       P10      P30      P60
      99999, 99999.90, 99999.90, 9999.99, 9999.99, 9999.99, 9999.99};

  // Static Fields used to initiate the MetaData
  private static String name= cName + "DB";
  private static String description = "OMNI data";
  private static String longDescription = "The OMNI data set contain several different " +
    "parameters with a one hour resolution. This data set is only a subset of the full OMNI set. " +
    "This data set contain the solar wind magnetic field data " + 
    "(F, B, Bx, ByGSE, BzGSE, ByGSM, BzGSM, sigmaF, sigmaB, sigmaBx, sigmaBy, sigmaBz), " +
    "solar wind plasma data (T, N, V, sigmaT, sigmaN, sigmaV), " +
    "the geomagnetic ring current index Dst, " +
    "and solar wind protons at 6 levels: >1MeV, >2MeV, >4MeV, >10MeV, >30MeV, >60MeV. ";
  private static Calendar st=Calendar.getInstance(TimeZone.getTimeZone("UTC"),new Locale("sv","SE"));
  private static Calendar et=Calendar.getInstance(TimeZone.getTimeZone("UTC"),new Locale("sv","SE"));
  private static boolean ct=true;
  private static long step=60*60*1000; //One hour timestep
  private static String[] fieldNames =
    { "F", "B", "Bx", "ByGSE", "BzGSE", "ByGSM", "BzGSM",
      "sigmaF", "sigmaB", "sigmaBx", "sigmaBy", "sigmaBz",
      "T", "N", "V", "sigmaT", "sigmaN", "sigmaV", "Dst",
      "P>1 MeV", "P>2 MeV", "P>4 MeV", "P>10 MeV", "P>30 MeV", "P>60 MeV"};
  private static String[] fieldUnits =
    { "nT", "nT", "nT", "nT", "nT", "nT", "nT",
      "nT", "nT", "nT", "nT", "nT",
      "K", "1/cm^3", "km/s", "K", "1/cm^3", "km/s", "nT",
      "1/(cm^2 s sr)", "1/(cm^2 s sr)", "1/(cm^2 s sr)", "1/(cm^2 s sr)", "1/(cm^2 s sr)", "1/(cm^2 s sr)"};
  private static String[] fieldTypes =
    { "Double", "Double", "Double", "Double", "Double", "Double", "Double",
      "Double", "Double", "Double", "Double", "Double",
      "Double", "Double", "Double", "Double", "Double", "Double", "Double",
      "Double", "Double", "Double", "Double", "Double", "Double"};
  private static int numOfFields = fieldNames.length;

  // Fields.
  private OMNI data;
  private Vector vec;

  // Constructors.
  public OMNIDB() {
    
  }
  public OMNIDB(int i) {
    
    //    acev = ACEV;
    //store(acev);
  }
  /**
   * When a new class is added to the database this
   * constructor should be used which includes the
   * information that goes into the meta data file.
   *
   * The arguments are: Vector of OMNI objects, the class name,
   * a short description, continuos time series (false/true), and
   * if a continous time series the time step.
   */
  public OMNIDB(Vector VEC) throws IOException {
    //, String des, boolean ct, long step) {
    //
    
    vec = VEC;
    raf = new RandomAccessFile(baseDir + cName + "DB.data","rw");
    // Then get its metadata
    try { // To load the meta data for this database from file
      meta = new MetaData(cName + "DB");
      isSet=true;
    }
    catch (IOException e) {
      System.err.println(e);
      System.out.println("Here we should handle the situation when no DB exists");
      try {
  meta = setDefaultMeta(((DatabaseObject) (vec.elementAt(0))).getDate(),((DatabaseObject) (vec.elementAt(vec.size()-1))).getDate()); //Takes static fields from XDB

  isSet=true;
      }
      catch (IOException ee) {
  System.err.println(ee);
  ee.printStackTrace();
      }
    }
    catch (ClassNotFoundException e) {
      System.err.println(e);
      e.printStackTrace();
    }

    store((Vector) vec); //Need to type cast so that store(Object obj) isn't used
  }

  // Methods.
  public Object read() throws IOException {
    long t = raf.readLong();
//    int numOfFields = OMNI.getNumberOfDataIndexes()
    double[] d = new double[numOfFields];
    for (int i=0;i<numOfFields;i++)
      d[i] = raf.readDouble();
    data = new OMNI(t,d);
//    data = new OMNI(raf.readLong(),
//                          raf.readDouble(),raf.readDouble(),raf.readDouble(),raf.readDouble(),raf.readDouble(),
//                          raf.readDouble(),raf.readDouble(),raf.readDouble(),raf.readDouble(),raf.readDouble());
    return (Object) data;
  }

  public void myHello() {
    System.out.println("My" + cName + "DB Hello");
  }

  public void write(Object obj) throws IOException {
    data = (OMNI) obj;
    raf.writeLong(data.getDate().getTime());
    for (int i=0;i<numOfFields;i++)
      raf.writeDouble(data.getData(i));
  }
  // Implemented for Database abstract getDate
  public Date getDate(Object obj) {
    return ((OMNI) obj).getDate();
  }

  /** This method is implemented for Database abstract metod.
   * It retrieves the latest data from the dataprovider and returns the vector of Database objects.
   */
  public Vector getLatestDatabaseObjects() {
    Vector vec = new Vector(24);
    Date mdate = new Date();
    double[] x = new double[numOfFields];
    for (int i=0;i<numOfFields;i++)
      x[i] = (double) i+1;

    OMNI omni = new OMNI(mdate,x);

    for(int i=0;i<8*30;i++) {
      vec.addElement(omni);
      mdate.setDate( (int) (mdate.getDate() + 10));
      omni.setDate(mdate);
      omni.setData(x);
    }
    vec.trimToSize();
    return vec;
  }

  /** Creates a metadata file for this class with the default values set
   *to the static values defined above. OBS!, check calendar ....
   */
  protected MetaData setDefaultMeta(Date sd, Date ed) throws IOException {

    st.setTime(sd);
    st.clear(st.MILLISECOND);
    et.setTime(ed);
    et.clear(et.MILLISECOND);

    MetaData m = new MetaData(name,description, longDescription, st, et, ct, step,fieldNames,fieldUnits,fieldTypes);
    return m;
  }

  /**
   * Returns an "empty" metadata object. All fields but Start and end time
   * has valid values (as set in the respective database class).
   * The times will be set to that what is supplied.
   * OBS! This is not stored in the MetaData file !!
   */
  protected MetaData getEmptyMetaData(Date sd, Date ed) {
    MetaData m = new MetaData();
    st.clear(st.MILLISECOND);
    st.setTime(sd);
    et.clear(et.MILLISECOND);
    et.setTime(ed);
    m.setMetaData(name,description, longDescription, st, et, ct, step, fieldNames, fieldUnits, fieldTypes );
    return m;
  }

}
