/*
 * GOES08Model.java
 *
 * Created on September 12, 2000, 10:10 AM
 */
 
package saaps.sapm;

import java.io.IOException;
import java.util.Calendar;
import java.util.Date;
import java.util.TimeZone;
import java.util.Vector;
import saaps.ann.*;
import saaps.ann.mlff.*;
import saaps.dbt.Database;
import saaps.util.TimeSeriesPlot;
import saaps.util2.TimeSeries;

/** 
 *
 * @author  saaps
 * @version 
 */
public class GOES08Model extends Object {

  // An environment property to define the path to the models.
  public static String modelDirVar="saaps.models";
  
  //The neural network.
  private Network net;
  
  //The input data.
  private Data input;
  
  //The output data.
  private Data output;
  
  //The input time series.
  private TimeSeries inputTS;
  
  //The output time series.
  private TimeSeries outputTS;
  
  //A calendar object for time computations.
  private Calendar calendar;
  
  //The longitude in degress east of the GOES-08 satellite.
  //The longitude can be set to othher values.
  private double longitude = 360.0 - 76.0;

  /** Creates new GOES08Model */
  public GOES08Model() {
    //Create the network.
    createNetwork();
    //Initialize the calendar object to be used for time compuations.
    initCalendar();
  }
  
  private void createNetwork()
  {
/*    //Get the place of the models.
    String dir = System.getProperty(modelDirVar);
    //Check that dir ends with an "/".
    if (!dir.endsWith("/"))
      dir = new String(dir + "/");
    */
    String dir = "/space/saaps/Models/";
    //Input-to-hidden weights.
    Weights[] w1 = WeightsCreator.read(dir + "dailyGOES08e2layer1.txt");
    //Hidden-to-output weights.
    Weights[] w2 = WeightsCreator.read(dir + "dailyGOES08e2layer2.txt");
    //Initialize layers.
    Layer[] l = new Layer[2];
    //Create layer 1.
    Neuron[] n = new Neuron[w1.length];
    for (int i=0;i<n.length;i++)
      n[i] = new TanhNeuron(w1[i]);
    l[0] = new Layer(n);
    //Create layer 2.
    n = new Neuron[w2.length];
    for (int i=0;i<n.length;i++)
      n[i] = new LinNeuron(w2[i]);
    l[1] = new Layer(n);
    //Create network.
    net = new Network(l);
  }
  
  private void initCalendar()
  {
    calendar = Calendar.getInstance(TimeZone.getTimeZone("GMT"));
  }
  
  public void getInputData(Date d1, Date d2)
  {
    //The time lag in days.
    int timeLag = 8;
    //The number of input parameters: n and V.
    int numOfInputParam = 2;
    //The number of inputs including the cosine and sine of local time.
    int numOfInputs = timeLag*numOfInputParam + 2;
    //Go back the time lag.
    calendar.clear();
    calendar.setTime(d1);
    calendar.add(Calendar.DATE,-timeLag);
    Date dd1 = calendar.getTime();
    //Get the time series.
    inputTS = getTimeSeries(dd1,d2);
    //Interpolate over the datagaps.
    inputTS.interpolate();
    //Resample to daily averages.
    inputTS.resample(0,1,0,23,TimeSeries.HOUR);
    //Normalize
    inputTS.setNormParam(0,0.0,40.0,-0.8,+0.8);
    inputTS.setNormParam(1,250.0,750.0,-0.8,+0.8);
    inputTS.normalize();
    //Add the local time fields.
    double[] lt = localTime(inputTS.getDates());
    inputTS.addField(cosTime(lt));
    inputTS.addField(sinTime(lt));
    //Create the input data.
    //n(t),n(t+1),...,n(t+7),v(t),...,v(t+7),cos(LT(t+7)),sin(LT(t+7))
    int[] time = {0,1,2,3,4,5,6,7,0,1,2,3,4,5,6,7,7,7};
    int[] field = {0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,2,3};
    TimeSeriesToDataMap map = new TimeSeriesToDataMap(time,field);
    input = map.convert(inputTS);
    net.setInput(input);
  }
  
  public Data getOutputData()
  {
    output = net.getOutput();
    return output;
  }
  
  public TimeSeries getOutput()
  {
    output = net.getOutput();
    Date[] d = inputTS.getDates();
    Date[] nd = new Date[output.getNumOfExamples()];
    for (int i=0;i<nd.length;i++)
      nd[i] = new Date(d[i+d.length-nd.length].getTime());
    outputTS = new TimeSeries(nd,output);
    outputTS.setNormParam(0,-0.8,+0.8,-1,+5);
    outputTS.pow(10.0,0);
    return outputTS;
  }
    
  
  private TimeSeries getTimeSeries(Date d1, Date d2)
  {
    Database db = Database.getInstance("ACEsweHourlyDB");
    Vector vec = null;
    try
    {
      vec = db.retreive(d1,d2);
    }
    catch (IOException ioe)
    {
      System.err.println(ioe);
    }
    return new TimeSeries(vec);
  }
  
  private double[] localTime(Date[] d)
  {
    double[] lt = new double[d.length];
    Calendar c = Calendar.getInstance(TimeZone.getTimeZone("GMT"));
    c.clear();
    for (int i=0;i<lt.length;i++)
    {
      c.setTime(d[i]);
      lt[i] = c.get(Calendar.HOUR_OF_DAY) + 
              c.get(Calendar.MINUTE)/60.0 +
              c.get(Calendar.SECOND)/3600.0 + 
              longitude/360.0*24.0;
      if (lt[i]>24.0)
        lt[i] -= 24.0;
    }
    return lt;
  }
  
  private double[] cosTime(double[] lt)
  {
    double[] c = new double[lt.length];
    for (int i=0;i<c.length;i++)
      c[i] = 0.8*Math.cos(lt[i]/12.0*Math.PI);
    return c;
  }
  
  private double[] sinTime(double[] lt)
  {
    double[] s = new double[lt.length];
    for (int i=0;i<s.length;i++)
      s[i] = 0.8*Math.sin(lt[i]/12.0*Math.PI);
    return s;
  }
  
  /**
  * @param args the command line arguments
  */
  public static void main (String args[]) {
    GOES08Model model = new GOES08Model();
    Calendar c = Calendar.getInstance(TimeZone.getTimeZone("GMT"));
    c.clear();
    c.set(2000,9-1,1,0,0,0);
    Date d1 = c.getTime();
    c.set(2000,9-1,10,0,0,0);
    Date d2 = c.getTime();
    model.getInputData(d1,d2);
//    model.input.print();
    Data output = model.getOutputData();
    TimeSeries outputTS = model.getOutput();
//    outputTS.print();
    //Plot of input data.
    saaps.util.TimeSeriesPlot tsPlotIn = new saaps.util.TimeSeriesPlot(model.inputTS.toOldTimeSeries());
    tsPlotIn.plotData();
    new GOES08ModelPlot(tsPlotIn);
    //Plot of output data.
    saaps.util.TimeSeriesPlot tsPlotOut = new saaps.util.TimeSeriesPlot(outputTS.toOldTimeSeries());
    tsPlotOut.setYLog(true);
    tsPlotOut.plotData();
    new GOES08ModelPlot(tsPlotOut);
  }

}