package saaps.util;

/**
 * The TimeSeries class contain n-dimensional time series
 * data. The data are ordered in time. The names and units
 * of the data fields are also stored in the <code>TimeSeries</code>
 * object.
 * <p>
 * Currently it has methods to extract daily sums of the
 * data.
 *
 * @author Peter Wintoft
 * @version 2000-01-20
 * @version 2000-08-04 (Added name and unit fields).
 * @see    saaps.util.TimeSeriesPlot
 */

import saaps.ann.NormParam;
import saaps.dbt.DatabaseObject;
import saaps.dbt.Database;
import java.util.Calendar;
import java.util.Date;
import java.util.TimeZone;
import java.util.Vector;
import java.util.Locale;
import java.io.Serializable;
import java.text.NumberFormat;

import Jama.Matrix;


public class TimeSeries implements Serializable
{
    // dates contain the dates for each observation
    private Date[] dates;
    // data contain the data.
    private Matrix data;
    // the number of data.
    private int numOfData;
    // the number of fields.
    private int numOfFields;
    // the number of data fields.
    private int numOfDataFields;
    // Time zone = UT.
    private TimeZone tz = TimeZone.getTimeZone("GMT");
    private Calendar cal = Calendar.getInstance(tz);
    // A short description of the data.
    private String shortDescription;
    // A long description of the data.
    private String longDescription;
    // The field names.
    private String[] fieldNames;
    // The list name.
    private String[] listNames;
    
    // The field units.
    private String[] fieldUnits;
    public static final int aveArrDptOffset = 1;
    public static final int aveArrStdvOffset = 2;
  /**
   * Describes how many different types of data is in the average
   *  array. ex. data,number of data points, standard deviation
   */
    public static final int aveArrTypes = 3;
    
  /**
   * Create a <code>TimeSeries</code> object from a Date vector and a double vector.
   *
   * @param d The dates.
   * @param m The data.
   */
    public TimeSeries(Date[] d, double[] m)
    {
        // Get the number of data and the number of data fields.
        numOfData = d.length;
        numOfFields = 2;
        numOfDataFields = numOfFields - 1;
        dates = d;
        data = new Matrix(m,m.length);
        //Set the default meta data.
        setDefaultMeta();
    }
    
  /**
   * Create a <code>TimeSeries</code> object from a Date vector and a double vector,
   * and specify the data field names and units.
   *
   * @param  d The dates.
   * @param  m The data.
   * @param  fn  The field name.
   * @param  fu  The field unit.
   */
    public TimeSeries(Date[] d, double[] m, String fn, String fu)
    {
        // Get the number of data and the number of data fields.
        numOfData = d.length;
        numOfFields = 2;
        numOfDataFields = numOfFields - 1;
        dates = d;
        data = new Matrix(m,m.length);
        fieldNames = new String[1];
        fieldUnits = new String[1];
        fieldNames[0] = fn;
        fieldUnits[0] = fu;
    }
    
  /**
   * Create the object from date array and the 2-D array.
   * Set also the field names.
   * <p>
   * <b>Note:</b> All 2-D array rows must have the same length.
   * @param d date array
   * @param m 2-D array of double data (all rows same length)
   * @param fn field names
   * @param fu field units
   */
    public TimeSeries(Date[] d, double[][] m, String[] fn, String[] fu) {
        numOfData = d.length;
        numOfFields = m[0].length+1;
        // java.lang.reflect.Array.getLength(m[0]);
        numOfDataFields = numOfFields - 1;
        dates = d;
        data = new Matrix(m);
        fieldNames = fn;
        fieldUnits = fu;
        listNames = null;
    }

  /**
   * @param fu field units
   */
    public TimeSeries(Date[] d, double[][] m, String[] fn, String[] fu, String[] ln) {
        numOfData = d.length;
        numOfFields = m[0].length+1;
        // java.lang.reflect.Array.getLength(m[0]);
        numOfDataFields = numOfFields - 1;
        dates = d;
        data = new Matrix(m);
        fieldNames = fn;
        fieldUnits = fu;
        listNames = ln;
    }

    
  /**
   * Create a TimeSeries object from a Vector of DatabaseObject:s
   *
   * @param dbov The DatabaseObject vector
   */
    public TimeSeries(Vector dbov)
    {
        
        // Get the number of data and the number of data fields.
        //Assume that the vector contains same objects all the way
        //which should be tre !!!
        numOfData = dbov.size();
        numOfFields = ((DatabaseObject) dbov.elementAt(0)).getNumberOfDataIndexes() +1;
        numOfDataFields = numOfFields - 1;
        
        //Create a matrix and fill it, fill the vector
        data = new Matrix(numOfData,numOfDataFields);
        dates = new Date[numOfData];
        
        for(int i=0;i<numOfData;i++) {
            dates[i]= ((DatabaseObject) dbov.elementAt(i)).getDate();
            for(int j=0;j<numOfDataFields;j++) {
                data.set(i,j,((DatabaseObject) dbov.elementAt(i)).getData(j));
            }
        }
        //Set the default meta data.
        setDefaultMeta();
    }
    
  /**
   * Create a TimeSeries object from a Vector of DatabaseObject:s
   *
   * @param dbov The DatabaseObject vector
   * @param fn Field names
   * @param fu Field units
   */
    public TimeSeries(Vector dbov,String[] fn,String[] fu) {
        this(dbov);
        fieldNames=fn;
        fieldUnits=fu;
    }
    
  /**
   * Create a TimeSeries object from an ObjectTimeSeries.
   *
   * @param ots The ObjectTimeSeries
   * @param fn Field names
   * @param fu Field units
   */
    public TimeSeries(ObjectTimeSeries ots,String[] fn,String[] fu) {
        
        fieldNames=fn;
        fieldUnits=fu;
        // Get the number of data and the number of data fields.
        //Assume that the vector contains same objects all the way
        //which should be tre !!!
        numOfData = ots.getLength();
        numOfDataFields = ots.getNumOfDataFields();
        numOfFields = numOfDataFields + 1;
        
        
        //Create a matrix and fill it, fill the vector
        data = new Matrix(numOfData,numOfDataFields);
        dates = new Date[numOfData];
        
        for(int i=0;i<numOfData;i++) {
            DatabaseObject tmp = ots.getDataObject(i);
            dates[i]= tmp.getDate();
            for(int j=0;j<numOfDataFields;j++) {
                data.set(i,j,tmp.getData(j));
            }
        }
    }
    
  /**
   * Create a TimeSeries object from a Vector of DatabaseObjects
   * Use only the indicated data field.
   * @param dbov The DatabaseObject vector
   * @param fieldIndex field number in <code>DatabaseObject</code> data
   */
    public TimeSeries(Vector dbov,int fieldIndex) throws TimeSeriesIndexException {
    /* Get the number of data and the number of data fields.
     * Assume that the vector contains same objects all the way
     * which should be tre !!!
     */
        numOfData = dbov.size();
        numOfFields = 2;
        numOfDataFields = numOfFields - 1;
        
        if(fieldIndex >= ((DatabaseObject) dbov.elementAt(0)).
        getNumberOfDataIndexes())
            throw (new TimeSeriesIndexException("index:" +
            fieldIndex + " out of bounds"));
        
        //Create a matrix and fill it, fill the vector
        data = new Matrix(numOfData,1);
        dates = new Date[numOfData];
    /* fieldIndex starts at 0 */
        for(int i=0;i<numOfData;i++) {
            dates[i]= ((DatabaseObject) dbov.elementAt(i)).getDate();
            data.set(i,0,((DatabaseObject) dbov.elementAt(i)).getData(fieldIndex));
        }
        //Set the default meta data.
        setDefaultMeta();
    }
    
  /**
   * Create a TimeSeries object from an array of Date objects
   * and a Matrix. The number of elements in the Date array must
   * equal the number of rows in the Matrix object.
   *
   * @param d Array of Date object
   * @param m Matrix of data.
   */
    public TimeSeries(Date[] d, Matrix m)
    {
        dates = d;
        data = m;
        numOfData = dates.length;
        numOfDataFields = data.getColumnDimension();
        numOfFields = numOfDataFields + 1;
        //Set the default meta data.
        setDefaultMeta();
    }
    
  /**
   * Create an empty time series object.
   */
    public TimeSeries(long st, long et, long si, int nodf)
    {
        //Ensure that there are integer steps from st to et.
        numOfData = (int) ((et-st)/si);
        et = st + ((long) numOfData)*si;
        
        //Initiate the time field.
        dates = new Date[numOfData];
        for (int i=0;i<numOfData;i++)
        {
            long t = (long) st + ((long) i)*si;
            dates[i] = new Date(t);
        }
        
        //Initiate the data fields.
        numOfDataFields = nodf;
        numOfFields = nodf + 1;
        data = new Matrix(numOfData,numOfDataFields);
        //Set the default meta data.
        setDefaultMeta();
    }
    
  /**
   * Set the default meta data for the time series.
   */
    private void setDefaultMeta()
    {
        shortDescription = "Unknown";
        longDescription = "Unknown";
        fieldNames = new String[numOfDataFields];
        fieldUnits = new String[numOfDataFields];
        for (int i=0;i<numOfDataFields;i++)
        {
            fieldNames[i] = Integer.toString(i);
            fieldUnits[i] = "";
        }
    }
    
  /**
   * Sets the field names array.
   * @param field names
   */
    public void setFieldNames(String[] fn) {
        fieldNames=fn;
    }
    
  /**
   * Sets the field units array.
   * @param field units
   */
    public void setFieldUnits(String[] fu) {
        fieldUnits=fu;
    }
    
  /*
  A method to set a Calendar object from a Date object
  and setting the hour, minute, and second fields all to
  zero.
   
  @param d the date
   */
    private Calendar setCalDay(Date d)
    {
        Calendar cal = Calendar.getInstance(tz);
        cal.setTime(d);
        cal.set(cal.HOUR,0);
        cal.set(cal.MINUTE,0);
        cal.set(cal.SECOND,0);
        cal.set(cal.MILLISECOND,0);
        return cal;
    }
    
  /**
   * Get the daily sum for each parameter for each day.
   * <p>
   * Future version should include a check that all days
   * contain the same number of data.
   */
    public TimeSeries dailySum()
    {
        // Get the number of days.
        System.out.println("TimeSeries.dailySum()");
        int numOfDays =(int) ( ( dates[dates.length-1].getTime() - dates[0].getTime() ) / 1000 / 3600 / 24 + 1 ) ;
        System.out.println("numOfDays=" + numOfDays);
        Date[] sumDates = new Date[numOfDays];
        Matrix sumData = new Matrix(numOfDays,numOfDataFields);
        // Get the day for the first day.
        cal = setCalDay(dates[0]);
        int d = cal.get(cal.DATE);
        // Go through the data.
        int dayIndex = 0;
        sumDates[dayIndex] = cal.getTime();
        Matrix tmp = new Matrix(1,numOfDataFields);
        for (int i=0;i<numOfData;i++)
        {
            cal = setCalDay(dates[i]);
            if (cal.get(cal.DATE) == d)
            {
                tmp = tmp.plus(data.getMatrix(i,i,0,numOfDataFields-1));
            }
            else
            {
                //        sumDates[dayIndex] = cal.getTime();
                sumData.setMatrix(dayIndex,dayIndex,0,numOfDataFields-1,tmp);
                dayIndex++;
                cal = setCalDay(dates[i]);
                d = cal.get(cal.DATE);
                sumDates[dayIndex] = cal.getTime();
                tmp = new Matrix(1,numOfDataFields);
            }
        }
        System.out.println("Finishing TimeSeries.dailySum()");
        return new TimeSeries(sumDates,sumData);
    }
    
  /**
   * Method to normalize the data according to a NormParam object.
   *
   * @param np A NormParam object.
   */
    public TimeSeries normalize(NormParam np)
    {
/*    Matrix ones = new Matrix(1,data.getColumnDimension(),1);
 
    Matrix xMin = (new Matrix(np.getMinValueIn(),np.getNumOfInputs()).times(ones);
 
    Matrix yMin = (new Matrix(data.getRowDimension(),
                              data.getColumnDimension(),
                              np.getMinValueOut());
    Matrix y = data.minus(xMin)
                   .times(1/(np.getMaxValueIn()-np.getMinValueIn()))
                   .times(np.getMaxValueOut()-np.getMinValueOut())
                   .plus(yMin);
 */
        double[] xMin = np.getMinValueIn();
        double[] xMax = np.getMaxValueIn();
        double[] yMin = np.getMinValueOut();
        double[] yMax = np.getMaxValueOut();
        
        Matrix y = new Matrix(data.getRowDimension(),data.getColumnDimension());
        for (int r=0;r<y.getRowDimension();r++)
            for (int c=0;c<y.getColumnDimension();c++)
                y.set(r,c,(data.get(r,c)-xMin[c])/(xMax[c]-xMin[c])*(yMax[c]-yMin[c])+yMin[c]);
        
        return new TimeSeries(dates,y);
    }
    
  /**
   * Method to create a time delay line. The resulting Matrix object
   * has m rows and n columns, where m equals the length of the delay
   * line and n equals the length of the TimeSeries object. The first
   * m-1 columns will contain NaNs.
   *
   * @param np A NormParam object.
   */
/*    public Matrix timeDelay(saaps.network.NormParam np)
    {
        int td = np.getTimeDelay();
        int rows = data.getRowDimension();
        int cols = data.getColumnDimension();
        Matrix res = new Matrix(cols*td,rows,Double.NaN);
        for (int i=0;i<td;i++)
        {
            int r1 = cols*i;
            int r2 = cols*(i+1) - 1;
            res.setMatrix(r1,r2,td-i-1,res.getColumnDimension()-1,data.getMatrix(0,rows-td+i,0,cols-1).transpose());
        }
        return res;
    }
*/    
  /**
   * Get the length of the TimeSeries object.
   */
    public int getLength()
    {
        return numOfData;
    }
    
  /**
   * Get the Date objects.
   */
    public Date[] getDates()
    {
        return dates;
    }
    
  /**
   * Get one Date object.
   */
    public Date getDate(int i)
    {
        return dates[i];
    }
    
  /**
   * Get the data as a Matrix object.
   */
    public Matrix getData()
    {
        return data;
    }
    
  /**
   * Get one data element.
   */
    public double getData(int i, int j)
    {
        return data.get(i,j);
    }
    
  /**
   * Get the data for one row.
   */
    public double[] getData(int i)
    {
        double[] res = new double[numOfDataFields];
        for (int j=0;j<numOfDataFields;j++)
            res[j] = data.get(i,j);
        return res;
    }
    
  /**
   * Get the number of data fields.
   */
    public int getNumOfDataFields()
    {
        return numOfDataFields;
    }
    
  /**
   * Get the short description of the data.
   *
   * @return The short description
   */
    public String getShortDescription()
    {
        return shortDescription;
    }
    
  /**
   * Get the long description of the data.
   *
   * @return The long description
   */
    public String getLongDescription()
    {
        return longDescription;
    }
    
  /**
   * Gets the field names array.
   * @return field names
   */
    public String[] getFieldNames() {
        return fieldNames;
    }
    
  /**
   * Gets the field units array.
   * @return field units
   */
    public String[] getFieldUnits() {
        return fieldUnits;
    }
    
  /**
   * Gets the field names array.
   * @return field names
   */
    public String[] getListNames() {
        return listNames;
    }

  /**
   * Set the short description of the data.
   *
   * @param  sd  The short description
   */
    public void setShortDescription(String sd)
    {
        shortDescription = new String(sd);
    }
    
  /**
   * Set the long description of the data.
   *
   * @param  ld  The short description
   */
    public void setLongDescription(String ld)
    {
        longDescription = new String(ld);
    }
    
  /**
   * Set one row of data.
   */
    public void setData(int i, double[] x)
    {
        for (int j=0;j<numOfDataFields;j++)
            data.set(i,j,x[j]);
    }
    
  /**
   * Set data for observation i and data field j.
   *
   * @param  i The index for one observation
   * @param  j The index of the data field
   * @param  x The value to be set
   */
    public void setData(int i, int j, double x)
    {
        data.set(i,j,x);
    }
    
  /**
   * Get the average of a time series object.
   */
    public double[] average(boolean ignoreNaN)
    {
        double[] res = new double[numOfDataFields];
        int count = 0;
        if (ignoreNaN)
        {
            for (int j=0;j<numOfDataFields;j++)
            {
                count = 0;
                boolean stillNaN = true;
                for (int i=0;i<numOfData;i++)
                {
                    if (!Double.isNaN(data.get(i,j)))
                    {
                        stillNaN = false;
                        res[j] += data.get(i,j);
                        count++;
                    }
                }
                if (stillNaN)
                    res[j] = Double.NaN;
                else
                    res[j] /= count;
            }
        }
        else
        {
            for (int j=0;j<numOfDataFields;j++)
            {
                for (int i=0;i<numOfData;i++)
                    res[j] += data.get(i,j);
                res[j] /= numOfData;
            }
        }
        
        return res;
    }
    
  /**
   * Gets the average of a time series object.
   * The returned array contains the averaged data, then the number of
   * datapoints used in the average for every datapoint, then the
   * standard deviation for every data point.
   * @param date centre time of data interval is returned here
   * @param ignoreNaN if <code>true</code> ignore NaN data
   * @return averaged data and additional information about the average
   */
    public double[] average(Calendar date,boolean ignoreNaN) {
        final int nof = numOfDataFields*aveArrTypes;
        double[] res = new double[nof];
        int[] count = new int[numOfDataFields];
        double[] squareSum = new double[numOfDataFields];
        
        if (ignoreNaN) {
      /* sum up */
            for(int i=0;i<numOfData;i++) {
                for(int j=0;j<numOfDataFields;j++) {
                    double tmp = data.get(i,j);
          /* Skip NaN:s */
                    if(!Double.isNaN(tmp)) {
                        res[j] += tmp;
                        squareSum[j] += Math.pow(tmp,2);
                        count[j]++;
                    }
                }
            }
            
      /* Average */
            for(int j=0;j<numOfDataFields;j++) {
                if (count[j]<1) { //All datapoints were NaN
                    res[j] = Double.NaN;
          /* when all fields are NaN we can leave the
           * other fields of the returned array untouched (zero)
           */
                }
                else {
                    res[j] /= count[j];
          /* number of datapoints in this average */
                    res[j+aveArrDptOffset*numOfDataFields] = count[j];
          /* standard deviation */
                    res[j+aveArrStdvOffset*numOfDataFields] = Math.sqrt((squareSum[j] - (count[j]*Math.pow(res[j],2)))/(count[j]-1));
                }
            }
        }
        else {
      /* sum up */
            for(int i=0;i<numOfData;i++) {
                for(int j=0;j<numOfDataFields;j++) {
                    res[j]+= data.get(i,j);
                    squareSum[j] += Math.pow(data.get(i,j),2);
                }
            }
            
      /* Average */
            for(int j=0;j<numOfDataFields;j++) {
                res[j] /= numOfData;
        /* number of datapoints in this average */
                res[j+aveArrDptOffset*numOfDataFields] = numOfData;
        /* standard deviation */
                res[j+aveArrStdvOffset*numOfDataFields] = Math.sqrt((squareSum[j] - (numOfData*Math.pow(res[j],2)))/(numOfData-1));
            }
        }
    /* Set the date/time to middle of data interval */
        date.setTime(new Date(dates[0].getTime() +
        (dates[numOfData-1].getTime() -
        dates[0].getTime())/2));
        
        return res;
    }
    
/**
 * Averages all fields in the data for the selected time interval.
 * The data is returned in mdate and mdata.
 * The time is set to the sample time.
 * @param sampleTime time of the resulting data.
 * @param wa time offset 1 (negative value means before sampleTime)
 * @param wb time offset 2 (positive value means after sampleTime)
 * @return the averaged object
 */
    public int average(Date sampleTime,long wa,long wb,Date mdate, double[] mdata) {
        
        int cnt[] = new int[numOfDataFields]; //Count the number of data points
        
      /* set time of resulting object */
        mdate.setTime(sampleTime.getTime());
        
        /* if only one datapoint return it */
        if(dates.length == 1) {
            mdate.setTime(dates[0].getTime());
            for(int i=0;i<numOfDataFields;i++) {
                mdata[i] = data.get(0,i);
            }
            return 0;
        }
        Date startTime = new Date(sampleTime.getTime() + wa);
        Date endTime = new Date(sampleTime.getTime() + wb);
        
        if(startTime.after(endTime)) {
            System.err.println("starttime is after endtime.");
            return -1;
        }
        /* clear mdata */
        for(int i=0;i<mdata.length;i++)
            mdata[i]=0;
        
        for(int i=0;i<dates.length;i++) {
            /* if inside time interval use it else if after endTime break loop */
            if((dates[i].compareTo(startTime) >= 0) && (dates[i].compareTo(endTime) <= 0)) {
                for(int j=0;j<numOfDataFields;j++) {
                    double tmp = data.get(i,j);
                    /* Skip NaN:s */
                    if(!Double.isNaN(tmp)) {
                        mdata[j]+= tmp;
                        cnt[j]++;
                    }
                }
            }
            else if(dates[i].after(endTime))
                break;
        }
        
        /* Average */
        for(int j=0;j<numOfDataFields;j++) {
            mdata[j]/= cnt[j];
        }
        return 0;
    }
 
    /**
     * Averages the selected fields in the data for the selected time interval.
     * The data is returned in mdate and mdata.
     * The time is set to the sample time.
     * @param fields which are included in the average
     * @param sampleTime time of the resulting data.
     * @param wa time offset 1 (negative value means before sampleTime)
     * @param wb time offset 2 (positive value means after sampleTime)
     * @return the averaged object
     */
    public int average(int[] fields,Date sampleTime,long wa,long wb,Date mdate, double[] mdata) {
        
        int cnt[] = new int[fields.length]; //Count the number of data points
        
      /* set time of resulting object */
        mdate.setTime(sampleTime.getTime());
        
        /* if only one datapoint return it */
        if(dates.length == 1) {
            mdate.setTime(dates[0].getTime());
            for(int i=0;i<fields.length;i++) {
                mdata[i] = data.get(0,fields[i]);
            } 
            return 0;
        }
        Date startTime = new Date(sampleTime.getTime() + wa);
        Date endTime = new Date(sampleTime.getTime() + wb);
        
        if(startTime.after(endTime)) {
            System.err.println("starttime is after endtime.");
            return -1;
        }
        /* clear mdata */
        for(int i=0;i<mdata.length;i++)
            mdata[i]=0;
        
        for(int i=0;i<dates.length;i++) {
            /* if inside time interval use it else if after endTime stop the loop */
            if((dates[i].compareTo(startTime) >= 0) && (dates[i].compareTo(endTime) <= 0)) {
                for(int j=0;j<fields.length;j++) {
                    double tmp = data.get(i,fields[j]);
                    /* Skip NaN:s */
                    if(!Double.isNaN(tmp)) {
                        mdata[j]+= tmp;
                        cnt[j]++;
                    }
                }
            }
            else if(dates[i].after(endTime))
                break;
        }
        
        /* Average */
        for(int j=0;j<fields.length;j++) {
            mdata[j]/= cnt[j];
        }
        return 0;
    }

  /**
   * Extract a time period.
   */
    public TimeSeries extract(long st, long et)
    {
        int i1 = 0;
        while (dates[i1].getTime() < st)
            i1++;
        int i2 = i1;
        while (dates[i2].getTime() < et)
            i2++;
        Date[] d = new Date[i2-i1+1];
        for (int i=0;i<d.length;i++)
            d[i] = dates[i1+i];
        Matrix m = data.getMatrix(i1,i2,0,numOfDataFields-1);
        return new TimeSeries(d,m);
    }
    
  /**
   * Sum over the data fields.
   */
    public TimeSeries sumFields()
    {
        int r = data.getRowDimension();
        int c = data.getColumnDimension();
        Matrix x = data.getMatrix(0,r-1,0,0);
        for (int i=1;i<c;i++)
            x = x.plus(data.getMatrix(0,r-1,i,i));
        return new TimeSeries(dates,x);
    }
    
  /**
   * Average over the data fields.
   */
    public TimeSeries averageFields()
    {
        int r = data.getRowDimension();
        int c = data.getColumnDimension();
        Matrix x = data.getMatrix(0,r-1,0,0);
        for (int i=1;i<c;i++)
            x = x.plus(data.getMatrix(0,r-1,i,i));
        x = x.times(1/(double) c);
        return new TimeSeries(dates,x);
    }
    
  /**
   * Get the last date in the time series.
   *
   * @return The last date.
   */
    public Date getLastDate()
    {
        return dates[dates.length-1];
    }
    
  /**
   * Get the sample interval.
   *
   * @return The sample interval.
   */
    public long getSampleInterval()
    {
        return dates[1].getTime() - dates[0].getTime();
    }
    
  /**
   * Resamples all fields in the data.
   *
   * XXX The list has to be scanned through for every new
   * sample.
   * XXX Should throw an Exception if it fails ! Fix it !!
   *
   * @param dt0 offset time from the first time in the series [ms]
   * @param dt new sample interval (timestep [ms])
   * @param wa time offset where resampling begins [ms]
   * @param wb time offset where resampling ends [ms]
   * @return -1 if failed 0 when normal operation.
   */
    public byte resample(long dt0,long dt,long wa,long wb) {
        
        /* Get the ave time */
        Date date = null;
        
        long timeSpan = dates[numOfData-1].getTime() - dates[0].getTime();
        System.out.println("TimeSpan:" + timeSpan);
        /* Create a list that can hold the resampled data objects */
        int objCnt = (int)( timeSpan/dt);
        if(objCnt < 2) {
            System.err.println("List is too short:" + objCnt);
            return -1;
        }
        System.out.println("num new objs:" + objCnt);
        
        Date[] newDates = new Date[objCnt];
        Matrix newData = new Matrix(objCnt,numOfDataFields);
        
        //First sample will be here
        Calendar mtime = Calendar.getInstance(TimeZone.getTimeZone("UTC"),new Locale("sv","SE"));
        mtime.setTime(new Date(dates[0].getTime()+dt0));
        double mdata[] = new double[numOfDataFields];
        Date mdate = new Date();
        for(int i=0;i<objCnt;i++) {
            average(mtime.getTime(),wa,wb,mdate,mdata);
            newDates[i]= new Date(mdate.getTime());
            for(int j=0;j<numOfDataFields;j++) {
                newData.set(i,j,mdata[j]);
            }
            //step time forward with the new sample interval
            mtime.add(Calendar.SECOND,(int) dt/1000);
            mtime.add(Calendar.MILLISECOND,(int) (dt - ((int)dt/1000)*1000));
        }
        
        dates = newDates;
        data = newData;
        numOfData = dates.length;
        //numOfFields = should be the same because it's the same data type !!
        //numOfDataFields = should be the same because it's the same data type !!
        
        return 0;
    }
    
      /**
   * Resamples selected fields in the data.
   *
   * XXX The list has to be scanned through for every new
   * sample.
   * XXX Should throw an Exception if it fails ! Fix it !!
   *
   * @param fields which will be included in the result
   * @param dt0 offset time from the first time in the series [ms]
   * @param dt new sample interval (timestep [ms])
   * @param wa time offset where resampling begins [ms]
   * @param wb time offset where resampling ends [ms]
   * @return -1 if failed 0 when normal operation.
   */
    public byte resample(int[] fields,long dt0,long dt,long wa,long wb) {
                
        long timeSpan = dates[numOfData-1].getTime() - dates[0].getTime();
        System.out.println("TimeSpan:" + timeSpan);
        /* Create a list that can hold the resampled data objects */
        int objCnt = (int)( timeSpan/dt);
        if(objCnt < 2) {
            System.err.println("List is too short:" + objCnt);
            return -1;
        }
        System.out.println("num new objs:" + objCnt);
        
        Date[] newDates = new Date[objCnt];
        Matrix newData = new Matrix(objCnt,java.lang.reflect.Array.getLength(fields));
        
        //First sample will be here
        Calendar mtime = Calendar.getInstance(TimeZone.getTimeZone("UTC"),new Locale("sv","SE"));
        mtime.setTime(new Date(dates[0].getTime()+dt0));
        double mdata[] = new double[java.lang.reflect.Array.getLength(fields)];
        Date mdate = new Date();
        for(int i=0;i<objCnt;i++) {
            average(fields,mtime.getTime(),wa,wb,mdate,mdata);
            newDates[i]= new Date(mdate.getTime());
            for(int j=0;j<java.lang.reflect.Array.getLength(fields);j++) {
                newData.set(i,j,mdata[j]);
            }
            //step time forward with the new sample interval
            mtime.add(Calendar.SECOND,(int) dt/1000);
            mtime.add(Calendar.MILLISECOND,(int) (dt - ((int)dt/1000)*1000));
        }
        
        dates = newDates;
        data = newData;
        numOfData = dates.length;
        numOfFields = java.lang.reflect.Array.getLength(fields) + 1 ;
        numOfDataFields = java.lang.reflect.Array.getLength(fields);
        String[] fn = new String[java.lang.reflect.Array.getLength(fields)];
        String[] fu = new String[java.lang.reflect.Array.getLength(fields)];
        for(int i=0;i<java.lang.reflect.Array.getLength(fields);i++) {
            fn[i]=fieldNames[fields[i]];
            fu[i]=fieldUnits[fields[i]];
        }
        fieldNames = fn;
        fieldUnits = fu;
        
        return 0;
    }
    
    public static void main(String[] args) {
        
        try {
            NumberFormat nf = NumberFormat.getInstance();
            nf.setMaximumFractionDigits(3);
            nf.setMinimumFractionDigits(3);
            
            Calendar sd = Calendar.getInstance(TimeZone.getTimeZone("UTC"),new Locale("sv","SE"));
            Calendar ed = Calendar.getInstance(TimeZone.getTimeZone("UTC"),new Locale("sv","SE"));
            sd.set(2000,6-1,1,0,0,0);
            sd.clear(sd.MILLISECOND);
            ed.set(2000,6-1,2,0,0,0);
            ed.clear(ed.MILLISECOND);
            long start = -10*60*1000;
            long end = 10*60*1000;
            
            Database db = Database.getInstance(args[0],true);
            
            TimeSeries ts = db.retrieveTS( sd.getTime(), ed.getTime(),false);
            //ots.print();
            ts.resample(new int[]{3,2,1,0},0,60*60*1000,-30*60*1000,30*60*1000);
            //ots.print();
            
            System.out.println("TimeSeries: " + ts.getShortDescription() + " long descr:" + ts.getLongDescription());
            
            String[] fn = ts.getFieldNames();
            for(int i=0;i<fn.length;i++) {
                System.out.print(fn[i]);
                if(i<(fn.length-1))
                    System.out.print(" : ");
            }
            System.out.println("");
            
            for(int i = 0;i<ts.getLength();i++) {
                System.out.print("d: " + ts.getDate(i));
                
                for(int j=0;j<ts.getNumOfDataFields();j++) {
                    System.out.print(" " + nf.format(ts.getData(i,j)));
                }
                System.out.println("");
            }
            
        }
        catch (Exception e) {
            System.err.println(e);
            e.printStackTrace();
        }
}
    
}
