package saaps.util;

import java.text.DateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.Locale;
import java.util.TimeZone;
import ptolemy.plot.Plot;
import saaps.util.TimeSeries;

/**
* A class to produce a plot of time series data. The class is an
* extension of the <code>ptolemy.plot.Plot</code> class which in
* turn extends the <code>java.awt.Panel</code> class.
* <p>
* The constructor takes a <code>saaps.util.TimeSeries</code> object in
* its argument. It the automatically selects the x-axis to represent time
* and the y-axis the data in the data fields of the <code>TimeSeries</code>
* object. The tick marks on the x-axis are set so that the time information
* becomes clear. E.g., if the data extends over several days then the tick
* marks are in days, while if it extends over one day or less then the tick
* marks are in hours. The x-label shows the start date of the plot. The y-label
* shows the name and unit of the plotted data. If there are several data
* fields plotted then each line has different colours and the y-label lists
* all data field names and units. The field name and units are extracted from
* the <code>TimeSeries</code> object.
*
* @author   Peter Wintoft
* @version  2000-04-10
* @see      ptolemy.plot.Plot
* @see      java.awt.Panel
* @see      saaps.util.TimeSeries
*/
public class TimeSeriesPlot extends Plot{

  // Calendar time.
  public static final int CAL_TIME = -1;
  // Relative time.
  public static final int REL_TIME = -2;
  // The x-axis data.
  double[] xData;
  double xOffset;
  double xOffsetUsed;
  // The time div.
  double timeDiv;
  // The y-axis data.
  double[][] yData;
  // The x-axis field.
  int xDataField;
  // The y-axis fields.
  int[] yDataFields;
  // If x data is time then use timeUnit.
  String timeUnit;
  // The x-axis label.
  String xAxisLabel;
  // The y-axis labels.
  String[] yAxisLabels;
  // X-log.
  boolean xLog = false;
  // Y-log.
  boolean yLog = false;
  // Lines.
  boolean lines = true;
  // Points.
  boolean points = false;
  // The first date in the time series.
  // Used for calendar time plotting.
  Date firstDate;
  // The date of time zero.
  Date zeroDate;

  /**
  * Create the default time series plot. Plot all data fields
  * as function of calendar time.
  *
  * @param  ts  The <code>TimeSeries</code> object to be plotted.
  */
  public TimeSeriesPlot(TimeSeries ts)
  {
    //Initialize plot.
    super();
    initPlot();

    //Set parameters.
    firstDate = ts.getDate(0);
    zeroDate = firstDate;
    //Assume time on x-axis.
    xDataField = -1;
    //Assume all data on y-axis.
    int n = ts.getNumOfDataFields();
    yDataFields = new int[n];
    //Assume only lines.
    lines = true;
    points = false;
    yAxisLabels = new String[yDataFields.length];
    String[] fieldNames = ts.getFieldNames();
    String[] listNames = ts.getListNames();
    String[] fieldUnits = ts.getFieldUnits();
    if (listNames == null) {
      for (int i=0;i<yDataFields.length;i++)
      {
        yDataFields[i] = i;
        yAxisLabels[i] = new String(fieldNames[yDataFields[i]] + " (" + fieldUnits[yDataFields[i]] + ")");
      }
    }
    else {
      for (int i=0;i<yDataFields.length;i++)
      {
        yDataFields[i] = i;
        yAxisLabels[i] = new String(listNames[yDataFields[i]] + " (" + fieldUnits[yDataFields[i]] + ")");
      }
    }
    //Assume linear scales.
    xLog = false;
    yLog = false;
   
    //Extract data.
    extractData(ts);
    
    //Set the x-axis label.
    if (xDataField<0)
      xAxisLabel = new String("Time (" + timeUnit + ")");
    else if (listNames == null)
      xAxisLabel = new String(fieldNames[xDataField] + " (" + fieldUnits[xDataField] + ")");
    else
      xAxisLabel = new String(listNames[xDataField] + " (" + fieldUnits[xDataField] + ")");
    
    //Set the used x-offset.
    xOffsetUsed = xOffset;

    //Log plot?
    setXLog(xLog);
    setYLog(yLog);

    //Plot the data.
    plotData();

    //Set labels.
    setXLabel();
    setYLabels();

    //Set the title.
    setTitle(ts.getShortDescription());
}

  /**
  * Create a custom plot. This is useful for e.g. plotting data fields as a function
  * of another data field.
  *
  * @param  ts  The <code>TimeSeries</code> object.
  * @param  xf  An integer pointing to which data field should be used on the x-axis.
  * @param  yfs An array of integers pointing to which data fields that should be on the y-axis.
  * @param  li  A boolean that is true if lines should be used.
  * @param  po  A boolean that is true if points should be used.
  * @param  xLabel  A string for the x-axis label.
  * @param  yLabels An array of strings for the y-axis labels.
  * @param  xLog  A boolean that is true of the x-axis should be logarithmic.
  * @param  yLog  A boolean that is true of the y-axis should be logarithmic.
  */
  public TimeSeriesPlot(TimeSeries ts, int xf, int[] yfs, boolean li, boolean po, String xLabel, String[] yLabels, boolean xL, boolean yL)
  {
    //Initialize plot.
    super();
    initPlot();

    //Set parameters.
    firstDate = ts.getDate(0);
    zeroDate = firstDate;
    xDataField = xf;
    yDataFields = new int[yfs.length];
    lines = li;
    points = po;
    xAxisLabel = new String(xLabel);
    yAxisLabels = new String[yLabels.length];
    for (int i=0;i<yfs.length;i++)
    {
      yDataFields[i] = yfs[i];
      yAxisLabels[i] = new String(yLabels[i]);
    }
    xLog = xL;
    yLog = yL;
   
    //Extract data.
    extractData(ts);
    
    //Set the used x-offset.
    xOffsetUsed = xOffset;

    //Log plot?
    setXLog(xLog);
    setYLog(yLog);

    //Plot the data.
    plotData();

    //Set labels.
    setXLabel();
    setYLabels();

    //Set the title.
    setTitle(ts.getShortDescription());
  }

  public TimeSeriesPlot(TimeSeries ts, int xf, int[] yfs, boolean li, boolean po, boolean xL, boolean yL)
  {
    //Initialize plot.
    super();
    initPlot();

    //Set parameters.
    firstDate = ts.getDate(0);
    zeroDate = firstDate;
    xDataField = xf;
    yDataFields = new int[yfs.length];
    lines = li;
    points = po;
    yAxisLabels = new String[yfs.length];
    String[] fieldNames = ts.getFieldNames();
    String[] listNames = ts.getListNames();
    String[] fieldUnits = ts.getFieldUnits();
    if (listNames == null) {
      for (int i=0;i<yfs.length;i++)
      {
        yDataFields[i] = yfs[i];
        yAxisLabels[i] = new String(fieldNames[yDataFields[i]] + " (" + fieldUnits[yDataFields[i]] + ")");
      }
    }
    else {
      for (int i=0;i<yfs.length;i++)
      {
        yDataFields[i] = yfs[i];
        yAxisLabels[i] = new String(listNames[yDataFields[i]] + " (" + fieldUnits[yDataFields[i]] + ")");
      }
    }
    xLog = xL;
    yLog = yL;
   
    //Extract data.
    extractData(ts);
    
    //Set the x-axis label.
    if (xDataField<0)
      xAxisLabel = new String("Time (" + timeUnit + ")");
    else if (listNames == null)
      xAxisLabel = new String(fieldNames[xDataField] + " (" + fieldUnits[xDataField] + ")");
    else
      xAxisLabel = new String(listNames[xDataField] + " (" + fieldUnits[xDataField] + ")");
    
    //Set the used x-offset.
    xOffsetUsed = xOffset;

    //Log plot?
    setXLog(xLog);
    setYLog(yLog);

    //Plot the data.
    plotData();

    //Set labels.
    setXLabel();
    setYLabels();

    //Set the title.
    setTitle(ts.getShortDescription());
  }
  
  private void initPlot()
  {
    setSize(800,300);
    setButtons(false);
  }
  
  public void setTimeSeries(TimeSeries ts)
  {
    //Get the first date.
    firstDate = ts.getDate(0);
    zeroDate = firstDate;

    //Assume to plot all fields.
    //Generate default y-fields and y-labels.
    int n = ts.getNumOfDataFields();
    yDataFields = new int[n];
    yAxisLabels = new String[n];
    for (int i=0;i<n;i++)
    {
      yDataFields[i] = i;
      yAxisLabels[i] = String.valueOf(i);
    }
    
    extractData(ts);
  }

  /**
  * Extract the data from the time series object.
  */
  private void extractData(TimeSeries ts)
  {
    //Initiate the data.
    xData = new double[ts.getLength()];
    yData = new double[ts.getLength()][];
    //Get x data.
    timeUnit = "seconds";
    if (xDataField == CAL_TIME || xDataField == REL_TIME)
    {
      Date[] d = ts.getDates();
      double x1 = (double) d[0].getTime()/1000;
      double x2 = (double) d[d.length-1].getTime()/1000;
      double xw = x2-x1;
      timeDiv = 1000;
      if (xw>60 & xw<=3600)
      {
        timeDiv = 1000*60;
        timeUnit = "minutes";
      }
      if (xw>3600 & xw<=3600*24)
      {
        timeDiv = 1000*3600;
        timeUnit = "hours";
      }
      if (xw>3600*24)
      {
        timeDiv = 1000*3600*24;
        timeUnit = "days";
      }
//      if (xDataField == CAL_TIME)
//      {
      xOffset = (double) d[0].getTime() / timeDiv;
//      xOffsetUsed = xOffset;
//        for (int i=0;i<d.length;i++)
//         xData[i] = (double) ((d[i].getTime() - d[0].getTime())) / timeDiv;
//      }
//      else
//      {
      for (int i=0;i<d.length;i++)
        xData[i] = (double) d[i].getTime() / timeDiv;
//      }
    }
    else
    {
      for (int i=0;i<ts.getLength();i++)
        xData[i] = ts.getData(i,xDataField);
    }
    //Get the y data.
    for (int i=0;i<ts.getLength();i++)
    {
      yData[i] = new double[yDataFields.length];
      for (int j=0;j<yDataFields.length;j++)
        yData[i][j] = ts.getData(i,yDataFields[j]);
    }
  }

  private void setXLabel()
  {
    //Set labels.
    if (xDataField==CAL_TIME)
    {
      Calendar cal = Calendar.getInstance(TimeZone.getTimeZone("GMT"));
      DateFormat df = DateFormat.getDateTimeInstance(DateFormat.MEDIUM,DateFormat.MEDIUM,new Locale("uk","UK"));
      df.setCalendar(cal);
      setXLabel(timeUnit + " since " + df.format(zeroDate));
    }
    else
    {
      if (xDataField==REL_TIME)
         setXLabel(timeUnit);
      else
        setXLabel(xAxisLabel);
    }
  }

  private void setYLabels()
  {
    for (int i=0;i<yAxisLabels.length;i++)
      addLegend(i,yAxisLabels[i]);
  }
  
  public boolean getLineState()
  {
    return lines;
  }
  
  public void setLineState(boolean l)
  {
    lines = l;
  }
  
  public boolean getPointState()
  {
    return points;
  }
  
  public void setPointState(boolean p)
  {
    points = p;
  }

  /**
  * Create the plot.
  */
  public void plotData()
  {
    clear(false);
    for (int j=0;j<yData[0].length;j++)
    {
      boolean connect = false;
      for (int i=0;i<xData.length;i++)
      {
        if (Double.isNaN(xData[i]) || Double.isNaN(yData[i][j]) ||
        (getXLog() && ((xData[i]<0) || (xData[i]==0))) || (getYLog() && ((yData[i][j]<0) || (yData[i][j]==0))) ) {
          connect = false;
        }
        else
        {
          addPoint(j,xData[i]-xOffsetUsed,yData[i][j],connect);
          connect = lines;
        }
      }
    }
    if (points)
      setMarksStyle("various");
    else
      setMarksStyle("none");
  }

  /**
  * Call this method to use calendar time on the x-axis.
  */
  public void useCalTime()
  {
    xDataField = CAL_TIME;
    xOffsetUsed = xOffset;
    zeroDate = firstDate; 
    setXLabel();
  }

  /**
  * Call this method to use relative time on the x-axis.
  * <p>
  * Relative time is used for <code>TimeSeries</code> objects where
  * the time field represents time reletive to an event. The event
  * time is equal to zero.
  */
  public void useRelTime()
  {
    xDataField = REL_TIME;
    xOffsetUsed = 0;
    setXLabel();
  }
  
  public void setZeroTime(Date d)
  {
    xOffsetUsed = (double) d.getTime() / timeDiv;
    zeroDate = d;
    setXLabel();
  }
}
