###
# $Id: Sedat.pm.in,v 1.5 2002/11/12 11:32:00 stamper Exp $
#
#  Purpose
#  
#   To provide various utility functions for the SEDAT CGI routines.
#   This file is parametrised with the string '@-PARAM-@' (without the dashes
#   present here to prevent unwanted substitution) and is processed using
#   'make' to generate the Sedat.pm file. 
#   
#  Author	   Richard Stamper
#
#  $Log: Sedat.pm.in,v $
#
#  Revision 1.6  2009/05/04 13:55:00  heynderickx
#  Added ODI_connect to open connection to the ODI database
#
#  Revision 1.5  2002/11/12 11:32:00  stamper
#  In relativeDir, eliminate redundant .. components.
#
#  Revision 1.4  2001/11/26 17:06:14  stamper
#  Use File::Spec->catfile rather than the local joinName function.
#  Minor typographical changes too.
#
#  Revision 1.3  2001/05/04 15:28:35  stamper
#  Set the umask to 02 so that group writability is enabled by default.
#  Check for null values in the returnField function.
#  Remove the exceptional processing for SYSTEM objects in objectName - this
#  now should be done in the calling routine where necessary.
#  Make nameParts handle any form with a '!' in.
#  recursivelyCopy no longer aborts the CGI transaction, but returns an error
#  to the caller so that rollback can be executed where necessary.
#  Make recursivelyDelete always remove the top-level directory.
#
#  Revision 1.2  2001/02/27 16:16:58  stamper
#  Corrected the relativeDir subroutine to handle correctly the cases when:
#   - the destination is in a subtree of the current directory
#   - the destination is a parent of the current directory
#
###
package Sedat;
use strict;
use vars qw(@ISA
	    @EXPORT
	    $VERSION
	    $OBJECTS
	    %TYPES
	    %PARAMS);

use Exporter;
use File::Copy;
use File::Basename;
use File::Spec;

use DBI;

# 
# Initialisation
# 
# The 'param' file is read and its contents parsed into the %PARAMS hash. 
# 
BEGIN {
    my ($line,$key,$value);
    my ($paramfile) = 'odi.param';

    # Set up default parameters
    $PARAMS{directory_sep} = '/';
    $PARAMS{server_dir}    = '/opt/sedat/server_dir';
    $PARAMS{server_fifo}   = '/var/tmp/sedat_queue';
    $PARAMS{object_dir}    = '/opt/sedat/objects';
    $PARAMS{bin_dir}       = '/opt/sedat/bin';

    # Read the system parameters file into a hash.
    open PARAMFILE,$paramfile or
	die "Failed to open system parameter file $paramfile\n";
    while ($line = <PARAMFILE>) {
	chomp $line;
	next if ($line =~ /^\#/);
	next if ($line =~ /^\s*$/);
	($key,$value) = split(/\s*:\s*/,$line,2);
	if (defined($key)) {
	    $value = '' if (!defined($value));
	    $PARAMS{lc($key)} = $value;
	}
    }
    close PARAMFILE;

    # Set the umask so that group write is enabled by default
    umask 02;

    # Export the STPDF parameters as environment variables if they exist
    $ENV{STPDFDYNAMIC} = $PARAMS{stpdfdynamic} if exists($PARAMS{stpdfdynamic});
    $ENV{STPDFMRDLRD}  = $PARAMS{stpdfmrdlrd}  if exists($PARAMS{stpdfmrdlrd});
    $ENV{STPDFGUNZIP}  = $PARAMS{stpdfgunzip}  if exists($PARAMS{stpdfgunzip});
    $ENV{STPDFTMPDIR}  = $PARAMS{stpdftmpdir}  if exists($PARAMS{stpdftmpdir});
}

$VERSION = 1.00;
@ISA= qw(Exporter);

#
# These are the methods visible to the outside
#
@EXPORT = qw(returnField
	     abortCGI
	     endCGI
	     joinName
	     objectName
             objectRealName
	     objectDir
	     ownerAndName
	     nameParts
	     checkConnection
	     recursivelyCopy
	     recursivelyDelete
	     systemParam
	     relativeDir
            ODI_connect
	     );

$OBJECTS = $PARAMS{object_dir};

%TYPES = ( B => '',
           P => 'projects',
	   D => 'datasets',
           Q => 'queries' ,
	   T => 'tools'   ,
	   F => 'figures' ,
	   L => 'logs' ,
	   U => 'users'    );

#
# systemParam - returns a system parameter read from the system 
#               configuration files
#
sub systemParam {
    my ($param) = @_;

    return $PARAMS{lc($param)};
}

#
# returnField - Prints the details of a field in a form acceptable to the
#               Java DataHandler
#
sub returnField {
    my ($field,$type,@values) = @_;

    print "$field\n";
    print "$type\n";
    if (@values) {
	@values = grep defined,@values;
	if ($#values >= 0) {
	    print join("\001",@values);
	    print "\001";
	}
    }
    print "\002\n";
}

#
# abortCGI - Halts a CGI, but also prints status information for the
#            Java DataHandler
#
sub abortCGI {
    my ($msg) = @_;

    returnField("status.code","I",1);
    returnField("status.message","T",$msg);
    exit(1);
}

#
# endCGI - Ends a CGI successfully and prints status information for the
#          Java DataHandler
#
sub endCGI {
    my ($msg) = @_;

    returnField("status.code","I",0);
    returnField("status.message","T",$msg);
    exit(0);
}

#
# ownerAndName - Join an object owner and object name to give a composite
#
sub ownerAndName {
    my ($owner,$name) = @_;

    return join('!',$owner,$name);
}

#
# joinName - Connects parts of a file specification in an appropriate
#            way for this operating system
#
sub joinName {
    return File::Spec->catfile(@_);
}

#
# objectName - Generate the location of an object, always a link to a 
#              directory.
#
sub objectName {
    my ($requester,$owner,$type,$name) = @_;

    foreach ($type) {
	/B/ && return $OBJECTS;
	/U/ && return File::Spec->catfile($OBJECTS,$TYPES{'U'},$name);
	return File::Spec->catfile($OBJECTS,$TYPES{'U'},$requester,
				   $TYPES{$type},ownerAndName($owner,$name));
    }
}

#
# objectRealName - Generate the location of an object, always a directory.
#
sub objectRealName {
    my ($owner,$type,$name) = @_;

    foreach ($type) {
	/B/ && return $OBJECTS;
	/U/ && return File::Spec->catfile($OBJECTS,$TYPES{'U'},$name);
        return File::Spec->catfile($OBJECTS,$TYPES{'U'},$owner,
				   $TYPES{$type},$name);
    }
}

#
# objectDir - Append a directory name for a given object type
#
sub objectDir {
    my ($base,$type) = @_;

    return File::Spec->catfile($base,$TYPES{$type});
}

#
# nameParts - Split an object name up into owner and object parts, when given
#             a full path name beginning with the OBJECTS root
#             
sub nameParts {
    my $fullname = shift;

    if (index($fullname,$OBJECTS) != 0) {
	return split(/!/,$fullname,2);
    } else {
	return split(/!/,substr($fullname,length($OBJECTS)),2);
    }
}

#
# checkConnection - Carry out standard checks for data items that all CGI
#                   communications from the SEDAT server should return
sub checkConnection {
    my ($user,$remoteuser,$session,$source) = @_;

    # Check user
    if (! defined $user) {
	abortCGI('Missing user name');
    }
    if (! defined $remoteuser) {
	abortCGI('Missing remote user name');
    }
    if ($remoteuser ne $user) {
	abortCGI('Mismatch between user name and remote user name');
    }

    # Check session
    if (! defined $session) {
    	abortCGI('Missing session id');
    }

    # Check source
    if (! defined $source) {
    	abortCGI('Missing source id');
    }
}

#
# recursivelyCopy - Copy an object directory recursively
#                   
sub recursivelyCopy {
    my ($objectOld,$objectNew) = @_;
    my ($file,@list,$fileOld,$fileNew);

    # Get the contents of the existing object
    opendir LIST,$objectOld or
	return 1,"Failed listing object $objectOld";
    @list = grep {!/^\./} readdir(LIST);
    closedir LIST;

    # Work through the list of contents
    foreach $file (@list) {
        # Make up the full path names for old and new versions of a thing
	$fileOld = joinName($objectOld,$file);
	$fileNew = joinName($objectNew,$file);

        # Process a thing depending on its type     
	if (!-l $fileOld && -d $fileOld) {	# If a directory
	    # Make a new directory and copy it recursively in turn
	    mkdir $fileNew, 0775 or
		return 1,"Failed creating directory $fileNew";
	    recursivelyCopy($fileOld,$fileNew);
	} elsif (-l $fileOld) {			# If a link
	    # 
	    # Link to the destination from the new version
	    symlink(readlink($fileOld), $fileNew) or
		return 1,"Failed creating link to ".readlink($fileOld);
	} elsif (-e $fileOld) {			# If a file
	    # Just copy the file
	    copy($fileOld,$fileNew) or
		return 1,"Failed copying file $fileOld";
	} else {				# Otherwise abort
	    return 1,"Failed to find file $fileOld";
	}
    }
    return (0,"OK");
}

#
# recursivelyDelete - Delete an object directory recursively
# 
sub recursivelyDelete {
    my $object = shift;
    my (@list,$file);

    # Get the contents of the existing object
    opendir LIST,$object or
	abortCGI("Failed listing object $object");
    @list = grep {!/^\.{1,2}$/} readdir(LIST);
    closedir LIST;

    # Work through the list of contents
    foreach $file (@list) {
	# Make up the full path name
	$file = joinName($object,$file);

	# Process according to its type
	if (!-l $file && -d $file) {		# If a real subdirectory
	    # Delete it recursively
	    recursivelyDelete($file);
	} elsif (-l $file || -e $file) {	# If a link or file, unlink
	    unlink($file) or
		abortCGI("Failed deleting file $file");
	} else {				# Otherwise abort
	    abortCGI("Failed to find file $file");
	}
    }
    rmdir $object;
}

#
# relativeDir - Express one path as relative to another
# 
sub relativeDir {
    # $here is where to begin the relative path
    # $there is where to get to
    my ($here,$there) = @_;
#    return File::Spec->abs2rel($there,$here);

    my (@here,@there,@result,@newresult,$t,$h);

    # Make the destination relative to here if not absolute
    if (substr($there,0,1) ne '/') {
	$there = File::Spec->catdir($here,$there);
    }

    # Break the paths into their components
    @here  = grep /.+/,split(m!$PARAMS{directory_sep}!,$here);
    @there = grep /.+/,split(m!$PARAMS{directory_sep}!,$there);
    @result = ();
    
    # Work through the components of here and there eliminating all those that
    # match from the beginning, and replacing the mismatches with '..'

    # Get the first component of there
    $t = shift(@there);
    while ($#here >= 0) {
	  # Get the next component of here
	  $h = shift(@here);
	  if (defined $t && $t eq $h) {		# If here and there match 
	      # discard the component
	      $t = shift(@there);			
	  } else {				# otherwise
	      # Record the need to go up a level
	      push @result,'..';
	      if (defined $t && $t ne '') {	# If the first mismatch
		  # Put back the mismatched component
		  unshift @there,$t;
		  # Flag the mismatch - '' will never match $h
		  $t = '';
	      }
	  }
    }

    # Put back any component removed from $there but not discarded
    unshift @there,$t if (defined $t && $t ne '');

    # Add the portion of @there that mismatched to the list of '..' elements
    push @result,@there;

#    # Work down the list eliminating unnecessary components
    @newresult = ();
    while ($t = shift @result) {
	 # Ignore a single .
	 if ($t eq '.') {
	     next;
	 # A .. is ignored if there is a previous element, which is removed
	 } elsif (($t eq '..') && 
		  ($#newresult >= 0) &&
		  ($newresult[-1] ne '..')) {
	     pop @newresult;
	 # Otherwise we keep the element
	 } else {
	     push @newresult,$t;
	 }
    }

    # Reform the components into a path
    return File::Spec->catfile(@newresult);
}

#
# ODI_connect - Open a connection to the ODI database
# 
sub ODI_connect {

# Remember to use lower case in $PARAMS{}!

    my $dsn = "DBI:mysql:host=$PARAMS{odi_hostname};mysql_socket=$PARAMS{odi_socket};port=$PARAMS{odi_port};database=$PARAMS{odi_database};mysql_local_infile=1";
    my $dbh = DBI->connect($dsn, $PARAMS{odi_username}, $PARAMS{odi_password});

    return $dbh;
}
