<?php
// Library functions for the ODI database.
// Peter Wintoft, Swedish Institute of Space Physics
// Daniel Heynderickx, DH Consultancy, Belgium
// 2009-05-13

// Set the correct line feed depending on OS.
if (substr(PHP_OS, 0, 3) == "WIN")
  define("LF", "\r\n");
else
  define("LF", "\n");

define("CDF1970","62167219200000.0");

// Connection parameters for a given user level.
// An array is returned with the ODI database parameters
// based on the environment variables and the user level.
function get_conn_array($user_level) {

    // List with parameters to be used to connect to the database.
    // If the environment variable is not required it is markes as false.
    $param=array(
        "host"=>array("ODI_HOST",false),
        "user"=>array("ODI_USER_".$user_level,true),
        "password"=>array("ODI_PW_".$user_level,true),
        "port"=>array("ODI_PORT",false),
        "socket"=>array("ODI_SOCKET",false),
        "database"=>array("ODI_DB",true)
    );

    $ca=array();
    
    foreach($param as $key=>$val) {
        list($env_var,$req) = $val;
        if($ev=getenv($env_var)) {
            $ca[$key] = $ev;
        }
        else {
            if($req) {
                die("Environment variable ".$env_var." must be set!".LF);
            }
            $ca[$key] = null;
        }
    }
    
   return $ca;
}

/* Remove
// Connect to database.
function connect_to_database($host,$user,$password,$database,$port,$socket) {
    $conn = mysqli_init();
    if (!$conn) die("Failed to initialize database connection");
    if (!$conn->options(MYSQLI_OPT_LOCAL_INFILE, 1)) die("Failed to set LOAD LOCAL option");
    if (!$conn->real_connect($host,$user,$password,$database,$port,$socket)) die("Unable to connect to database");
    return $conn;
}
End remove.*/

// Connect to database.
// Input: An array with connection parameters (see get_conn_array).
// Return: MySQL connection.
function connect_to_database_array($ca) {
    $conn = mysqli_init();
    if (!$conn) die("Failed to initialize database connection");
    if (!$conn->options(MYSQLI_OPT_LOCAL_INFILE, 1)) die("Failed to set LOAD LOCAL option");
    if (!$conn->real_connect(
        $ca["host"],$ca["user"],$ca["password"],$ca["database"],$ca["port"],$ca["socket"])) die("Unable to connect to database");
    return $conn;
}

// Create a column name with data type to be used in creating dataset_* table.
// Input: $name, string with the base part of the name.
//        $num, an array with numbers.
//        $type, the MySQL data type.
// Return: The column name.
function create_column_name($name,$num,$type) {
    $col_name = $name;
    if(count($num)>0) {
        foreach($num as $n) {
            $col_name .= "_".$n;
        }
    }
    $col_name .= " ".$type.",";
    return $col_name;
}

// Clean the skeleton file to enable further parsing.
// Input: $sk_file_name, The skeleton file name.
// Return: An array with the contents of the skeleton file.
function clean_skeleton($sk_file_name) {
    // Read in all rows and make the data consistent.
    $sk_file=fopen($sk_file_name,"r");
    $row_num=0;
    while($row=fgets($sk_file)) {
        // Remove leading and trailing spaces.
        $row=trim($row);
        // Skip comment lines and blank lines.
        if(strlen($row)>0 & preg_match("/^!/",$row)==0) {
            // If the data is split over several rows then merge.
            // First check if it just one string enclosed in "".
            if(preg_match('/".+"$/',$row)) {
                // Then check if the line following is a continuation.
                $row_next=trim(fgets($sk_file));
                // Is it a global attribute continuation or a variable attribute continuation?
                if(preg_match('/\d+:.*/',$row_next) | preg_match('/\w+\s+.+/',$row_next))
                    $row .= " ".$row_next;
                // Else treat it as a new row.
                else {
                    $row_all[$row_num] = $row;
                    $row_num++;
                    $row = $row_next;
                }
            }            
            // If row ends with "-" then read next row and append to current row.
            while(preg_match('/(.*)"\s+\-$/',$row,$match_1)) {
                $row_next=trim(fgets($sk_file));
                preg_match('/"(.+)/',$row_next,$match_2);
                $row = $match_1[1].$match_2[1];
            }
            // Add rows to row_all.
            $row_all[$row_num]=$row;
            $row_num++;
        }
    }    
    fclose($sk_file);
    return $row_all;
}

/**
* Get the data and field names for a set of variables for a dataset.
*
* @param mysqli $conn A MySQL connection.
* @param string $dataset The SEDAT name of the dataset.
* @param double $epoch_start_cdf Start epoch in CDF epoch format.
* @param double $epoch_end_cdf End epoch in CDF epoch format.
* @param int $max_num_rec Maximum number of records to be read.
* @param array $meta An array with variable metadata (name,dims,sizes).
* @return array An array with a MySQL result set and field name for each variable.
* @author Peter Wintoft.
*/
function get_data_and_fields($conn,$dataset,$epoch_start_cdf,$epoch_end_cdf,$max_num_rec,$meta) {

    $q = "SELECT name FROM dataset WHERE sedat_name=".quote($dataset);
    if (!($res=$conn->query($q))) printf("Query: %s\nError: %s\n",$q,$conn->error);
    $row = $res->fetch_row();
    $dataset = $row[0];

    $field_name = array();
    $q = "SELECT ";
    foreach($meta as $m) {
        switch ($m["dims"]) {
            case 0:
                if(!strcasecmp("epoch",$m["name"])==TRUE) {
                    $q.= "concat(epoch,'.',lpad(millisec,3,'0')),";
                }
                else {
                    $q.= s_to_us(strtolower($m["name"])).",";
                }
                $field_name[] = $m["name"];
                break;
            case 1:
                $nrv_values = get_nrv_values($conn,$dataset,$m["name"]);
                for($i=1;$i<=$m["sizes"];$i++) {
                    $q.= s_to_us(strtolower($m["name"]))."_".$i.",";
                    $field_name[] = $m["name"]." (".$nrv_values[$i-1].")";
                }
                break;
            case 2:
                $nrv_values = get_nrv_values($conn,$dataset,$m["name"]);
                list($i_max,$j_max) = explode($m["sizes"],",");
                $k=0;
                for($i=1;$i<=$i_max;$i++) {
                    for($j=1;$j<=$j_max;$j++) {
                        $q.= s_to_us(strtolower($m["name"]))."_".$i."_".$j.",";
                        $field_name[] = $m["name"]." (".$nrv_values[$k++].")";
                    }
                }
                break;
        }
    }
    $q = trim($q,",");
    $q.= " FROM ".$dataset;
    $q.= " WHERE cdf_epoch>=".sprintf("%f",$epoch_start_cdf);
    $q.= " AND cdf_epoch<=".sprintf("%f",$epoch_end_cdf);
    $q.= " ORDER BY cdf_epoch";
    $q.= " LIMIT ".no_quote($max_num_rec);
        
    if (!($res=$conn->query($q))) printf("Query: %s\nError: %s\n",$q,$conn->error);

    return array($res,$field_name);
}

/**
* Get the metadata for a list of variables.
*
* @param mysqli $conn A MySQL connection.
* @param string $dataset The SEDAT name of the dataset.
* @param array $var An array with variable names.
* @return array An array with metadata for each variable.
* @author Peter Wintoft.
*/
function get_variable_metadata($conn,$dataset,$var) {
    $meta = array();
    foreach($var as $v) {
        $q = "SELECT vm.name,vm.dims,vm.sizes FROM variable_metadata AS vm";
        $q.= " LEFT JOIN dataset ON vm.dataset_id=dataset.id";
        $q.= " WHERE dataset.sedat_name=".quote($dataset);
        $q.= " AND vm.name=".quote($v);
        if (!($res=$conn->query($q))) printf("Query: %s\nError: %s\n",$q,$conn->error);
        if($row=$res->fetch_assoc()) {
            $meta[] = $row;
        }
        else {
            printf("Query: %s\nError: No data!\n",$q);
        }
    }
    return $meta;
}

/**
* A function for reading ODI data into an array.
*
* @param string $dataset The SEDAT name of the dataset.
* @param string $epoch_start The start epoch in format "YYYY-MM-DD hh:mm:ss.millisec".
* @param string $epoch_end The end epoch in format "YYYY-MM-DD hh:mm:ss.millisec".
* @param int $max_num_rec The maximum number of records to be read.
* @param string $var The name of a variable to be read.
* @param string $var,... Unlimited optional variables to be read.
* @return array A 2-dimensional array with data. The record number increases along the first dimension, and the variables along the second dimension.
*/
function odi_read_data() {

    $arg_list = func_get_args();
    $dataset = array_shift($arg_list);
    $epoch_start = array_shift($arg_list);
    $epoch_end = array_shift($arg_list);
    $max_num_rec = array_shift($arg_list);
    $var = $arg_list;

    $data = array();
    
    // Connect to the database at user level 3.
    $conn_array=get_conn_array(3);
    $conn=connect_to_database_array($conn_array);

    // Convert epoch range to CDF epoch.
    list($date,$ms) = explode(".",$epoch_start);
    $epoch_start_cdf = date_to_cdfepoch($date,$ms);
    list($date,$ms) = explode(".",$epoch_end);
    $epoch_end_cdf   = date_to_cdfepoch($date,$ms);

    // Get the metadata for each variable.
    $meta = get_variable_metadata($conn,$dataset,$var);
    
    // Get the data and field names.
    list($res,$field_name) = get_data_and_fields($conn,$dataset,$epoch_start_cdf,$epoch_end_cdf,$max_num_rec,$meta);

    while($row=$res->fetch_row()) {
        $data[] = $row;
    }
    
    return $data;
}

// A function to create the ODI metadata from a CDF skeleton table.
// The metadata are stored into the ODI tables dataset, dataset_metadata,
// variable_metadata, and variable_attribute.
// Input: $conn, A MySQL database connection.
//        $dataset_name, The name of the dataset (e.g. dataset_index_Dst).
//        $platform, The platform (e.g. index, XMM, ...).
//        $platform_type, The platform type (e.g. virtual, satellite, ...).
//        $instrument, The instrument (e.g. Dst, ERMD, ...).
//        $sk_file_name, The skeleton file name.
//        $sedat_name, The SEDAT name.
//        $sedat_name, The SEDAT name.
//        $sedat_name, The SEDAT name.
//        $sedat_name, The SEDAT name.
//        $sedat_name, The SEDAT name.
// Peter Wintoft, Swedish Institute of Space Physics.
// Created: 2009-05-09.
// Updated: 2009-06 (D. Heynderickx)
function skt2odi($conn, $dataset_name, $platform, $platform_type, $instrument, $sk_file_name,
                 $sedat_name, $sedat_regions, $sedat_particles, $sedat_emin, $sedat_emax)
{
    
    $test_mode = FALSE;
    
    if($test_mode) {
        fwrite(STDOUT,"===========================================\n");
        fwrite(STDOUT,"Test mode.\n");
        fwrite(STDOUT,"Printing contents of cleaned skeleton file.\n");
        fwrite(STDOUT,"-------------------------------------------\n");
    }
    
    // Clean the skeleton file and put into an array.
    $row_all = clean_skeleton($sk_file_name);
    
    if($test_mode) {
        fwrite(STDOUT,"-------------------------------------------\n");
        fwrite(STDOUT,"Finished printing skeleton file.\n\n\n\n\n\n\n\n");
        fwrite(STDOUT,"===========================================\n");
        fwrite(STDOUT,"Listing interpreted skeleton file.\n");
        fwrite(STDOUT,"-------------------------------------------\n");
    }

    // Also check if the "dataset_*" table already exists, then delete it.
    // This can happen even when there is no entry in "dataset" table.
    if(!$test_mode) {
        $q = 'SHOW TABLES LIKE '.quote($dataset_name);
        if (!($res=$conn->query($q))) printf("Query: %s\nError: %s\n",$q,$conn->error);
        $row=$res->fetch_assoc();
        if($row) {
            fwrite(STDOUT,$dataset_name." exists.  Deleting ...\n");
            $q = 'DROP TABLE '.$dataset_name;
            if (!($res=$conn->query($q))) exit("Query: %s\nError: %s\n".$q.$conn->error);
        }
    
        // Insert an entry into the dataset table.
        fwrite(STDOUT,"Creating ".$dataset_name." and metadata ...\n");
        $dataset_id = insert_dataset_entry($conn,$dataset_name);
    
        // Add the SEDAT specific metadata
        $meta = array("dataset" => $dataset_id,
                      "attribute" => "sedat_name",
                      "value" => $sedat_name);
        update_dataset_entry($conn, $meta);
    
        $meta["attribute"] = "sedat_regions";
        $meta["value"] = $sedat_regions;
        update_dataset_entry($conn, $meta);
    
        $meta["attribute"] = "sedat_particles";
        $meta["value"] = $sedat_particles;
        update_dataset_entry($conn, $meta);
    
        $meta["attribute"] = "sedat_emin";
        $meta["value"] = $sedat_emin;
        update_dataset_entry($conn, $meta);
    
        $meta["attribute"] = "sedat_emax";
        $meta["value"] = $sedat_emax;
        update_dataset_entry($conn, $meta);
    
        // Insert the platform.
        $global = array();
        $global["dataset"]=$dataset_id;
        $global["attribute"]="platform";
        $global["entry_number"]=1;
        $global["data_type"]="CDF_CHAR";
        $global["value"]=$platform;
        insert_dataset_metadata_entry($conn,$global);
    
        // Insert the platform type.
        $global["attribute"]="platform_type";
        $global["value"]=$platform_type;
        insert_dataset_metadata_entry($conn,$global);
    
        // Insert the instrument.
        $global["attribute"]="instrument";
        $global["value"]=$instrument;
        insert_dataset_metadata_entry($conn,$global);
    }
    
    // Loop over all rows, interpret, and store into database.
    $block="";          // Block flag: head, global, zvariable, ...
    $in_block=FALSE;    // True when in a block.
    $nrv_index=0;       // Index used to count nrv values when they exist.
    foreach($row_all as $row) {
        // Get lines starting with "#" indicating start of block.
        if(preg_match("/#(\w+)/",$row,$match)>0) {
            $block=$match{1};
            #But still not in block.
            $in_block=FALSE;
        }
        // Read data if in a block.
        if($in_block) {
            // Check type of block.
            switch($block) {
                // header block.
                case "header":
                    // Get tokens separated by ":".
                    $meta=array();
                    $meta["dataset"]=$dataset_id;
                    $meta["attribute"]=strtolower(trim(strtok($row,":")));
                    $meta["value"]=trim(strtok(":"));
                    if(strlen($meta["value"])>0 & strcasecmp($meta["attribute"],"cdf name")!=0) {
                        if($test_mode) {
                            fwrite(STDOUT,"================\nHeader block\n-------------\n");
                            print_r($meta);
                        }
                        else {
                            update_dataset_entry($conn,$meta);
                        }
                    }
                    // Get the variables count line.
                    if(preg_match_all('/(\d+)\/(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\/z\s+(\d+)(.*)/',$row,$match)) {
                        $variable_count=array();
                        $variable_count["rVariables"]=$match[1][0];
                        $variable_count["zVariables"]=$match[2][0];
                        $variable_count["gAttributes"]=$match[3][0];
                        $variable_count["vAttributes"]=$match[4][0];
                        $variable_count["records"]=$match[5][0];
                        $variable_count["dims"]=$match[6][0];
                        if(isset($match[7][0])) {
                            $variable_count["sizes"]=trim($match[7][0]);
                        }
                        else {
                            $variable_count["sizes"]=NULL;
                        }
                        if($test_mode) {
                            fwrite(STDOUT,"===================\nVariable count line\n=============\n");
                            print_r($variable_count);
                        }
                    }
                    break;
                // Global attributes block.
                case "GLOBALattributes":
                    // Extract the global attributes.
                    // If the row starts with a digit then it is the same
                    // attribute with a new entry number.
                    if(preg_match_all('/^(\d+):\s+(.+)\s+\{\s+(.*)\s+\}/',$row,$match)) {
                        $global["entry_number"]=$match[1][0];
                        $global["data_type"]=$match[2][0];
                        $global["value"]=$match[3][0];
                    }
                    else {
                        preg_match_all('/"(.+)"\s+(\d+):\s+(.+)\s+\{\s+(.*)\s+\}/',$row,$match);
                        $global["attribute"]=$match[1][0];
                        $global["entry_number"]=$match[2][0];
                        $global["data_type"]=$match[3][0];
                        $global["value"]=$match[4][0];
                    }
                    if($test_mode) {
                        fwrite(STDOUT,"================\nGlobal attributes block\n-------------\n");
                        print_r($global);
                    }
                    else {
                        insert_dataset_metadata_entry($conn,$global);
                    }
                    break;
                // rVariable block.
                case "variables":
                    // rVariable.
                    if(preg_match_all('/"(.+)"\s+(\w+)\s+(\d+)\s+(\w)(.*)/',$row,$match)) {
                        $var = array();
                        $var["dataset"]=$dataset_id;
                        $var["name"]=$match[1][0];
                        $var["data_type"]=$match[2][0];
                        $var["number_elements"]=$match[3][0];
                        $var["record_variance"]=$match[4][0];
                        $var["dimension_variances"]=str_replace(" ",",",trim($match[5][0]));
                        $var["dims"]=$variable_count["dims"];
                        $var["sizes"]=$variable_count["sizes"];
                        if($test_mode) {
                            fwrite(STDOUT,"================\nrVariable block\n-------------\n");
                            print_r($var);
                        }
                        else {
                            insert_variable_metadata_entry($conn,$var);
                        }
                        $nrv_index=0;
                    }
                    // Get variable attributes.
                    if(preg_match_all('/("\w+")\s+(\w+)\s+\{\s+(.*)\s+\}/',$row,$match)) {
                        $varatt = array();
                        $varatt["variable_metadata"]=$var["name"];
                        $varatt["attribute"]=$match[1][0];
                        $varatt["data_type"]=$match[2][0];
                        $varatt["value"]=$match[3][0];
                        if($test_mode) {
                            fwrite(STDOUT,"================\nVariable attributes block\n-------------\n");
                            print_r($varatt);
                        }
                        else {
                            insert_variable_attribute_entry($conn,$varatt,$dataset_id);
                        }
                    }
                    // Get nrv values.
                    if(preg_match_all('/^\[(.*)\]\s*=\s*\{*\s*(.*)\s*\}*/',$row,$match)) {
                        // Remove a possible trailing "}".
                        $nrv_index++;
                        $nrv_pos=trim($match[1][0]," }");
                        $nrv_value=trim($match[2][0]," }");
                        if($test_mode) {
                            fwrite(STDOUT,"================\nNRV values\n-------------\n");
                            fwrite(STDOUT,$var["name"]." <".$nrv_index."> <".$nrv_pos."> <".$nrv_value.">\n");
                        }
                        else {
                            insert_variable_nrv_entry($conn,$var["name"],$nrv_index,$nrv_pos,$nrv_value,$dataset_id);
                        }
                    }
                    break;
                // zVariable block.
                case "zVariables":
                    // Variable name row always contain "Number Elements" and "Dims".
                    if(preg_match_all('/"(.+)"\s+(\w+)\s+(\d+)\s+(\d+)\s+(.*)/',$row,$match)) {
                        $var = array();
                        $var["dataset"]=$dataset_id;
                        $var["name"]=$match[1][0];
                        $var["data_type"]=$match[2][0];
                        $var["number_elements"]=$match[3][0];
                        $var["dims"]=$match[4][0];
                        $rest=$match[5][0];
                        if($var["dims"]>0) {
                            $var["sizes"]="";
                            for($i=0;$i<$var["dims"];$i++) {
                                preg_match_all('/(\d+)\s+(.*)/',$rest,$match);
                                $var["sizes"] .= $match[1][0].",";
                                $rest=$match[2][0];
                            }
                            $var["sizes"]=trim($var["sizes"],",");
                            $rest=$match[2][0];
                        }
                        else {
                            $var["sizes"]="";
                        }
                        if(preg_match_all('/(\w)\s+(.*)/',$rest,$match)) {
                            $var["record_variance"]=$match[1][0];
                            $var["dimension_variances"]=str_replace(" ",",",trim($match[2][0]));
                        }
                        else {
                            $var["record_variance"]=$rest;
                            $var["dimension_variances"]="";
                        }
                        if($test_mode) {
                            fwrite(STDOUT,"================\nzVariable block\n-------------\n");
                            print_r($var);
                        }
                        else {
                            insert_variable_metadata_entry($conn,$var);
                        }
                        $nrv_index=0;
                    }
                    // Get variable attributes.
                    if(preg_match_all('/("\w+")\s+(\w+)\s+\{\s+(.*)\s+\}/',$row,$match)) {
                        $varatt = array();
                        $varatt["variable_metadata"]=$var["name"];
                        $varatt["attribute"]=$match[1][0];
                        $varatt["data_type"]=$match[2][0];
                        $varatt["value"]=$match[3][0];
                        if($test_mode) {
                            fwrite(STDOUT,"================\nVariable attributes block\n-------------\n");
                            print_r($varatt);
                        }
                        else {
                            insert_variable_attribute_entry($conn,$varatt,$dataset_id);
                        }
                    }
                    // Get nrv values.
                    if(preg_match_all('/^\[(.*)\]\s*=\s*\{*\s*(.*)\s*\}*/',$row,$match)) {
                        // Remove a possible trailing "}".
                        $nrv_index++;
                        $nrv_pos=trim($match[1][0]," }");
                        $nrv_value=trim($match[2][0]," }");
                        if($test_mode) {
                            fwrite(STDOUT,"================\nNRV values\n-------------\n");
                            fwrite(STDOUT,$var["name"]." <".$nrv_index."> <".$nrv_pos."> <".$nrv_value.">\n");
                        }
                        else {
                            insert_variable_nrv_entry($conn,$var["name"],$nrv_index,$nrv_pos,$nrv_value,$dataset_id);
                        }
                    }
                    break;
            } // End switch.
        }
        $in_block=TRUE;
    }

    if(!$test_mode) {
        // Create dataset table.
        create_dataset_table($conn,$dataset_id,$dataset_name);
    }

    fwrite(STDOUT,"Done!\n");

}


// Create a "dataset_*" table.
// The column names are constructed from "data" (CDF VAR_TYPE="data") variable names and
// the NRV index if it is larger than 0.
// The tables always have the columns cdf_epoch, epoch, millisec, and dataset_file_id.
// Updated version with improved query syntax (DH)
function create_dataset_table($conn,$dataset_id,$dataset_name) {
    
    // Get the dataset id and majority.
    $q = 'SELECT majority FROM dataset WHERE id=' . $dataset_id;
    if (!($result=$conn->query($q))) printf("Query: %s\nError: %s\n",$q,$conn->error);
    $row = $result->fetch_assoc();
    $first_dim_slow = TRUE;         // If ROW_MAJOR then the first dimension changes the slowest.
    if(strncasecmp($row["majority"],"col",3) == 0) {
        $first_dim_slow = FALSE;    // If COLUMN_MAJOR then the first dimension changes the fastest.
    }
    
    // Create columns for all variables that are record variant except Epoch.
    $q  = 'SELECT vm.name AS column_name,';
    $q .= ' vm.sizes AS size,';
    $q .= ' dt.mysql_type AS data_type,';
    $q .= ' vm.dimension_variances AS dim_var';
    $q .= ' FROM variable_metadata AS vm';
    $q .= ' LEFT JOIN data_type AS dt ON (vm.data_type_id=dt.id)';
    $q .= ' WHERE vm.dataset_id=' . $dataset_id;
    $q .= ' AND vm.record_variance="T"';
    $q .= ' AND STRCMP(LOWER(vm.name),"epoch")!=0';
    if (!($result=$conn->query($q))) printf("Query: %s\nError: %s\n",$q,$conn->error);
    $q = 'CREATE TABLE '.$dataset_name.' ( ';
    $q .= 'cdf_epoch DOUBLE NOT NULL,';
    $q .= 'epoch DATETIME,';
    $q .= 'millisec SMALLINT,';
    while($row=$result->fetch_assoc()) {
        // Get the name part and the data type to form the query.
        $col_name = s_to_us(strtolower($row["column_name"]));
        $data_type = $row["data_type"];
        // If "size" is set then it is a multidimensional variable else it is a scalar.
        if($row["size"]) {
            $size=explode(",",$row["size"]);
            switch(count($size)) {
                case 1:
                    for($i=1;$i<=$size[0];$i++) {
                        $q .= create_column_name($col_name,array($i),$data_type);
                    }
                    break;
                case 2:
                    if($first_dim_slow) {
                        for($i0=1;$i0<=$size[0];$i0++) {
                            for($i1=1;$i1<=$size[1];$i1++) {
                                $q .= create_column_name($col_name,array($i0,$i1),$data_type);
                            }
                        }
                    }
                    else {
                        for($i1=1;$i1<=$size[1];$i1++) {
                            for($i0=1;$i0<=$size[0];$i0++) {
                                $q .= create_column_name($col_name,array($i0,$i1),$data_type);
                            }
                        }
                    }
                    break;
                default:
                    fwrite(STDERR,"Dimension > 2 not supported!\n");
                    exit(0);
            }
        }
        else {
            $q .= create_column_name($col_name,null,$data_type);
        }
    }
    $q .= 'dataset_file_id INT UNSIGNED,';
    $q .= 'PRIMARY KEY(cdf_epoch),';
    $q .= 'KEY(dataset_file_id)';
    $q .= ') ENGINE=INNODB';
    if (!($conn->query($q))) printf("Query: %s\nError: %s\n",$q,$conn->error);
    $q = 'ALTER TABLE '.$dataset_name.' ADD FOREIGN KEY (dataset_file_id) ';
    $q .= 'REFERENCES dataset_file (id) ON DELETE CASCADE';
    if (!($conn->query($q))) printf("Query: %s\nError: %s\n",$q,$conn->error);
}



// Delete the "dataset_*" table together with all related entries.
function delete_dataset($conn,$dataset_name) {
    $q = 'DROP TABLE '.no_quote($dataset_name);
    if (!($conn->query($q))) printf("Query: %s\nError: %s\n",$q,$conn->error);
    $q = 'DELETE FROM dataset WHERE name='.quote($dataset_name);
    if (!($conn->query($q))) printf("Query: %s\nError: %s\n",$q,$conn->error);
}

// True if dataset name exists in dataset table.
function exists_dataset($conn,$dataset_name) {
    $q = 'SELECT COUNT(id) AS num FROM dataset WHERE name='.quote($dataset_name);
    if (!($res=$conn->query($q))) printf("Query: %s".LF."Error: %s".LF,$q,$conn->error);
    $row=$res->fetch_assoc();
    return ($row["num"]>0);
}

// Command line arguments.
function get_cl_arguments($argv,$arg_array) {
    // Check number of input arguments.
    if (count($argv)<(sizeof($arg_array)+1)) {
        fwrite(STDOUT,"Error: Too few input arguments.\n");
        fwrite(STDOUT,"Usage: ".$argv[0]);
        foreach($arg_array as $a)
            fwrite(STDOUT," <".$a.">");
        fwrite(STDOUT,"\n");
        exit(0);
    }

    // Input parameters.
    array_shift($argv); #First argument is the script name.
    foreach($arg_array as $a)
        $args[$a]=array_shift($argv);
    return $args;
}

// Insert a row in the dataset table.
function insert_dataset_entry($conn,$dataset_name) {
    $dataset_id = FALSE;
    $q = 'INSERT INTO dataset (name,creation_date)';
    $q .= ' VALUES ('.quote($dataset_name).',NOW())';
    if (!($conn->query($q)))
      printf("Query: %s".LF."Error: %s".LF,$q,$conn->error);
    else
      $dataset_id = $conn->insert_id;

    return $dataset_id;
}    

// Insert a row in the dataset_metadata table.
function insert_dataset_metadata_entry($conn,$meta) {
    $q = 'INSERT IGNORE INTO dataset_metadata';
    $q .= ' (dataset_id,attribute,data_type_id,entry_number,value)';
    $q .= ' SELECT ' . $meta["dataset"] . ',';
    $q .= quote($meta["attribute"]).',data_type.id,'.$meta["entry_number"].',';
    $q .= quote($meta["value"]);
    $q .= ' FROM data_type ';
    $q .= ' WHERE data_type.cdf_type='.quote($meta["data_type"]);
    if (!($conn->query($q))) printf("Query: %s".LF."Error: %s".LF,$q,$conn->error);
}

// Insert a row in the nrv_value table.
function insert_variable_nrv_entry($conn,$variable_name,$nrv_index,$nrv_pos,$nrv_value,$dataset_id) {
    $q = 'INSERT IGNORE INTO variable_nrv (variable_metadata_id,nrv_index,nrv_pos,nrv_value) ';
    $q .= 'SELECT variable_metadata.id,'.$nrv_index.','.quote($nrv_pos).','.quote($nrv_value);
    $q .= ' FROM variable_metadata WHERE';
    $q .= ' variable_metadata.name='.quote($variable_name);
    $q .= ' AND variable_metadata.dataset_id=' . $dataset_id;
    if (!($conn->query($q))) printf("Query: %s".LF."Error: %s".LF,$q,$conn->error);
}

// Insert a row in the variable_metadata table.
function insert_variable_metadata_entry($conn,$metadata) {
    $q = 'INSERT IGNORE INTO variable_metadata ';
    $q .= '(name,dataset_id,data_type_id,number_elements,dims,sizes,record_variance,dimension_variances) ';
    $q .= 'SELECT ';
    $q .= '"'.$metadata["name"].'",';
    $q .= '"'.$metadata["dataset"].'",';
    $q .= 'data_type.id,';
    $q .= $metadata["number_elements"].',';
    $q .= $metadata["dims"].',';
    $q .= '"'.$metadata["sizes"].'",';
    $q .= '"'.$metadata["record_variance"].'",';
    $q .= '"'.$metadata["dimension_variances"].'" ';
    $q .= 'FROM data_type ';
    $q .= 'WHERE data_type.cdf_type="'.$metadata["data_type"].'"';
    if (!($conn->query($q)))
        printf("Query: %s\nError: %s\n",$q,$conn->error);
}

// Insert a row in the variable_attribute table.
function insert_variable_attribute_entry($conn,$metadata,$dataset_id) {
    $q = 'INSERT IGNORE INTO variable_attribute ';
    $q .= '(name,variable_metadata_id,data_type_id,value) ';
    $q .= 'SELECT '.quote($metadata["attribute"]).',variable_metadata.id';
    $q .= ',data_type.id,'.quote($metadata["value"]);
    $q .= ' FROM variable_metadata, data_type';
    $q .= ' WHERE variable_metadata.name='.quote($metadata["variable_metadata"]);
    $q .= ' AND variable_metadata.dataset_id=' . $dataset_id;
    $q .= ' AND data_type.cdf_type='.quote($metadata["data_type"]);
    if (!($conn->query($q))) printf("Query: %s\nError: %s\n",$q,$conn->error);
}


function update_dataset_entry($conn,$meta) {
    $q = 'UPDATE dataset SET '.no_quote(s_to_us($meta["attribute"])).'='.quote($meta["value"]);
    $q .= ' WHERE dataset.id='.$meta["dataset"];
    if (!($conn->query($q))) printf("Query: %s\nError: %s\n",$q,$conn->error);
}

/**
* Get the NRV values for a variable.
* If the LABL_PTR_? attribute is not defined the returned values will just be
* an array of number.
*
* @param ? $conn
* @param string $dataset
* @param string $variable_name
* @return array $nrv_values
* @author Peter Wintoft
*/
function get_nrv_values($conn,$dataset,$variable_name) {
    // Get the sizes of the variable.
    $q = "SELECT vm.id AS id, vm.sizes AS sizes";
    $q.= " FROM variable_metadata AS vm";
    $q.= " LEFT JOIN dataset AS ds ON vm.dataset_id=ds.id";
    $q.= " WHERE ds.name=".quote($dataset);
    $q.= " AND vm.name=".quote($variable_name);
    if (!($res=$conn->query($q))) printf("Query: %s\nError: %s\n",$q,$conn->error);
    $var_id=0;
    $n=0;
    if($row=$res->fetch_assoc()) {
        $var_id=$row["id"];
        $sizes = explode(",",$row["sizes"]);
        $n = array_product($sizes);
    }
    $nrv_values = array();
    // Get the NRV values.
    $q = "SELECT va.value AS nrv_var";
    $q.= " FROM variable_attribute AS va";
    $q.= " WHERE va.variable_metadata_id=".$var_id;
    $q.= " AND va.name='labl_ptr_1'";
    if (!($res=$conn->query($q))) printf("Query: %s\nError: %s\n",$q,$conn->error);
    if($row=$res->fetch_assoc()) {
        $q = "SELECT nrv.nrv_value FROM variable_nrv AS nrv";
        $q.= " LEFT JOIN variable_metadata AS vm ON nrv.variable_metadata_id=vm.id";
        $q.= " WHERE vm.name=".quote($row["nrv_var"]);
        if (!($res=$conn->query($q))) printf("Query: %s\nError: %s\n",$q,$conn->error);
        while($row=$res->fetch_row()) {
            $nrv_values[] = trim($row[0]);
        }
    }
    if(count($nrv_values)<$n) {
        $nrv_values = range(1,$n);
    }
    return $nrv_values;
}

// A function to ensure that a string is enclosed in quotes (").
function quote($s) {
    return '"'.trim($s,'"').'"';
}

// A function to remove quotes (").
function no_quote($s) {
    return trim($s,'\'"');
}

// A function to replace space " " with underscore "_".
function s_to_us($s) {
// New code.
    return trim(preg_replace('/\W+/','_',$s),'_');
// End new code.
}



  function date_to_cdfepoch($date, $msec)
  {
//
// Convert date, msec to cdfepoch
//
// Inputs: $date  UTC date string of the form YYYY-MM-DD hh:mm:ss
//         $msec  milliseconds
//
// Returns: cdfepoch as float
//
// Created: 05/05/2009 by D. Heynderickx (DH Consultancy)
//
    $cdf_1970 = 62167219200000.0;

    $epoch = 1000.0 * gmmktime(substr($date, 11, 2), substr($date, 14, 2), substr($date, 17, 2),
                               substr($date, 5, 2), substr($date, 8, 2), substr($date, 0, 4));
    $epoch += $cdf_1970 + $msec;

    return $epoch;
  }
  
  /**
  * Convert CDF Epoch to UTC date string of the form YYYY-MM-DD hh:mm:ss
  *
  * @param double $cdf_epoch
  * @return array ($epoch,$millisec)
  * @author Peter Wintoft
  */
  function cdfepoch_to_date($cdf_epoch) {
//    $epoch = $cdf_epoch - CDF1970;
    $epoch = $cdf_epoch - 62167219200000.0;
    $time = (int) ($epoch/1000.0);
    $ms = $epoch - $time*1000;
    $date = gmdate("Y-m-d H:i:s",$time);
    return array($date,$ms);
  }


// Check if dataset is listed in table dataset and return dataset_id
  function checkTable($link, $tablename, $lhandle)
  {
    $tableid = FALSE;

    $query = "SELECT * FROM dataset WHERE " .
             "name='" . $tablename . "'";
    $result = mysqli_query($link, $query);
    if (mysqli_num_rows($result) > 0)
    {
      $row = mysqli_fetch_array($result, MYSQLI_ASSOC);
      $tableid = $row["id"];
      mysqli_free_result($result);
    }
    else
      fprintf($lhandle, "  Table " . $tablename . " not found" . LF);

    return $tableid;
  }


// Produce a list of files already ingested for the dataset
  function tfileList($link, $pattern)
  {
    $files = array();
    $query = "SELECT * FROM dataset_file WHERE filename LIKE '" . $pattern . "'";
    $result = mysqli_query($link, $query);

    while ($row = mysqli_fetch_array($result, MYSQLI_ASSOC))
    {
      $files["id"][] = $row["id"];
      $files["dfname"][] = $row["filename"];
      $files["fdate"][] = $row["filedate"];
    }
    mysqli_free_result($result);

    return $files;
  }



// Load temporary files load.tmp into dataset_* and update dataset and dataset_files
  function getFiles($link, $dfiles, $tfiles, $lHandle, $tablename, $tableid, $platform, $instrument, $sk_file)
  {
    $nf = 0;
    foreach ($dfiles as $dfile)
    {
      $fid = FALSE;

      $filename = $dfile[0];
      $filedate = $dfile[1];
      $mtime = strtotime($filedate);
      $rfname = $dfile[2];

      $getFile = false;
      $key = FALSE;
      if (count($tfiles) > 0) $key = array_search($filename, $tfiles["dfname"]);
      if ($key !== FALSE)
      {
        $fid = $tfiles["id"][$key];
        $sfiledate = $tfiles["fdate"][$key];            
        if ($mtime > strtotime($sfiledate)) $getFile = true;
      }
      else
        $getFile = true;

      if ($getFile)
      {
        $sqlerror = FALSE;
        fprintf($lHandle, "%s", $rfname . "  ");
        if ($fid !== FALSE)
        {
          $query = "UPDATE dataset_file SET filedate='" . $filedate . "' " .
                   "WHERE id=" . $fid . " LIMIT 1";
          mysqli_query($link, $query);
          if (mysqli_affected_rows($link) > 0)
            $status = "updated";
          else
          {
            $status = "update error";
            $sqlerror = mysqli_error($link);
          }
        }
        else
        {
          $query = "INSERT INTO dataset_file VALUES(DEFAULT,'" . $filename .
                   "'," . $tableid . ",'" . $filedate . "')";
          mysqli_query($link, $query);
          if (mysqli_affected_rows($link) > 0)
          {
            $fid = mysqli_insert_id($link);
            $status = "loaded";
          }
          else
          {
            $status = "insert error";
            $sqlerror = mysqli_error($link);
          }
        }
        if (strpos($status, "error") === FALSE)
        {
          $result = parseFile($rfname, $fid, $platform, $instrument, $sk_file);
          if ($result)
          {
            $query = "LOAD DATA LOCAL INFILE 'load.tmp' " .
                     "REPLACE INTO TABLE `" . $tablename .
                     "` FIELDS TERMINATED BY ',' " .
                     "OPTIONALLY ENCLOSED BY '\"' " .
                     "LINES TERMINATED BY '" . LF . "'";
            mysqli_query($link, $query);
            if (mysqli_affected_rows($link) < 0)
            {
              $status = "load error";
              $sqlerror = mysqli_error($link);
            }
            else
              $nf++;
            @unlink("load.tmp");
          }
          else
            $status = "read error";
        }
        if ((strpos($status, "error") !== FALSE) && ($fid !== FALSE))
        {
          $query = "DELETE FROM dataset_file WHERE id=" . $fid . " LIMIT 1";
          mysqli_query($link, $query);
        }
        fprintf($lHandle, "%s", $status);
        if ($sqlerror !== FALSE)
          fprintf($lHandle, "  %s", $sqlerror);
        fprintf($lHandle, "%s", LF);
      }
    }

    return $nf;
  }


// Update table dataset with start and end epoch and modification date
  function dataset_update($link, $tablename, $tableid)
  {
    $query = "SELECT MIN(cdf_epoch), MAX(cdf_epoch) FROM " . $tablename;
    $result = mysqli_query($link, $query);
    if (mysqli_num_rows($result) > 0)
    {
      $row = mysqli_fetch_array($result, MYSQLI_NUM);
      mysqli_free_result($result);
      $query = "UPDATE dataset SET epoch_start=" . $row[0] .
               ", epoch_end=" . $row[1] . ", modification_date=NOW() " .
               "WHERE id=" . $tableid . " LIMIT 1";
      mysqli_query($link, $query);
    }
  }


// Construct a list of files to be ingested, convert them to temporary file and ingest into the database
  function ingest_data($link, $tablename, $basedir, $platform, $instrument, $filepattern, $sk_file)
  {
    $nf = 0;
    $lname = $instrument . ".log";
    $lhandle = fopen($lname, "a");
    fprintf($lhandle, LF . "--" . LF . "-- " . date("d/m/Y H:i:s") . LF .
                      "--" . LF . LF);

    $tfiles = tfileList($link, $filepattern);
    
    $tableid = checkTable($link, $tablename, $lhandle);

    if ($tableid !== FALSE)
    {
      $dfiles = rfileList($platform, $instrument, $basedir, $filepattern);
      $nf = getFiles($link, $dfiles, $tfiles, $lhandle, $tablename, $tableid, $platform, $instrument, $sk_file);
      if ($nf > 0) dataset_update($link, $tablename, $tableid);
    }

    fclose($lhandle);

    return $nf;
  }

?>