# Create the tables to hold the ODI metadata.
#
# Peter Wintoft(1) and Daniel Heynderickx(2).
# (1) Swedish Institute of Space Physics.
# (2) DH Consultancy.
#
# Created: 2009-04-21.

###################################################################
#
# Create tables.
#
###################################################################

###################################################################
# Table: dataset.
# A table containing the names of the data tables.

CREATE TABLE dataset
(
    id INT UNSIGNED NOT NULL AUTO_INCREMENT,
    name VARCHAR(255) UNIQUE,
    epoch_start DOUBLE,
    epoch_end DOUBLE,
    creation_date DATETIME,
    modification_date DATETIME,
    data_encoding TEXT,
    majority TEXT,
    format TEXT,
    sedat_name TEXT,
    sedat_regions TEXT,
    sedat_particles TEXT,
    sedat_emin DOUBLE,
    sedat_emax DOUBLE,
    PRIMARY KEY (id)
) ENGINE=INNODB;

###################################################################

CREATE TABLE dataset_file
(
    id INT UNSIGNED NOT NULL AUTO_INCREMENT,
    filename TEXT,
    dataset_id INT UNSIGNED NOT NULL,
    filedate DATETIME,
    PRIMARY KEY (id),
    KEY (dataset_id)
) ENGINE=INNODB;

ALTER TABLE dataset_file
  ADD FOREIGN KEY (dataset_id) REFERENCES dataset (id) ON DELETE CASCADE;

###################################################################
# Table: data_type.
# A table containing the CDF data types. It also contains the equivalent MySQL types.
# Examples: CDF_CHAR, CDF_EPOCH.

CREATE TABLE data_type
(
    id INT UNSIGNED NOT NULL AUTO_INCREMENT,
    cdf_type TEXT,
    mysql_type TEXT,
    sedat_type TEXT,
    idl_type TEXT,
    PRIMARY KEY (id)
) ENGINE=INNODB;

###################################################################
# Table: dataset_metadata.
# A table containing the metadata related to a specific dataset.
# In the CDF skeleton these are the global attributes.

CREATE TABLE dataset_metadata
(
    id INT UNSIGNED NOT NULL AUTO_INCREMENT,
    dataset_id INT UNSIGNED,
    attribute TEXT,
    data_type_id INT UNSIGNED,
    entry_number INT UNSIGNED,
    value TEXT,
    PRIMARY KEY (id),
    KEY (dataset_id),
    KEY (data_type_id)
) ENGINE=INNODB;

ALTER TABLE dataset_metadata
  ADD FOREIGN KEY (dataset_id) REFERENCES dataset (id) ON DELETE CASCADE,
  ADD FOREIGN KEY (data_type_id) REFERENCES data_type (id) ON DELETE CASCADE;


###################################################################
# Table: variable_metadata.
# A table holding the CDF specific metadata for each variable in
# the database. Each variable belongs to a specific platform and
# category, and has a specific CDF data type.
#
# platform_id: key to the platform.
# category_id: key to the category.
# data_type_id: key to the CDF data type.
# id: unique id.
# name: the CDF variable name.
# number_elements: the CDF "Number Elements" which is "1" for all
#                  data types except CDF_CHAR where it is the length of the string.
# dims: the dimension. "0" for scalar, "1" for vector, "2" for matrix.
# sizes: the number of elements in each dimension. For a scalar it is the
#        empty string "". For a vector it is "n" where n>0, and for a
#        matrix it is "n m" where n,m>0.
# record_variance: either "T" or "F" for record variant or not.
# dimension_variances: for a scalar it is the empty string "",
#                      for a vector it is "F" or "T",
#                      and for a matrix it can be "F F", "F T", "T F", or "T T".

CREATE TABLE variable_metadata
(
    id INT UNSIGNED NOT NULL AUTO_INCREMENT,
    dataset_id INT UNSIGNED NOT NULL,
    data_type_id INT UNSIGNED NOT NULL,
    name TEXT NOT NULL,
    number_elements INT UNSIGNED NOT NULL,
    dims INT UNSIGNED NOT NULL,
    sizes TEXT,
    record_variance ENUM('F','T'),
    dimension_variances TEXT,
    PRIMARY KEY (id),
    KEY (dataset_id),
    KEY (data_type_id)
) ENGINE=INNODB;

ALTER TABLE variable_metadata
  ADD FOREIGN KEY (dataset_id) REFERENCES dataset (id) ON DELETE CASCADE,
  ADD FOREIGN KEY (data_type_id) REFERENCES data_type (id) ON DELETE CASCADE;


###################################################################
# Table: variable_attribute.
# A table holding the variable attributes.
# Compare with CDF variable attributes.

CREATE TABLE variable_attribute
(
    id INT UNSIGNED NOT NULL AUTO_INCREMENT,
    name TEXT,
    variable_metadata_id INT UNSIGNED NOT NULL,
    data_type_id INT UNSIGNED NOT NULL,
    value TEXT,
    PRIMARY KEY (id),
    KEY (variable_metadata_id),
    KEY (data_type_id)
) ENGINE=INNODB;

ALTER TABLE variable_attribute
    ADD FOREIGN KEY (variable_metadata_id) REFERENCES variable_metadata (id) ON DELETE CASCADE,
    ADD FOREIGN KEY (data_type_id) REFERENCES data_type (id) ON DELETE CASCADE;
    
#########

CREATE TABLE variable_nrv
(
    id INT UNSIGNED NOT NULL AUTO_INCREMENT,
    variable_metadata_id INT UNSIGNED,
    nrv_index INT UNSIGNED,
    nrv_pos TEXT,
    nrv_value TEXT,
    PRIMARY KEY (id),
    KEY (variable_metadata_id)
) ENGINE=INNODB;

ALTER TABLE variable_nrv
    ADD FOREIGN KEY (variable_metadata_id) REFERENCES variable_metadata (id) ON DELETE CASCADE;

###################################################################
# Populate the metadata tables with the "generic" metadata.
#
# Peter Wintoft, Swedish Institute of Space Physics.
# Daniel Heynderickx, DH Consultancy.

###################################################################
# Insert data into data_type.

INSERT INTO data_type (cdf_type,mysql_type,sedat_type,idl_type) VALUES
('CDF_BYTE', 'TINYINT', 'I', 'BYTE'),
('CDF_INT1', 'TINYINT', 'I', 'INT'),
('CDF_UINT1', 'TINYINT UNSIGNED', 'I', 'UINT'),
('CDF_INT2', 'SMALLINT', 'I', 'INT'),
('CDF_UINT2', 'SMALLINT UNSIGNED', 'I', 'UINT'),
('CDF_INT4', 'INT', 'I', 'LONG'),
('CDF_UINT4', 'INT UNSIGNED', 'I', 'ULONG'),
('CDF_REAL4', 'FLOAT', 'R', 'FLOAT'),
('CDF_FLOAT', 'FLOAT', 'R', 'FLOAT'),
('CDF_REAL8', 'DOUBLE', 'D', 'DOUBLE'),
('CDF_DOUBLE', 'DOUBLE', 'D', 'DOUBLE'),
('CDF_EPOCH', 'DATETIME_millisec_SMALLINT', 'D', 'DOUBLE'),
('CDF_EPOCH16', 'DATETIME_pico_BIGINT', 'D', 'DOUBLE'),
('CDF_CHAR', 'TEXT', 'C', 'STRING'),
('CDF_UCHAR', 'TEXT', 'C', 'STRING');
    